/* 
   Unix SMB/CIFS implementation.
   Character set conversion Extensions
   Copyright (C) Igor Vergeichik <iverg@mail.ru> 2001
   Copyright (C) Andrew Tridgell 2001
   Copyright (C) Simo Sorce 2001
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

*/
/* MKPROTO: libsamba_proto.h		(do not remove this line) */

#include "includes.h"
#include <errno.h>
#include "libsamba.h"
#include "byteorder.h"
#include "tng_misc.h"

#ifndef EILSEQ
#define EILSEQ EIO
#endif


/**
 * @file
 *
 * @brief Character-set conversion routines built on our iconv.
 * 
 * @note Samba's internal character set (at least in the 3.0 series)
 * is always the same as the one for the Unix filesystem.  It is
 * <b>not</b> necessarily UTF-8 and may be different on machines that
 * need i18n filenames to be compatible with Unix software.  It does
 * have to be a superset of ASCII.  All multibyte sequences must start
 * with a byte with the high bit set.
 *
 * @sa lib/iconv.c
 */

void tng_read_len_fixup_ascii(const char *str, int *len, int flags)
{
	DEBUG(4, ("tng_read_len_fixup_ascii: len==%d\n", *len));

	if (*len == 0)
		return;
	if (*len < -1)
	{
		DEBUG(1, ("WARNING: tng_read_len_fixup_ascii: "
			  "len==%d, flags=0%o\n", *len, flags));
		generate_backtrace();
		*len = 0;
		return;
	}
	if (*len == -1)
	{
		if (! (flags & STR_TERMINATE))
			DEBUG(1, ("WARNING: tng_len_fixup_ascii: len=-1, but "
				  "no STR_TERMINATE\n"));
		*len = strlen(str) + 1;
		return;
	}
	if (flags & STR_TERMINATE)
	{
		*len = strnlen(str, *len - 1) + 1;
	}
}

/*
 * str has to be already aligned!!
 * So please feed it aligned in!
 */
void tng_read_len_fixup_ucs2(const smb_ucs2_t *str, int align,
			     int *len, int flags)
{
	DEBUG(4, ("tng_read_len_fixup_ucs2: len==%d\n", *len));
	/* len must be at least one ucs2 character bigger than
	 * align (which is mostly a "fixup" constant here).
	 */
	if (((*len >= 0) && (*len < (align+2))))
	{
		*len = 0;
		return;
	}

	if (*len < -1)
	{
		DEBUG(1, ("WARNING: tng_read_len_fixup_ucs2: "
			  "len==%d, flags=0%o\n", *len, flags));
		generate_backtrace();
		*len = 0;
		return;
	}
	if (*len == -1)
	{
		if (! (flags & STR_TERMINATE))
			DEBUG(1, ("WARNING: tng_len_fixup_ucs2: len=-1, but "
				  "no STR_TERMINATE\n"));
		*len = strlen_w(str) + 1;
		*len *= 2;

		/* Ignore align, get out, we use the buffer's len */
		return;
	}

	*len -= align;

	*len &= ~1;

	if (flags & STR_TERMINATE)
	{
		*len = strnlen_w(str, *len/2 - 1) + 1;
		*len *= 2;
	}
}




static smb_iconv_t conv_handles[NUM_CHARSETS][NUM_CHARSETS];


/**
 * Return the name of a charset to give to iconv().
 **/
static const char *charset_name(charset_t ch)
{
	const char *ret = NULL;

	if (ch == CH_UCS2) ret = "UCS-2LE";
	else if (ch == CH_UNIX) ret = lp_unix_charset();
	else if (ch == CH_DOS) ret = lp_dos_charset();
	else if (ch == CH_DISPLAY) ret = lp_display_charset();
	else if (ch == CH_UTF8) ret = "UTF8";
	else if (ch == CH_UINT16) ret = "UCS-2-INTERNAL";


	if (!ret || !*ret) ret = "ASCII";
	return ret;
}


static smb_iconv_t conv_handles_get(charset_t from, charset_t to)
{
	static int initialized = False;
	const char *n1;
	const char *n2;
	smb_iconv_t descriptor;

	/* This block might go away? */
	if (!initialized) {
		initialized = True;

		charconv_reset();
	}

	descriptor = conv_handles[from][to];

	if (descriptor)
		return descriptor;

	n1 = charset_name(from);
	n2 = charset_name(to);

	descriptor = smb_iconv_open(n2,n1);
	if (descriptor == (smb_iconv_t)-1) {
		DEBUG(0,("Conversion from %s to %s not supported\n",
			 n1, n2));
		return NULL;
	}

	conv_handles[from][to] = descriptor;

	return descriptor;
}

/**
 * \brief Reset the charconv-tables to the empty state
 *
 * conv_handles_get reloads the table on demand.
 **/
void charconv_reset(void)
{
	int c1, c2;

	for (c1=0;c1<NUM_CHARSETS;c1++) {
		for (c2=0;c2<NUM_CHARSETS;c2++) {
			if (conv_handles[c1][c2])
				smb_iconv_close(conv_handles[c1][c2]);

			conv_handles[c1][c2] = NULL;
		}
	}
}


static BOOL suppress_next_error;

void convert_suppress_possible_error(BOOL set_or_clear_supress)
{
	suppress_next_error = set_or_clear_supress;
}


static void convert_report_error(charset_t from, charset_t to,
		      void const *src, size_t srclen, 
		      size_t destlen)
{
    	const char *reason="unknown error";
	int my_errno = errno;

	switch(my_errno) {
		case EINVAL:
			reason="Incomplete multibyte sequence";
			break;
		case E2BIG:
			reason="No more room"; 
		   	DEBUG(0, ("convert_string: Required %d, available %d\n",
				srclen, destlen));
			/* we are not sure we need srclen bytes,
			   may be more, may be less.
			   We only know we need more than destlen
			   bytes ---simo */
			break;
		case EILSEQ:
			reason="Illegal multibyte sequence";
			break;
	}
	DEBUG(0, ("Conversion error for %s -> %s: %s:\n", 
		  charset_name(from), charset_name(to), reason));
	dump_data(0, src, srclen);
	generate_backtrace();
}


/**
 * Convert string from one encoding to another, making error checking etc
 *
 * @param src pointer to source string (multibyte or singlebyte)
 * @param srclen length of the source string in bytes
 * @param dest pointer to destination string (multibyte or singlebyte)
 * @param destlen maximal length allowed for string
 * @returns the number of bytes occupied in the destination
 **/
size_t convert_string(charset_t from, charset_t to,
		      void const *src, size_t srclen, 
		      void *dest, size_t destlen)
{
	size_t i_len, o_len;
	size_t retval;
	const char* inbuf = (const char*)src;
	char* outbuf = (char*)dest;
	smb_iconv_t descriptor;

	if (srclen == (size_t)-1)
		srclen = strlen(src)+1;

	descriptor = conv_handles_get(from, to);

	if (descriptor == (smb_iconv_t)-1 || descriptor == (smb_iconv_t)0) {
		DEBUG(0,("convert_string_internal: Conversion not supported.\n"));
		goto use_as_is;
	}

	i_len=srclen;
	o_len=destlen;
	retval = smb_iconv(descriptor,  &inbuf, &i_len, &outbuf, &o_len);
	if(retval==(size_t)-1) {
		convert_report_error(from, to, src, srclen, destlen);
	}
	return destlen-o_len;

    use_as_is:
	/* conversion not supported, use as is */
	{
		size_t len = MIN(srclen,destlen);
		if (len)
			memcpy(dest,src,len);
		return len;
	}
}

/**
 * Convert between character sets, allocating a new buffer for the result.
 *
 * @param srclen length of source buffer.
 * @param dest always set at least to NULL
 * @note -1 is not accepted for srclen.
 *
 * @returns Size in bytes of the converted string; or -1 in case of error.
 **/

size_t convert_string_allocate(charset_t from, charset_t to,
		      		void const *src, size_t srclen, void **dest)
{
	size_t i_len, o_len, destlen;
	size_t retval;
	char *outbuf;
	smb_iconv_t descriptor;

	*dest = NULL;

	if (src == NULL || srclen == (size_t)-1)
		return (size_t)-1;

	descriptor = conv_handles_get(from, to);

	if (descriptor == (smb_iconv_t)-1 || descriptor == (smb_iconv_t)0) {
		/* conversion not supported, return -1*/
		DEBUG(3, ("convert_string_allocate: conversion not supported!\n"));
		return -1;
	}

	destlen = MAX(srclen, 512);
	outbuf = NULL;
convert:
	destlen = destlen * 2;
	if (!g_xrenew(char, &outbuf, destlen)) {
		DEBUG(0, ("convert_string_allocate: realloc failed!\n"));
		SAFE_FREE(outbuf);
		return (size_t)-1;
	}
	i_len = srclen;
	o_len = destlen;
	{
		const char *tmp_inbuf = src;
		char *tmp_outbuf = outbuf;

		retval = smb_iconv(descriptor,
				   &tmp_inbuf, &i_len,
				   &tmp_outbuf, &o_len);
	}
	if(retval == (size_t)-1) {
		if (errno == E2BIG)
			goto convert;
		convert_report_error(from, to, src, srclen, 0);
		return (size_t)-1;
	}
	
	destlen = destlen - o_len;
	*dest = g_renew(char, outbuf, destlen);
	if (destlen && !*dest) {
		DEBUG(0, ("convert_string_allocate: out of memory!\n"));
		SAFE_FREE(outbuf);
		return (size_t)-1;
	}

	return destlen;
}


size_t unix_strupper(const char *src, size_t srclen, char *dest, size_t destlen)
{
	size_t size;
	smb_ucs2_t *buffer;
	
	size = convert_string_allocate(CH_UNIX, CH_UINT16, src, srclen,
				       (void **) &buffer);
	if (size == (size_t) -1) {
		smb_panic("failed to create UCS2 buffer");
	}
	if (!strupper_w(buffer) && (dest == src)) {
		free(buffer);
		return srclen;
	}
	
	size = convert_string(CH_UINT16, CH_UNIX, buffer, size, dest, destlen);
	free(buffer);
	return size;
}

size_t unix_strlower(const char *src, size_t srclen, char *dest, size_t destlen)
{
	size_t size;
	smb_ucs2_t *buffer;
	
	size = convert_string_allocate(CH_UNIX, CH_UINT16, src, srclen,
				       (void **) &buffer);
	if (size == (size_t) -1) {
		smb_panic("failed to create UCS2 buffer");
	}
	if (!strlower_w(buffer) && (dest == src)) {
		free(buffer);
		return srclen;
	}
	size = convert_string(CH_UINT16, CH_UNIX, buffer, size, dest, destlen);
	free(buffer);
	return size;
}


size_t ucs2_align(const void *base_ptr, const void *p, int flags)
{
	if (! (flags & (STR_ALIGN|STR_NOALIGN)))
	{
		DEBUG(1, ("WARNING: ucs2_align: alignment not specified\n"));
		generate_backtrace();
	}
	if (flags & (STR_NOALIGN|STR_ASCII))
		return 0;
	return PTR_DIFF(p, base_ptr) & 1;
}


/**
 * Copy a string from a char* unix src to a dos codepage string destination.
 *
 * @return the number of bytes occupied by the string in the destination.
 *
 * @param flags can include
 * <dl>
 * <dt>STR_TERMINATE</dt> <dd>means include the null termination</dd>
 * <dt>STR_UPPER</dt> <dd>means uppercase in the destination</dd>
 * </dl>
 *
 * @param dest_len the maximum length in bytes allowed in the
 * destination.  If @p dest_len is -1 then no maximum is used.
 **/
size_t push_ascii(void *dest, const char *src, size_t dest_len, int flags)
{
	size_t src_len = strlen(src);
	pstring tmpbuf;

	/* treat a pstring as "unlimited" length */
	if (dest_len == (size_t)-1)
		dest_len = sizeof(pstring);

	if (flags & STR_UPPER) {
		pstrcpy(tmpbuf, src);
		strupper(tmpbuf);
		src = tmpbuf;
	}

	if (flags & STR_TERMINATE)
		src_len++;

	return convert_string(CH_UNIX, CH_DOS, src, src_len, dest, dest_len);
}

int push_ascii_fstring(void *dest, const char *src)
{
	return push_ascii(dest, src, sizeof(fstring), STR_TERMINATE);
}

int push_ascii_pstring(void *dest, const char *src)
{
	return push_ascii(dest, src, sizeof(pstring), STR_TERMINATE);
}

int push_pstring(void *dest, const char *src)
{
	return push_ascii(dest, src, sizeof(pstring), STR_TERMINATE);
}


/**
 * Copy a string from a dos codepage source to a unix char* destination.
 *
 * The resulting string in "dest" is always null terminated.
 *
 * @param flags can have:
 * <dl>
 * <dt>STR_TERMINATE</dt>
 * <dd>STR_TERMINATE means the string in @p src
 * is null terminated, and src_len is ignored.</dd>
 * </dl>
 *
 * @param src_len is the length of the source area in bytes.
 * @returns the number of bytes occupied by the string in @p src.
 **/
size_t pull_ascii(char *dest, const void *src, size_t dest_len, size_t src_len, int flags)
{
	size_t ret;

	if (dest_len == (size_t) -1) {
		dest_len = sizeof(pstring);
	}

	if (flags & STR_TERMINATE) {
		src_len = strlen(src) + 1;
	}

	ret = convert_string(CH_DOS, CH_UNIX, src, src_len, dest, dest_len);

	if (dest_len)
		dest[MIN(ret, dest_len-1)] = 0;

	return src_len;
}

size_t pull_ascii_pstring(char *dest, const void *src)
{
	return pull_ascii(dest, src, sizeof(pstring), -1, STR_TERMINATE);
}

size_t pull_ascii_fstring(char *dest, const void *src)
{
	return pull_ascii(dest, src, sizeof(fstring), -1, STR_TERMINATE);
}

/**
 * Copy a string from a char* src to a unicode destination.
 *
 * @returns the number of bytes occupied by the string in the destination.
 *
 * @param flags can have:
 *
 * <dl>
 * <dt>STR_TERMINATE <dd>means include the null termination.
 * <dt>STR_UPPER     <dd>means uppercase in the destination.
 * <dt>STR_NOALIGN   <dd>means don't do alignment.
 * </dl>
 *
 * @param dest_len is the maximum length allowed in the
 * destination. If dest_len is -1 then no maxiumum is used.
 **/
size_t push_ucs2(const void *base_ptr, void *dest, const char *src, size_t dest_len, int flags)
{
	size_t len=0;
	size_t src_len = strlen(src);
	pstring tmpbuf;

	/* treat a pstring as "unlimited" length */
	if (dest_len == (size_t) -1) {
		dest_len = sizeof(pstring);
	}

	if (flags & STR_UPPER) {
		pstrcpy(tmpbuf, src);
		strupper(tmpbuf);
		src = tmpbuf;
	}

	if (flags & STR_TERMINATE) {
		src_len++;
	}

	if (ucs2_align(base_ptr, dest, flags)) {
		*(char *)dest = 0;
		dest = (void *)((char *)dest + 1);
		if (dest_len) dest_len--;
		len++;
	}

	/* ucs2 is always a multiple of 2 bytes */
	dest_len &= ~1;

	len += convert_string(CH_UNIX, CH_UCS2, src, src_len, dest, dest_len);
	return len;
}


/**
 Copy a string from a char* src to a UTF-8 destination.
 Return the number of bytes occupied by the string in the destination
 Flags can have:
  STR_TERMINATE means include the null termination
  STR_UPPER     means uppercase in the destination
dest_len is the maximum length allowed in the destination. If dest_len
 is -1 then no maxiumum is used.
**/

size_t push_utf8(void *dest, const char *src, size_t dest_len, int flags)
{
	size_t src_len = strlen(src);
	pstring tmpbuf;

	/* treat a pstring as "unlimited" length */
	if (dest_len == (size_t) -1) {
		dest_len = sizeof(pstring);
	}

	if (flags & STR_UPPER) {
		pstrcpy(tmpbuf, src);
		strupper(tmpbuf);
		src = tmpbuf;
	}

	if (flags & STR_TERMINATE) {
		src_len++;
	}

	return convert_string(CH_UNIX, CH_UTF8, src, src_len, dest, dest_len);
}

size_t push_utf8_fstring(void *dest, const char *src)
{
	return push_utf8(dest, src, sizeof(fstring), STR_TERMINATE);
}

size_t push_utf8_pstring(void *dest, const char *src)
{
	return push_utf8(dest, src, sizeof(pstring), STR_TERMINATE);
}


/**
 * Copy a string from a unix char* src to a UTF-8 destination, allocating a buffer
 *
 * @param dest always set at least to NULL 
 *
 * @returns The number of bytes occupied by the string in the destination
 **/

size_t push_utf8_allocate(char **dest, const char *src)
{
	size_t src_len = strlen(src)+1;

	*dest = NULL;
	return convert_string_allocate(CH_UNIX, CH_UTF8, src, src_len, (void **)dest);	
}

/**
 Copy a string from a ucs2 source to a unix char* destination.
 Flags can have:
  STR_TERMINATE means the string in src is null terminated.
  STR_NOALIGN   means don't try to align.
 if STR_TERMINATE is set then src_len is ignored if it is -1.
src_len is the length of the source area in bytes
 Return the number of bytes occupied by the string in src.
 The resulting string in "dest" is always null terminated.
**/

size_t pull_ucs2(const void *base_ptr, char *dest, const void *src, size_t dest_len, size_t src_len, int flags)
{
	size_t ret, moved;

	if (dest_len == (size_t)-1) {
		dest_len = sizeof(pstring);
	}

	moved = ucs2_align(base_ptr, src, flags);
	if (moved) {
		src = (const void *)((const char *)src + moved);
		if (src_len > 0) src_len--;
	}

	if (flags & STR_TERMINATE) {
		src_len = strlen_w(src)*2 + 2;
	}

	/* ucs2 is always a multiple of 2 bytes */
	src_len &= ~1;
	
	ret = convert_string(CH_UCS2, CH_UNIX, src, src_len, dest, dest_len);
	if (dest_len) dest[MIN(ret, dest_len-1)] = 0;

	return src_len + moved;
}

size_t pull_ucs2_pstring(char *dest, const void *src)
{
	return pull_ucs2(NULL, dest, src, sizeof(pstring), -1, STR_TERMINATE | STR_NOALIGN);
}

size_t pull_ucs2_fstring(char *dest, const void *src)
{
	return pull_ucs2(NULL, dest, src, sizeof(fstring), -1, STR_TERMINATE | STR_NOALIGN);
}

/****************************************************************************
copy a string from a utf-8 source to a unix char* destination
flags can have:
  STR_TERMINATE means the string in src is null terminated
if STR_TERMINATE is set then src_len is ignored
src_len is the length of the source area in bytes
return the number of bytes occupied by the string in src
the resulting string in "dest" is always null terminated
****************************************************************************/
int pull_utf8(char *dest, const void *src, int dest_len, int src_len, int flags)
{
	int ret;

	if (dest_len == -1) {
		dest_len = sizeof(pstring);
	}

	if (flags & STR_TERMINATE) {
		src_len = strlen(src) + 1;
	}

	ret = convert_string(CH_UTF8, CH_UNIX, src, src_len, dest, dest_len);
	if (dest_len) dest[MIN(ret, dest_len-1)] = 0;

	return src_len;
}

int pull_utf8_pstring(char *dest, const void *src)
{
	return pull_utf8(dest, src, sizeof(pstring), -1, STR_TERMINATE);
}

int pull_utf8_fstring(char *dest, const void *src)
{
	return pull_utf8(dest, src, sizeof(fstring), -1, STR_TERMINATE);
}


/**
 * Copy a string from a UTF-8 src to a unix char * destination, allocating a buffer
 *
 * @param dest always set at least to NULL 
 *
 * @returns The number of bytes occupied by the string in the destination
 **/

size_t pull_utf8_allocate(void **dest, const char *src)
{
	size_t src_len = strlen(src)+1;
	*dest = NULL;
	return convert_string_allocate(CH_UTF8, CH_UNIX, src, src_len, dest);	
}
 
/**
 Copy a string from a char* src to a unicode or ascii
dos codepage destination choosing unicode or ascii based on the 
 flags in the SMB buffer starting at base_ptr.
 Return the number of bytes occupied by the string in the destination.
flags can have:
  STR_TERMINATE means include the null termination.
  STR_UPPER     means uppercase in the destination.
  STR_ASCII     use ascii even with unicode packet.
  STR_NOALIGN   means don't do alignment.
dest_len is the maximum length allowed in the destination. If dest_len
 is -1 then no maxiumum is used.
**/

size_t push_string(const void *base_ptr, void *dest, const char *src, size_t dest_len, int flags)
{
	size_t ret;
	if (!(flags & STR_ASCII) && \
	    ((flags & STR_UNICODE || \
	      (SVAL(base_ptr, smb_flg2) & FLAGS2_UNICODE_STRINGS))))
	{
		ret = push_ucs2(base_ptr, dest, src, dest_len, flags);
	}
	else
	{
		ret = push_ascii(dest, src, dest_len, flags);
	}

	DEBUG(5, ("push_string([%s], %d, 0x%x) returning %u\n",
		  src, dest_len, flags, ret));
	dump_data(9, dest, ret);

	return ret;
}


/**
 Copy a string from a unicode or ascii source (depending on
 the packet flags) to a char* destination.
 Flags can have:
  STR_TERMINATE means the string in src is null terminated.
  STR_UNICODE   means to force as unicode.
  STR_ASCII     use ascii even with unicode packet.
  STR_NOALIGN   means don't do alignment.
 if STR_TERMINATE is set then src_len is ignored is it is -1
 src_len is the length of the source area in bytes.
 Return the number of bytes occupied by the string in src.
 The resulting string in "dest" is always null terminated.
**/

int pull_string(const void *base_ptr, char *dest, const void *src, int dest_len, int src_len, 
		int flags)
{
	if (!(flags & STR_ASCII) && \
	    ((flags & STR_UNICODE || \
	      (SVAL(base_ptr, smb_flg2) & FLAGS2_UNICODE_STRINGS)))) {
		return pull_ucs2(base_ptr, dest, src, dest_len, src_len, flags);
	}
	return pull_ascii(dest, src, dest_len, src_len, flags);
}

size_t align_string(const void *base_ptr, const char *p, int flags)
{
	if (!(flags & STR_ASCII) && \
	    ((flags & STR_UNICODE || \
	      (SVAL(base_ptr, smb_flg2) & FLAGS2_UNICODE_STRINGS)))) {
		return ucs2_align(base_ptr, p, flags);
	}
	return 0;
}



/**
 convert from ucs2 to unix charset and return the
 allocated and converted string or NULL if an error occurred.
 you must provide a zero terminated string.
 the returning string will be zero terminated.
**/

char *acnv_u2ux(const smb_ucs2_t *src)
{
	size_t slen;
	size_t dlen;
	void *dest;
	
	slen = (strlen_w(src) + 1) * sizeof(smb_ucs2_t);
	dlen = convert_string_allocate(CH_UCS2, CH_UNIX, src, slen, &dest);
	if (dlen == (size_t)-1)
		return NULL;
	return dest;
}

/**
 Convert from unix to ucs2 charset and return the
 allocated and converted string or NULL if an error occurred.
 You must provide a zero terminated string.
 The returning string will be zero terminated.
**/

smb_ucs2_t *acnv_uxu2(const char *src)
{
	size_t slen;
	size_t dlen;
	void *dest;
	
	slen = strlen(src) + 1;
	dlen = convert_string_allocate(CH_UNIX, CH_UCS2, src, slen, &dest);
	if (dlen == (size_t)-1)
		return NULL;
	return dest;
}

/****************************************************************************
convert from ucs2 to dos charset and return the
allocated and converted string or NULL if an error occurred.
you must provide a zero terminated string.
the returning string will be zero terminated.
****************************************************************************/
char *acnv_u2dos(const smb_ucs2_t *src)
{
	size_t slen;
	size_t dlen;
	void *dest;
	
	slen = (strlen_w(src) + 1) * sizeof(smb_ucs2_t);
	dlen = convert_string_allocate(CH_UCS2, CH_DOS, src, slen, &dest);
	if (dlen == (size_t)-1)
		return NULL;
	return dest;
}

/**
 Convert from dos to ucs2 charset and return the
 allocated and converted string or NULL if an error occurred.
 You must provide a zero terminated string.
 The returning string will be zero terminated.
**/

smb_ucs2_t *acnv_dosu2(const char *src)
{
	size_t slen;
	size_t dlen;
	void *dest;
	
	slen = strlen(src) + 1;
	dlen = convert_string_allocate(CH_DOS, CH_UCS2, src, slen, &dest);
	if (dlen == (size_t)-1)
		return NULL;
	else
		return dest;
}
