/* 
   Unix SMB/Netbios implementation.
   Samba TNG daemon utility functions
   Copyright (C) Andrew Tridgell 1992-2000
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/
/* MKPROTO: libsamba_proto.h		(do not remove this line) */

#include "includes.h"
#include <errno.h>
#include <signal.h>
#include "libsamba.h"
#include "tng_misc.h"


/*******************************************************************
close the low 3 fd's and open dev/null in their place
********************************************************************/
void close_low_fds(void)
{
	int fd;
	int i;
	close(0);
	close(1);
#ifndef __INSURE__
	close(2);
#endif
	/* try and use up these file descriptors, so silly
	   library routines writing to stdout etc won't cause havoc */
	for (i = 0; i < 3; i++)
	{
		fd = sys_open("/dev/null", O_RDWR, 0);
		if (fd < 0)
			fd = sys_open("/dev/null", O_WRONLY, 0);
		if (fd < 0)
		{
			DEBUG(0, ("Can't open /dev/null\n"));
			return;
		}
		if (fd != i)
		{
			DEBUG(0, ("Didn't get file descriptor %d\n", i));
			return;
		}
	}
}


/****************************************************************************
 Become a daemon, discarding the controlling terminal.
****************************************************************************/

void become_daemon(void)
{
	if (sys_fork())
	{
		_exit(0);
	}

	/* detach from the terminal */
#ifdef HAVE_SETSID
	setsid();
#elif defined(TIOCNOTTY)
	{
		int i = sys_open("/dev/tty", O_RDWR, 0);
		if (i != -1)
		{
			ioctl(i, (int)TIOCNOTTY, (char *)0);
			close(i);
		}
	}
#endif /* HAVE_SETSID */

	/* Close fd's 0,1,2. Needed if started by rsh */
	close_low_fds();
}


void daemon_signal_setup(void)
{
	/* POSIX demands that signals are inherited. If the invoking process has
	 * these signals masked, we will have problems, as we won't recieve them.
	 */
#ifdef SIGHUP
	BlockSignals(False, SIGHUP);
#endif
#ifdef SIGUSR1
	BlockSignals(False, SIGUSR1);
#endif

#if defined(SIGUSR2)
	/* We are no longer interested in USR2 */
	BlockSignals(True,SIGUSR2);
#endif

#ifdef SIGPIPE
	/* we are never interested in SIGPIPE */
	BlockSignals(True,SIGPIPE);
#endif
}




static struct timeval start_time;


/* child_start is for migrational purposes */
void child_start(void)
{
	GetTimeOfDay(&start_time);

	DEBUG(3, ("New child at PID %d\n", getpid()));
}


void child_exit(int exit_val)
{
	struct timeval end_time;
	unsigned int secs;
	int msec;

	GetTimeOfDay(&end_time);

	msec = (end_time.tv_usec - start_time.tv_usec) / 1000;
	secs = end_time.tv_sec - start_time.tv_sec;
	if (msec < 0)
	{
		msec += 1000;
		secs--;
	}

	DEBUG((exit_val ? 0 : 3),
	      ("Child %d exiting with code %d after %u.%03d seconds.\n",
	       getpid(), exit_val, secs, msec));

	exit(exit_val);
}


int child_create(int (* start_fn) (void *data), void *data,
		 pid_t *child_pid)
{
	pid_t pid;
	int retval;

	g_return_val_if_fail(start_fn, -1);

	pid = sys_fork();

	if (pid == -1)
	{
		int err = errno;
		DEBUG(0, ("fork failed: %s\n", strerror(err)));
		return err;
	}

	if (pid > 0)
	{
		DEBUG(4, ("fork succeeded: New child's PID: %d\n",
			  pid));
		if (child_pid)
			*child_pid = pid;
		return 0;
	}

	/* Now in child */
	child_start();

	retval = start_fn(data);

	child_exit(retval);

	/* not reached */
	return -1;
}
