/* 
   Generic Handle system
   Copyright (C) Elrond <elrond@samba-tng.org>  2003
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
/* MKPROTO: NONE			(do not remove this line) */

#include "includes.h"
#include "libsamba.h"
#include "tng_misc.h"
#include "dlinklist.h"
#include "handle.h"

typedef struct _HandleData HandleData;

struct _HandleData
{
	HandleData *next, *prev;
	char *name;
	void *data;
	HandleDataFreeFn free_fn;
};


struct _Handle
{
	Handle *parent;
	Handle *childs;
	/* sibling list */
	Handle *next, *prev;
	HandleData *data;
	BOOL closed;
};


static void handle_try_close(Handle *hnd);


Handle *handle_new(Handle *parent)
{
	Handle *hnd;

	hnd = g_new(Handle, 1);
	if (!hnd)
		return NULL;
	hnd->parent = NULL;
	hnd->childs = NULL;
	hnd->next = hnd->prev = NULL;
	hnd->data = NULL;
	hnd->closed = False;

	handle_set_parent(hnd, parent);

	return hnd;
}

void handle_set_parent(Handle *hnd, Handle *parent)
{
	if (!hnd)
		return;
	if (parent == hnd->parent)
		return;
	if (hnd->parent)
	{
		DLIST_REMOVE(hnd->parent->childs, hnd);
		/* See, if the parent is now unreferenced... */
		handle_try_close(hnd->parent);
	}
	hnd->parent = parent;
	if (parent)
	{
		DLIST_ADD(parent->childs, hnd);
	}
}

void handle_set_data(Handle *hnd, const char *name,
		     void *data, HandleDataFreeFn free_fn)
{
	HandleData *hnd_data;

	if (!hnd || !name || !data)
		return;
	hnd_data = g_new(HandleData, 1);
	if (!hnd_data)
		return;
	hnd_data->name = strdup(name);
	hnd_data->data = data;
	hnd_data->free_fn = free_fn;
	DLIST_ADD(hnd->data, hnd_data);
}

void *handle_get_data(const Handle *hnd, const char *name, int check_parents)
{
	if (!name || !hnd || hnd->closed)
		return NULL;
	while(hnd)
	{
		HandleData *d;
		for (d = hnd->data; d; d = d->next)
		{
			if (strcsequal(d->name, name))
				return d->data;
		}
		if (check_parents)
			hnd = hnd->parent;
		else
			hnd = NULL;
	}
	return NULL;
}

static void handle_try_close(Handle *hnd)
{
	HandleData *d;
	if (!hnd->closed || hnd->childs)
		return;

	while ((d = hnd->data))
	{
		DLIST_REMOVE(hnd->data, d);
		if (d->free_fn)
			d->free_fn(d->data);
		safe_free(d->name);
		free(d);
	}
	if (hnd->parent)
	{
		DLIST_REMOVE(hnd->parent->childs, hnd);
		handle_try_close(hnd->parent);
	}
	free(hnd);
}

void handle_close(Handle *hnd)
{
	if (!hnd)
		return;
	hnd->closed = True;
	handle_try_close(hnd);
}

void handle_set_name(Handle *hnd, const char *name)
{
	handle_set_data(hnd, "__handle__/name(str)", strdup(name), free);
}

const char *handle_get_name(const Handle *hnd)
{
	const char *n;

	if (!hnd)
		return "(NULL)";

	if (hnd->closed)
		return "(closed)";

	n = handle_get_data(hnd, "__handle__/name(str)", False);

	if (!n)
		return "(unnamed)";

	return n;
}
