/* 
   Unix SMB/Netbios implementation.
   Version 1.9.
   SMB msrpcent generic functions
   Copyright (C) Luke Kenneth Casson Leighton 1996-2001
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/
/* MKPROTO: rpc_client_proto.h		(do not remove this line) */

#include "includes.h"
#include "libsamba.h"
#include "tng_misc.h"
#include "byteorder.h"
#include "client.h"
#include "rpc_parse.h"
#include "rpc_client.h"
#include "pathnames.h"
#include "others.h"


/****************************************************************************
  read an msrpc pdu from a fd. 
  The timeout is in milliseconds. 
****************************************************************************/
BOOL receive_msrpc(int fd, prs_struct * data, unsigned int timeout)
{
	BOOL ok;
	size_t len;
	RPC_HDR hdr;

	prs_init(data, UNMARSHALL);
	prs_realloc_data(data, 16);

	ok = (read_socket_with_timeout(fd, data->data, 16, 16, timeout) == 16);

	if (!ok)
	{
		prs_free_data(data);
		return False;
	}

	if (!smb_io_rpc_hdr("hdr", &hdr, data, 0))
	{
		prs_free_data(data);
		return False;
	}

	len = hdr.frag_len - 16;
	if (len > 0)
	{
		size_t ret;
		prs_realloc_data(data, hdr.frag_len);
		ret = read_socket_data(fd, prs_data_rw(data, 16, len), len);
		if (ret != len)
		{
			prs_free_data(data);
			return False;
		}
		prs_set_start(data, 0);
		data->offset = hdr.frag_len;
		data->end = hdr.frag_len;
		return True;
	}

	prs_free_data(data);
	return False;
}

/****************************************************************************
  send an smb to a fd and re-establish if necessary
****************************************************************************/
BOOL msrpc_send(int fd, prs_struct * ps)
{
	size_t len;
	ssize_t ret;
	char *outbuf;

	/* if there is no data then say success */
	if (!ps || !ps->data) return False;

	len = prs_buf_len(ps);
	outbuf = ps->data;

	DEBUG(10, ("ncalrpc_l_send_prs: data: %p len %d\n", outbuf, len));
	dump_data(10, outbuf, len);

	if (len == 0)
	{
		return False;
	}

	ret = write_socket(fd, outbuf, len);
	prs_free_data(ps);

	return (ret == len);
}

/****************************************************************************
  receive msrpc packet
****************************************************************************/
BOOL msrpc_receive(int fd, prs_struct * ps)
{
	int len;

	DEBUG(10, ("ncalrpc_l_receive: %d\n", __LINE__));

	if (!receive_msrpc(fd, ps, 0))
	{
		DEBUG(1,("msrpc_receive: failed\n"));
		return False;
	}

	len = prs_buf_len(ps);

	if (ps->data == NULL || len <= 0)
	{
		DEBUG(10, ("ncalrpc_l_receive: no data\n"));
		return False;
	}

	dump_data(10, ps->data, len);

	DEBUG(10, ("ncalrpc_l_receive: len %d\n", len));

	prs_debug_out(ps, "ncalrpc_l_receive_prs", 200);

	return True;
}

/****************************************************************************
open the msrpcent sockets
****************************************************************************/
static BOOL ncalrpc_l_connect(struct msrpc_local *msrpc,
		const char *srv_name,
		const char *pipe_name)
{
	fstring path;
	fstring pname;
	fstrcpy(pname, pipe_name);
	strlower(pname);
	slprintf(path, sizeof(path) - 1, "%s/%s", pathname_msrpc_sock_dir, pname);

	fstrcpy(msrpc->srv_name, srv_name);
	fstrcpy(msrpc->pipe_name, pipe_name);

	msrpc->fd = open_pipe_sock(path);

	if (msrpc->fd == -1)
	{
		return False;
	}

	return True;
}


/****************************************************************************
close the socket descriptor
****************************************************************************/
static void ncalrpc_l_close_socket(struct msrpc_local *msrpc)
{
	if (msrpc->fd != -1)
	{
		close(msrpc->fd);
	}
	msrpc->fd = -1;
}



static BOOL ncalrpc_l_authenticate(struct msrpc_local *msrpc)
{
	int sock = msrpc->fd;
	uint32 len;
	char *data;
	prs_struct ps;
#ifndef NOONS_SMALL_BUILD
	uint32 status;
#endif
	vuser_struct *vuser = NULL;
	uint16 command = AGENT_CMD_CON;
	RPC_HDR hdr;

#ifdef NO_NPEMU_SECURITY
	return True;
#endif
	vuser = get_valid_vuser_struct(msrpc->nt.vuid);

	if (! vuser)
	{
		DEBUG(1, ("ncalrpc_l_authenticate: No vuser for vuid %x\n",
			  msrpc->nt.vuid));
	}

	/* create user credentials, but place data at an offset
	 * of 16 bytes into the buffer, enough for a PDU header - later
	 */
	if (!create_user_creds(&ps, msrpc->srv_name, msrpc->pipe_name,
				lp_namedpipes_version(), command, vuser))
	{
		DEBUG(0, ("could not parse credentials\n"));
		close(sock);
		vuid_free_vuser_struct(vuser);
		return False;
	}

	vuid_free_vuser_struct(vuser);

	/* now add the PDU header, at the beginning (ps.offset=0)
	 */
	len = ps.offset;

	make_rpc_hdr(&hdr, RPC_INHERITCTX,
		RPC_FLG_FIRST | RPC_FLG_LAST, 0, len, 0);

	ps.offset = 0;
	smb_io_rpc_hdr("hdr", &hdr, &ps, 0);
	ps.offset = len;

	data = prs_data_rw(&ps, 0, len);

#ifdef DEBUG_PASSWORD
	DEBUG(100, ("data len: %d\n", len));
	dump_data(100, data, len);
#endif

	if (write_socket(sock, data, len) <= 0)
	{
		DEBUG(0, ("write failed\n"));
		prs_free_data(&ps);
		return False;
	}
	prs_free_data(&ps);

	return True;
#if 0
	if (lp_namedpipes_version() == 0x1)
	{
		/* version 1 of the namedpipes interface we don't
		 * receive a status back.  if one is needed, it's
		 * going to have to be sent back as a PDU and i
		 * can't be bothered to do that right now.  we'll see...
		 */
		return True;
	}

	len = read_socket_data(sock, (char*)&status, sizeof(status));

	return len == sizeof(status) && status == 0x0;
#endif
}

/****************************************************************************
shutdown a msrpcent structure
****************************************************************************/
void ncalrpc_l_shutdown(struct msrpc_local *msrpc)
{
	DEBUG(10, ("msrpc_shutdown\n"));
	if (!msrpc)
		return;
	ncalrpc_l_close_socket(msrpc);
	ZERO_OBJPN(msrpc);
}

/****************************************************************************
initialise a msrpcent structure
****************************************************************************/
struct msrpc_local *ncalrpc_l_initialise(uint16 vuid)
{
	struct msrpc_local *msrpc;

	msrpc = g_new(struct msrpc_local, 1);
	if (!msrpc)
		return NULL;

	ZERO_STRUCTP(msrpc);

	msrpc->fd = -1;
	msrpc->initialised = 1;

	if (vuid != UID_FIELD_INVALID)
	{
		msrpc->nt.vuid = vuid;
		vuid_get_usr_sesskey(vuid, msrpc->nt.usr_sess_key);
	}
	else
	{
		NET_USER_INFO_3 usr;
		uid_t uid = getuid();
		gid_t gid = getgid();
		char *name = uidtoname(uid);

		ZERO_STRUCT(usr);

		msrpc->nt.vuid = register_vuid(uid, gid,
					       name, name, 
					       False, 0,
					       &usr);
	}

	return msrpc;
}


/****************************************************************************
establishes a connection right up to doing tconX, reading in a password.
****************************************************************************/
BOOL ncalrpc_l_establish_connection(struct msrpc_local *msrpc,
				    const char *srv_name,
				    const char *pipe_name)
{
	if (strnequal("\\PIPE\\", pipe_name, 6))
	{
		pipe_name = &pipe_name[6];
	}

	if (strnequal("\\\\", srv_name, 2))
	{
		srv_name = &srv_name[2];
	}

	DEBUG(5, ("ncalrpc_l_establish_connection: connecting to %s[%s]\n",
		  srv_name, pipe_name));

	/* establish connection */

	if (!msrpc->initialised)
	{
		return False;
	}

	if (msrpc->fd == -1)
	{
		if (!ncalrpc_l_connect(msrpc, srv_name, pipe_name))
		{
			DEBUG(1,
			      ("ncalrpc_l_establish_connection: failed %s)\n",
			       pipe_name));

			return False;
		}
	}

	if (!ncalrpc_l_authenticate(msrpc))
	{
		DEBUG(0, ("authenticate failed\n"));
		close(msrpc->fd);
		msrpc->fd = -1;
		return False;
	}

	return True;
}


/****************************************************************************
 Server side functions...
****************************************************************************/


/*
 * Read credentials from fd (c), allocate vuid for them
 */

BOOL ncalrpc_srv_get_user_creds(int c, uint16 *vuid)
{
	pstring buf;
	int rl;
	uint32 len;
	BOOL new_con = False;
#ifndef NOONS_SMALL_BUILD
	uint32 status;
#endif

	CREDS_CMD cmd;
	RPC_HDR hdr;
	prs_struct ps;

	ZERO_STRUCT(cmd);

#ifdef NO_NPEMU_SECURITY
	{
		NET_USER_INFO_3 usr;
		uid_t uid = getuid();
		gid_t gid = getgid();
		char *name = uidtoname(uid);

		ZERO_STRUCT(usr);

		DEBUG(0, ("named pipe emulation security disabled\n"));
		(*vuid) = register_vuid(uid, gid,
					name, name,
					False, 0,
					&usr);
		return True;
	}
#endif

	/* first grab a PDU of 16 bytes.  NOTE - we don't do any
	 * detection here!  it's gonna be of type RPC_INHERITCTX
	 * OR ELSE!
	 */

	DEBUG(10, ("get_user_creds: first request (RPC_HDR)\n"));

	rl = read_socket_data(c, (char *)&buf, sizeof(RPC_HDR));

	if (rl != sizeof(RPC_HDR))
	{
		DEBUG(0, ("Unable to read length\n"));
		dump_data(0, buf, sizeof(RPC_HDR));
		return False;
	}

	prs_init(&ps, UNMARSHALL);
	if (!prs_append_data(&ps, buf, rl))
	{
		DEBUG(0, ("get_user_creds: prs_append_data failed\n"));
		prs_free_data(&ps);
		return False;
	}

	if (!smb_io_rpc_hdr("hdr", &hdr, &ps, 0))
	{
		DEBUG(0, ("Unable to parse credentials\n"));
		prs_free_data(&ps);
		return False;
	}

	prs_free_data(&ps);

	/* now we know how much data there is - minus the PDU header */

	len = hdr.frag_len - 16;

	if (len > sizeof(buf))
	{
		DEBUG(0, ("length %d too long\n", len));
		return False;
	}

	rl = read(c, buf, len);

	if (rl < 0)
	{
		DEBUG(0, ("Unable to read from connection\n"));
		return False;
	}

	if (rl < len)
	{
		DEBUG(3, ("get_user_creds: Got only %d bytes, expected %d\n",
			  rl, len));
	}

#ifdef DEBUG_PASSWORD
	dump_data(100, buf, rl);
#endif

	/* make a static data parsing structure from the api_fd_reply data */
	prs_init(&ps, UNMARSHALL);
	if (!prs_append_data(&ps, buf, rl))
	{
		DEBUG(0, ("get_user_creds: prs_append_data failed\n"));
		prs_free_data(&ps);
		return False;
	}

	if (!creds_io_cmd("creds", &cmd, &ps, 0))
	{
		DEBUG(0, ("Unable to parse credentials\n"));
		prs_free_data(&ps);
		return False;
	}

	if (ps.offset != rl)
	{
		DEBUG(0, ("Buffer size %d %d!\n", ps.offset, rl));
		prs_free_data(&ps);
		return False;
	}

	prs_free_data(&ps);

	switch (cmd.command)
	{
		case AGENT_CMD_CON:
		case AGENT_CMD_CON_ANON:
		{
			new_con = True;
			break;
		}
		case AGENT_CMD_CON_REUSE:
		{
			new_con = True;
			break;
		}
		default:
		{
			DEBUG(0, ("unknown command %d\n", cmd.command));
			return False;
		}
	}

	/* set the client hostname from the connection */
	if (!strequal(cmd.hostname, "."))
	{
		/* treat \\. - loopback - differently */
		set_remote_machine (cmd.hostname);
	}

	switch (cmd.version)
	{
		case 0x0:
		{
			/* obtain the remote process id and vuid */
			if (cmd.vuser == NULL)
			{
				DEBUG(0, ("get_user_creds: cmd.vuser == NULL\n"));
				(*vuid) = UID_FIELD_INVALID;
			}
			else
			{
				if (cmd.vuser->usr == NULL)
				{
					DEBUG(0, ("get_user_creds: TODO - do unix lookup like in sessetup\n"));
				}
				(*vuid) = make_vuid(cmd.vuser);
			}
			break;
		}
	}

	creds_free_contents(&cmd);

#if 0
	if (cmd.version == 0x1)
	{
		/* version 1 of the namedpipes interface we don't
		 * send a status back.  if one is needed, it's
		 * going to have to be sent back as a PDU and i
		 * can't be bothered to do that right now.  we'll see...
		 */
		return new_con;
	}
	status = new_con ? 0x0 : 0x1;

	if (write_socket(c, (char *)&status, sizeof(status)) !=
	    sizeof(status))
	{
		return False;
	}
#endif
	return new_con;
}
