/* 
   Unix SMB/Netbios implementation.
   Samba utility functions
   Copyright (C) Andrew Tridgell 1992-2000
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/
/* MKPROTO: libsamba_proto.h		(do not remove this line) */

#include "includes.h"
#include <errno.h>
#include <signal.h>
#include <stdarg.h>
#if defined(HAVE_VALGRIND_MEMCHECK_H) && !defined(__STRICT_ANSI__)
#include <valgrind/memcheck.h>
#else
#define VALGRIND_CHECK_READABLE(a,b) (NULL)
#endif
#include "tng_misc.h"
#include "libsamba.h"
#include "byteorder.h"
#include "client.h"
#include "rpc_reg.h"

#if (defined(HAVE_NETGROUP) && defined (WITH_AUTOMOUNT))
#ifdef WITH_NISPLUS_HOME
#ifdef BROKEN_NISPLUS_INCLUDE_FILES
/*
 * The following lines are needed due to buggy include files
 * in Solaris 2.6 which define GROUP in both /usr/include/sys/acl.h and
 * also in /usr/include/rpcsvc/nis.h. The definitions conflict. JRA.
 * Also GROUP_OBJ is defined as 0x4 in /usr/include/sys/acl.h and as
 * an enum in /usr/include/rpcsvc/nis.h.
 */

#if defined(GROUP)
#undef GROUP
#endif

#if defined(GROUP_OBJ)
#undef GROUP_OBJ
#endif

#endif /* BROKEN_NISPLUS_INCLUDE_FILES */

#include <rpcsvc/nis.h>

#else /* !WITH_NISPLUS_HOME */

#include "rpcsvc/ypclnt.h"

#endif /* WITH_NISPLUS_HOME */
#endif /* HAVE_NETGROUP && WITH_AUTOMOUNT */


int Protocol = PROTOCOL_COREPLUS;

/* a default finfo structure to ensure all fields are sensible */
file_info def_finfo = { -1, 0, 0, 0, 0, 0, 0, "", "" };

/* this is used by the chaining code */
int chain_size = 0;


/*
   case handling on filenames 
*/
int case_default = CASE_LOWER;

/* the following control case operations - they are put here so the
   client can link easily */

pstring user_socket_options = DEFAULT_SOCKET_OPTIONS;

pstring global_myname = "";
fstring global_myworkgroup = "";

char *daynames[] = { "Mon", "Tue", "Wed", "Thu", "Fri", "Sat", "Sun" };
char *daynames_short[] = { "M", "Tu", "W", "Th", "F", "Sa", "Su" };

/*********************************************************************
 Memory allocation - see comment in tng_misc.h for details.
 Do not use the g__ functions directly.
 ********************************************************************/

#define MAX_ALLOC_SIZE (1024*1024*256)
#define CHECK_ALLOC_OVERFLOW(type, count) do { \
	if (count > MAX_ALLOC_SIZE / type) { \
		DEBUG(0, ("attempted memory overflow: %u * %u bytes\n", \
			  type, count));				\
		return 0; \
	} } while (0)

 void *g__new(size_t type, size_t count)
{
	CHECK_ALLOC_OVERFLOW(type, count);
	return malloc(type * count);
}

 void *g__new0(size_t type, size_t count)
{
	CHECK_ALLOC_OVERFLOW(type, count);
	return calloc(count, type);
}

 BOOL g__xrenew(size_t type, void *ptr, size_t count)
{
	void **oldptr = ptr;
	void *newptr;
	CHECK_ALLOC_OVERFLOW(type, count);
	newptr = realloc(*oldptr, type * count);
	if (!newptr)
		return False;
	*oldptr = newptr;
	return True;
}

/*************************************************************
 initialise password databases, domain names, domain sid.
**************************************************************/
BOOL init_myworkgroup(void)
{
	fstrcpy(global_myworkgroup, lp_workgroup());

	if (strequal(global_myworkgroup, "*"))
	{
		DEBUG(0,
		      ("ERROR: a workgroup name of * is no longer supported\n"));
		return False;
	}
	return True;
}

/****************************************************************************
 Find a suitable temporary directory. The result should be copied immediately
 as it may be overwritten by a subsequent call.
  ****************************************************************************/

char *tmpdir(void)
{
	char *p;
	if ((p = getenv("TMPDIR")))
	{
		return p;
	}
	return "/tmp";
}

/****************************************************************************
 Determine whether we are in the specified group.
****************************************************************************/

BOOL in_group(gid_t group, gid_t current_gid, int ngroups, gid_t * groups)
{
	int i;

	if (group == current_gid)
		return (True);

	for (i = 0; i < ngroups; i++)
		if (group == groups[i])
			return (True);

	return (False);
}


/****************************************************************************
gets either a hex number (0xNNN) or decimal integer (NNN).
****************************************************************************/
uint32 get_number(const char *tmp)
{
	uint32 ret;
	if (strnequal(tmp, "0x", 2))
	{
		ret = strtoul(tmp, (char **)NULL, 16);
		DEBUG(10, ("get_number: %s -> 0x%x\n", tmp, ret));
	}
	else
	{
		ret = strtoul(tmp, (char **)NULL, 10);
		DEBUG(10, ("get_number: %s -> %d\n", tmp, ret));
	}
	return ret;
}

/****************************************************************************
 Like atoi but gets the value up to the separator character.
****************************************************************************/

char *Atoic(char *p, int *n, char *c)
{
	if (!isdigit((int)*p))
	{
		DEBUG(5, ("Atoic: malformed number\n"));
		return NULL;
	}

	(*n) = (int)get_number(p);

	if (strnequal(p, "0x", 2))
	{
		p += 2;
	}

	while ((*p) && isdigit((int)*p))
	{
		p++;
	}

	if (strchr(c, *p) == NULL)
	{
		DEBUG(5,
		      ("Atoic: no separator characters (%s) not found\n", c));
		return NULL;
	}

	return p;
}

uint32 *add_num_to_list(uint32 **num, int *count, int val)
{
	if (!g_xrenew(uint32, num, *count + 1))
	{
		return NULL;
	}
	(*num)[(*count)] = val;
	(*count)++;

	return (*num);
}

/*************************************************************************
 Reads a list of numbers.
 *************************************************************************/

char *get_numlist(char *p, uint32 **num, int *count)
{
	int val;

	if (num == NULL || count == NULL)
	{
		return NULL;
	}

	(*count) = 0;
	(*num) = NULL;

	while ((p = Atoic(p, &val, ":,")) != NULL && (*p) != ':')
	{
		if (add_num_to_list(num, count, val) == NULL)
		{
			return NULL;
		}
		p++;
	}

	return p;
}

/*******************************************************************
 Check if a file exists - call vfs_file_exist for samba files.
********************************************************************/

BOOL file_exist(const char *fname, SMB_STRUCT_STAT *sbuf)
{
	SMB_STRUCT_STAT st;
	if (!sbuf)
		sbuf = &st;

	if (sys_stat(fname, sbuf) != 0)
		return (False);

	return (S_ISREG(sbuf->st_mode));
}

/*******************************************************************
 Check a files mod time.
********************************************************************/

time_t file_modtime(const char *fname)
{
	SMB_STRUCT_STAT st;

	if (sys_stat(fname, &st) != 0)
		return (0);

	return (st.st_mtime);
}

/*******************************************************************
 Check if a directory exists.
********************************************************************/

BOOL directory_exist(char *dname, SMB_STRUCT_STAT * st)
{
	SMB_STRUCT_STAT st2;
	BOOL ret;

	if (!st)
		st = &st2;

	if (sys_stat(dname, st) != 0)
		return (False);

	ret = S_ISDIR(st->st_mode);
	if (!ret)
		errno = ENOTDIR;
	return ret;
}

/*******************************************************************
return a string representing an attribute for a file
********************************************************************/
char *attrib_string(uint16 mode)
{
	static struct field_info attrs[] =
	{
		/* descriptions unused at the moment */
		{ aVOLID,  "V", "Volume label" },
		{ aDIR,    "D", "Directory" },
		{ aARCH,   "A", "Archive" },
		{ aHIDDEN, "H", "Hidden" },
		{ aSYSTEM, "S", "System" },
		{ aRONLY,  "R", "Read-only" },
		{ 0, NULL, NULL }
	};

	return bit_field_to_str (mode, attrs, NULL);
}

/*******************************************************************
  set the length and marker of an smb packet
********************************************************************/
void smb_setlen(char *buf, int len)
{
	_smb_setlen(buf, len);

	SCVAL(buf,4,0xFF);
	SCVAL(buf,5,'S');
	SCVAL(buf,6,'M');
	SCVAL(buf,7,'B');
}

/*******************************************************************
  setup the word count and byte count for a smb message
********************************************************************/
int set_message(char *buf, int num_words, int num_bytes, BOOL zero)
{
	if (zero)
		memset(buf + smb_size, '\0', num_words * 2 + num_bytes);
	SCVAL(buf,smb_wct,num_words);
	SSVAL(buf, smb_vwv + num_words * SIZEOFWORD, num_bytes);
	smb_setlen(buf, smb_size + num_words * 2 + num_bytes - 4);
	return (smb_size + num_words * 2 + num_bytes);
}


/*******************************************************************
reduce a file name, removing .. elements.
********************************************************************/
void dos_clean_name(char *s)
{
	char *p = NULL;

	DEBUG(3, ("dos_clean_name [%s]\n", s));

	/* remove any double slashes */
	all_string_sub_g(s, "\\\\", "\\", 0);

	while ((p = strstr(s, "\\..\\")) != NULL)
	{
		pstring s1;

		*p = 0;
		pstrcpy(s1, p + 3);

		if ((p = strrchr(s, '\\')) != NULL)
			*p = 0;
		else
			*s = 0;
		pstrcat(s, s1);
	}

	trim_string(s, NULL, "\\..");

	all_string_sub_g(s, "\\.\\", "\\", 0);

	DEBUGADD(3, ("        after: [%s]\n", s));
}

/****************************************************************************
 Transfer some data between two fd's.
****************************************************************************/
static
SMB_OFF_T transfer_file(int infd, int outfd, SMB_OFF_T n, char *header,
			int headlen, int align)
{
	static char *buf = NULL;
	static int size = 0;
	char *buf1, *abuf;
	SMB_OFF_T total = 0;

	DEBUG(4,
	      ("transfer_file n=%.0f  (head=%d) called\n", (double)n,
	       headlen));

	if (size == 0)
	{
		size = lp_readsize();
		size = MAX(size, 1024);
	}

	while (!buf && size > 0)
	{
		buf = g_new(char, size + 8);
		if (!buf)
			size /= 2;
	}

	if (!buf)
	{
		DEBUG(0, ("Can't allocate transfer buffer!\n"));
		exit(1);
	}

	abuf = buf + (align % 8);

	if (header)
		n += headlen;

	while (n > 0)
	{
		int s = (int)MIN(n, (SMB_OFF_T) size);
		int ret, ret2 = 0;

		ret = 0;

		if (header && (headlen >= MIN(s, 1024)))
		{
			buf1 = header;
			s = headlen;
			ret = headlen;
			headlen = 0;
			header = NULL;
		}
		else
		{
			buf1 = abuf;
		}

		if (header && headlen > 0)
		{
			ret = MIN(headlen, size);
			memcpy(buf1, header, ret);
			headlen -= ret;
			header += ret;
			if (headlen <= 0)
				header = NULL;
		}

		if (s > ret)
			ret += read(infd, buf1 + ret, s - ret);

		if (ret > 0)
		{
			ret2 =
				(outfd >=
				0 ? write_data(outfd, buf1, ret) : ret);
			if (ret2 > 0)
				total += ret2;
			/* if we can't write then dump excess data */
			if (ret2 != ret)
				transfer_file(infd, -1, n - (ret + headlen),
					      NULL, 0, 0);
		}
		if (ret <= 0 || ret2 != ret)
			return (total);
		n -= ret;
	}
	return (total);
}

/*******************************************************************
 Sleep for a specified number of milliseconds.
********************************************************************/

void msleep(int t)
{
#ifdef HAVE_OS_WIN32
	Sleep(t);
#else
	int tdiff = 0;
	struct timeval tval, t1, t2;
	fd_set fds;

	GetTimeOfDay(&t1);
	GetTimeOfDay(&t2);

	while (tdiff < t)
	{
		tval.tv_sec = (t - tdiff) / 1000;
		tval.tv_usec = 1000 * ((t - tdiff) % 1000);

		FD_ZERO(&fds);
		errno = 0;
		sys_select_intr(0, &fds, &tval);

		GetTimeOfDay(&t2);
		tdiff = TvalDiff(&t1, &t2);
	}
#endif
}


/****************************************************************************
 Put up a yes/no prompt
****************************************************************************/
BOOL yesno(char *fmt, ...)
{
	pstring ans;
	va_list args;

	va_start(args, fmt);
	vfprintf(stdout, fmt, args);
	fflush(stdout);
	va_end(args);

	if (!fgets(ans, sizeof(ans) - 1, stdin))
		return (False);

	if (*ans == 'y' || *ans == 'Y')
		return (True);

	return (False);
}



/****************************************************************************
set the length of a file from a filedescriptor.
Returns 0 on success, -1 on failure.
****************************************************************************/

/* tpot vfs need to recode this function */

int set_filelen(int fd, SMB_OFF_T len)
{
/* According to W. R. Stevens advanced UNIX prog. Pure 4.3 BSD cannot
   extend a file with ftruncate. Provide alternate implementation
   for this */

#ifdef HAVE_FTRUNCATE_EXTEND
	return sys_ftruncate(fd, len);
#else
	SMB_STRUCT_STAT st;
	char c = 0;
	SMB_OFF_T currpos = sys_lseek(fd, (SMB_OFF_T) 0, SEEK_CUR);

	if (currpos == -1)
		return -1;
	/* Do an fstat to see if the file is longer than
	   the requested size (call ftruncate),
	   or shorter, in which case seek to len - 1 and write 1
	   byte of zero */
	if (sys_fstat(fd, &st) < 0)
		return -1;

#ifdef S_ISFIFO
	if (S_ISFIFO(st.st_mode))
		return 0;
#endif

	if (st.st_size == len)
		return 0;
	if (st.st_size > len)
		return sys_ftruncate(fd, len);

	if (sys_lseek(fd, len - 1, SEEK_SET) != len - 1)
		return -1;
	if (write(fd, &c, 1) != 1)
		return -1;
	/* Seek to where we were */
	if (sys_lseek(fd, currpos, SEEK_SET) != currpos)
		return -1;
	return 0;
#endif
}


/****************************************************************************
free memory, checks for NULL
****************************************************************************/
void safe_free(void *p)
{
	if (p != NULL)
	{
		free(p);
	}
}

/****************************************************************************
 Get my own name and IP.
****************************************************************************/

BOOL get_myname(char *my_name)
{
	struct hostent *hp;
	pstring hostname;

	*hostname = 0;

	/* get my host name */
	if (gethostname(hostname, sizeof(hostname)) == -1)
	{
		DEBUG(0, ("gethostname failed\n"));
		return False;
	}

	/* get host info */
	if ((hp = Get_Hostbyname(hostname)) == 0)
	{
		DEBUG(0, ("Get_Hostbyname: Unknown host %s\n", hostname));
		return False;
	}
	/* Ensure null termination. */
	hostname[sizeof(hostname) - 1] = '\0';

	if (my_name)
	{
		/* split off any parts after an initial . */
		char *p = strchr(hostname, '.');
		if (p)
			*p = 0;

		fstrcpy(my_name, hostname);
	}

#if 0
	if (ip)
		putip((char *)ip, (char *)hp->h_addr);
#endif

	return (True);
}

/****************************************************************************
 Interpret a protocol description string, with a default.
****************************************************************************/

enum protocol_types interpret_protocol(char *str, enum protocol_types def)
{
	static struct field_info protos[] =
	{
		{ PROTOCOL_NT1,     "NT1", NULL },
		{ PROTOCOL_LANMAN2, "LANMAN2", NULL },
		{ PROTOCOL_LANMAN1, "LANMAN1", NULL },
		{ PROTOCOL_COREPLUS, "COREPLUS", NULL },
		{ PROTOCOL_COREPLUS, "CORE+", NULL },
		{ PROTOCOL_CORE, "CORE", NULL },
		{ 0, NULL, NULL }
	};
	enum protocol_types p = str_to_enum_field (str, protos, 0);
	if (!p)
	{
		DEBUG(0, ("Unrecognised protocol level %s\n", str));
		return def;
	}
	return p;
}

/****************************************************************************
 Return true if a string could be a pure IP address.
****************************************************************************/

BOOL is_ipaddress(const char *str)
{
  BOOL pure_address = True;
  int i;
  
  for (i=0; pure_address && str[i]; i++)
    if (!(isdigit((int)str[i]) || str[i] == '.'))
      pure_address = False;

  /* Check that a pure number is not misinterpreted as an IP */
  pure_address = pure_address && (strchr(str, '.') != NULL);

  return pure_address;
}

/****************************************************************************
interpret an internet address or name into an IP address in 4 byte form
****************************************************************************/

uint32 interpret_addr(const char *str)
{
	struct hostent *hp;
	uint32 res;
	int i;
	BOOL pure_address = True;

	if (strcmp(str, "0.0.0.0") == 0)
		return (0);
	if (strcmp(str, "255.255.255.255") == 0)
		return (0xFFFFFFFF);

	for (i = 0; pure_address && str[i]; i++)
		if (!(isdigit((int)str[i]) || str[i] == '.'))
			pure_address = False;

	/* if it's in the form of an IP address then get the lib to interpret it */
	if (pure_address)
	{
		res = inet_addr(str);
	}
	else
	{
		/* otherwise assume it's a network name of some sort and use 
		   Get_Hostbyname */
		if ((hp = Get_Hostbyname(str)) == 0)
		{
			DEBUG(3, ("Get_Hostbyname: Unknown host. %s\n", str));
			return 0;
		}
		if (hp->h_addr == NULL)
		{
			DEBUG(3,
			      ("Get_Hostbyname: host address is invalid for host %s\n",
			       str));
			return 0;
		}
		putip((char *)&res, (char *)hp->h_addr);
	}

	if (res == (uint32)-1)
		return (0);

	return (res);
}

/*******************************************************************
  a convenient addition to interpret_addr()
  ******************************************************************/
struct in_addr *interpret_addr2(const char *str)
{
	static struct in_addr ret;
	uint32 a = interpret_addr(str);
	ret.s_addr = a;
	return (&ret);
}

/*******************************************************************
 Check if an IP is the 0.0.0.0
  ******************************************************************/
BOOL is_zero_ip(struct in_addr ip)
{
	uint32 a;
	putip((char *)&a, (char *)&ip);
	return (a == 0);
}

/*******************************************************************
 Set an IP to 0.0.0.0
 ******************************************************************/

void zero_ip(struct in_addr *ip)
{
        static BOOL init;
        static struct in_addr ipzero;

        if (!init) {
                ipzero = *interpret_addr2("0.0.0.0");
                init = True;
        }

        *ip = ipzero;
}

#if (defined(HAVE_NETGROUP) && defined(WITH_AUTOMOUNT))
/******************************************************************
 Remove any mount options such as -rsize=2048,wsize=2048 etc.
 Based on a fix from <Thomas.Hepper@icem.de>.
*******************************************************************/

static void strip_mount_options(pstring *str)
{
	if (**str == '-')
	{
		char *p = *str;
		while (*p && !isspace(*p))
			p++;
		while (*p && isspace(*p))
			p++;
		if (*p)
		{
			pstring tmp_str;

			pstrcpy(tmp_str, p);
			pstrcpy(*str, tmp_str);
		}
	}
}

/*******************************************************************
 Patch from jkf@soton.ac.uk
 Split Luke's automount_server into YP lookup and string splitter
 so can easily implement automount_path(). 
 As we may end up doing both, cache the last YP result. 
*******************************************************************/

#ifdef WITH_NISPLUS_HOME
char *automount_lookup(const char *user_name)
{
	static fstring last_key = "";
	static pstring last_value = "";

	char *nis_map = (char *)lp_nis_home_map_name();

	char nis_domain[NIS_MAXNAMELEN + 1];
	char buffer[NIS_MAXATTRVAL + 1];
	nis_result *result;
	nis_object *object;
	entry_obj *entry;

	strncpy(nis_domain, (char *)nis_local_directory(), NIS_MAXNAMELEN);
	nis_domain[NIS_MAXNAMELEN] = '\0';

	DEBUG(5, ("NIS+ Domain: %s\n", nis_domain));

	if (strcmp(user_name, last_key))
	{
		slprintf(buffer, sizeof(buffer) - 1, "[%s=%s]%s.%s", "key",
			 user_name, nis_map, nis_domain);
		DEBUG(5, ("NIS+ querystring: %s\n", buffer));

		if (result = nis_list(buffer, RETURN_RESULT, NULL, NULL))
		{
			if (result->status != NIS_SUCCESS)
			{
				DEBUG(3,
				      ("NIS+ query failed: %s\n",
				       nis_sperrno(result->status)));
				fstrcpy(last_key, "");
				pstrcpy(last_value, "");
			}
			else
			{
				object = result->objects.objects_val;
				if (object->zo_data.zo_type == ENTRY_OBJ)
				{
					entry =
						&object->zo_data.objdata_u.
						en_data;
					DEBUG(5,
					      ("NIS+ entry type: %s\n",
					       entry->en_type));
					DEBUG(3,
					      ("NIS+ result: %s\n",
					       entry->en_cols.en_cols_val[1].
					       ec_value.ec_value_val));

					pstrcpy(last_value,
						entry->en_cols.
						en_cols_val[1].ec_value.
						ec_value_val);
					pstring_sub(last_value, "&",
						    user_name);
					fstrcpy(last_key, user_name);
				}
			}
		}
		nis_freeresult(result);
	}

	strip_mount_options(&last_value);

	DEBUG(4, ("NIS+ Lookup: %s resulted in %s\n", user_name, last_value));
	return last_value;
}
#else /* WITH_NISPLUS_HOME */
char *automount_lookup(const char *user_name)
{
	static fstring last_key = "";
	static pstring last_value = "";

	int nis_error;		/* returned by yp all functions */
	char *nis_result;	/* yp_match inits this */
	int nis_result_len;	/* and set this */
	char *nis_domain;	/* yp_get_default_domain inits this */
	char *nis_map = (char *)lp_nis_home_map_name();

	if ((nis_error = yp_get_default_domain(&nis_domain)) != 0)
	{
		DEBUG(3, ("YP Error: %s\n", yperr_string(nis_error)));
		return last_value;
	}

	DEBUG(5, ("NIS Domain: %s\n", nis_domain));

	if (!strcmp(user_name, last_key))
	{
		nis_result = last_value;
		nis_result_len = strlen(last_value);
		nis_error = 0;
	}
	else
	{
		if ((nis_error = yp_match(nis_domain, nis_map,
					  user_name, strlen(user_name),
					  &nis_result, &nis_result_len)) != 0)
		{
			DEBUG(3,
			      ("YP Error: \"%s\" while looking up \"%s\" in map \"%s\"\n",
			       yperr_string(nis_error), user_name, nis_map));
		}
		if (!nis_error && nis_result_len >= sizeof(pstring))
		{
			nis_result_len = sizeof(pstring) - 1;
		}
		fstrcpy(last_key, user_name);
		strncpy(last_value, nis_result, nis_result_len);
		last_value[nis_result_len] = '\0';
	}

	strip_mount_options(&last_value);

	DEBUG(4, ("YP Lookup: %s resulted in %s\n", user_name, last_value));
	return last_value;
}
#endif /* WITH_NISPLUS_HOME */
#endif


/*******************************************************************
are two IPs on the same subnet?
********************************************************************/
BOOL same_net(struct in_addr ip1, struct in_addr ip2, struct in_addr mask)
{
	uint32 net1, net2, nmask;

	nmask = ntohl(mask.s_addr);
	net1 = ntohl(ip1.s_addr);
	net2 = ntohl(ip2.s_addr);

	return ((net1 & nmask) == (net2 & nmask));
}


/****************************************************************************
check if a process exists. Does this work on all unixes?
****************************************************************************/

BOOL process_exists(pid_t pid)
{
#ifdef HAVE_OS_WIN32
	return False;
#else
	return (kill(pid, 0) == 0 || errno != ESRCH);
#endif
}

/*******************************************************************
something really nasty happened - panic!
********************************************************************/
void smb_panic(char *why)
{
	char *cmd = lp_panic_action();
	if (cmd && *cmd)
	{
		system(cmd);
	}
	DEBUG(0, ("PANIC: %s\n", why));
	dbgflush();
	abort();
}


/*******************************************************************
a readdir wrapper which just returns the file name
********************************************************************/
char *readdirname(DIR * p)
{
	SMB_STRUCT_DIRENT *ptr;
	char *dname;

	if (!p)
		return (NULL);

	ptr = (SMB_STRUCT_DIRENT *) sys_readdir(p);
	if (!ptr)
		return (NULL);

	dname = ptr->d_name;

#ifdef NEXT2
	if (telldir(p) < 0)
		return (NULL);
#endif

#ifdef HAVE_BROKEN_READDIR
	/* using /usr/ucb/cc is BAD */
	dname = dname - 2;
#endif

	{
		static pstring buf;
		memcpy(buf, dname, NAMLEN(ptr) + 1);
		dname = buf;
	}

	return (dname);
}


/****************************************************************************
 Simple routine to do POSIX file locking. Cruft in NFS and 64->32 bit mapping
 is dealt with in posix.c
****************************************************************************/

BOOL fcntl_lock(int fd, int op, SMB_OFF_T offset, SMB_OFF_T count, int type)
{
#ifdef HAVE_OS_WIN32
	DEBUG(0, ("fcntl_lock: not yet implemented\n"));
	return True;
#else
	SMB_STRUCT_FLOCK lock;
	int ret;

	DEBUG(8,
	      ("fcntl_lock %d %d %.0f %.0f %d\n", fd, op, (double)offset,
	       (double)count, type));

	lock.l_type = type;
	lock.l_whence = SEEK_SET;
	lock.l_start = offset;
	lock.l_len = count;
	lock.l_pid = 0;

	errno = 0;

	ret = fcntl(fd, op, &lock);

	if (errno != 0)
		DEBUG(3,
		      ("fcntl_lock: fcntl lock gave errno %d (%s)\n", errno,
		       strerror(errno)));

	/* a lock query */
	if (op == SMB_F_GETLK)
	{
		if ((ret != -1) &&
		    (lock.l_type != F_UNLCK) &&
		    (lock.l_pid != 0) && (lock.l_pid != sys_getpid()))
		{
			DEBUG(3,
			      ("fcntl_lock: fd %d is locked by pid %d\n", fd,
			       (int)lock.l_pid));
			return (True);
		}

		/* it must be not locked or locked by me */
		return (False);
	}

	/* a lock set or unset */
	if (ret == -1)
	{
		DEBUG(3,
		      ("fcntl_lock: lock failed at offset %.0f count %.0f op %d type %d (%s)\n",
		       (double)offset, (double)count, op, type,
		       strerror(errno)));
		return (False);
	}

	/* everything went OK */
	DEBUG(8, ("fcntl_lock: Lock call successful\n"));

	return (True);
#endif
}


void out_ascii(FILE * f, const uchar * buf, int len)
{
	int i;
	for (i = 0; i < len; i++)
	{
		fprintf(f, "%c", isprint(buf[i]) ? buf[i] : '.');
	}
}

void out_struct(FILE * f, const char *buf1, int len, int per_line)
{
	const uchar *buf = (const uchar *)buf1;
	int i;

	if (len <= 0)
	{
		return;
	}

	fprintf(f, "{\n\t");
	for (i = 0; i < len;)
	{
		fprintf(f, "0x%02X", (int)buf[i]);
		i++;
		if (i != len)
		{
			fprintf(f, ", ");
		}
		if (i % per_line == 0 && i != len)
		{
			fprintf(f, "\n\t");
		}
	}
	fprintf(f, "\n};\n");
}

void out_data(FILE * f, const char *buf1, int len,
	      int per_line, const char *indstr)
{
	const uchar *buf = (const uchar *)buf1;
	int i = 0;
	if (len <= 0)
	{
		return;
	}
	if (f ==  NULL)
	{
		return;
	}
	if (buf1 == NULL)
	{
		DEBUG(1, ("out_data: NULL, len=%d\n", len));
		return;
	}
	if (indstr == NULL)
	{
		indstr = "";
	}

	fprintf(f, "%s[%03X] ", indstr, i);
	for (i = 0; i < len;)
	{
		fprintf(f, "%02X ", (int)buf[i]);
		i++;
		if (i % (per_line / 2) == 0)
			fprintf(f, " ");
		if (i % per_line == 0)
		{
			out_ascii(f, &buf[i - per_line], per_line / 2);
			fprintf(f, " ");
			out_ascii(f, &buf[i - per_line / 2], per_line / 2);
			fprintf(f, "\n");
			if (i < len)
				fprintf(f, "%s[%03X] ", indstr, i);
		}
	}
	if ((i % per_line) != 0)
	{
		int n;

		n = per_line - (i % per_line);
		fprintf(f, " ");
		if (n > (per_line / 2))
			fprintf(f, " ");
		while (n--)
		{
			fprintf(f, "   ");
		}
		n = MIN(per_line / 2, i % per_line);
		out_ascii(f, &buf[i - (i % per_line)], n);
		fprintf(f, " ");
		n = (i % per_line) - n;
		if (n > 0)
			out_ascii(f, &buf[i - n], n);
		fprintf(f, "\n");
	}
}

void print_asc(int level, uchar const *buf, size_t len)
{
	size_t i;
	for (i = 0; i < len; i++)
	{
		DEBUGADD(level, ("%c", isprint(buf[i]) ? buf[i] : '.'));
	}
}

void dump_data(int level, const void *buf1, size_t len)
{
	uchar const *buf = (uchar const *)buf1;
	size_t i = 0;
	const uchar *pfirstproblem;

	pfirstproblem = (void *)VALGRIND_CHECK_READABLE(buf, len);

	if (pfirstproblem)
	{
		size_t firstproblem = pfirstproblem - buf;

		DEBUG(0, ("dump_data: Only the first %u bytes of %u are "
			  "valid, please check. Dumping only those.\n",
			  firstproblem, len));
		len = MIN(len, firstproblem);
		level = 0; /* Dump it! */
	}

	if (! DEBUGLVL(level))
		return;

	if (buf == NULL)
	{
		DEBUGADD(level, ("dump_data: NULL, len=%u\n", len));
		return;
	}
	if (len <= 0)
	{
		DEBUGADD(level, ("dump_data: len=0\n"));
		return;
	}

	DEBUGADD(level, ("[%03X] ", i));
	for (i = 0; i < len;)
	{
		DEBUGADD(level, ("%02X ", (int)buf[i]));
		i++;
		if (i % 8 == 0)
			DEBUGADD(level, (" "));
		if (i % 16 == 0)
		{
			print_asc(level, &buf[i - 16], 8);
			DEBUGADD(level, (" "));
			print_asc(level, &buf[i - 8], 8);
			DEBUGADD(level, ("\n"));
			if (i < len)
				DEBUGADD(level, ("[%03X] ", i));
		}
	}

	if (i % 16 != 0)	/* finish off a non-16-char-length row */
	{
		size_t n;

		n = 16 - (i % 16);
		DEBUGADD(level, (" "));
		if (n > 8)
			DEBUGADD(level, (" "));
		while (n--)
			DEBUGADD(level, ("   "));

		n = MIN(8, i % 16);
		print_asc(level, &buf[i - (i % 16)], n);
		DEBUGADD(level, (" "));
		n = (i % 16) - n;
		if (n > 0)
			print_asc(level, &buf[i - n], n);
		DEBUGADD(level, ("\n"));
	}
}

void dump_data_pw(const char *msg, const void *data, size_t len)
{
#ifdef DEBUG_PASSWORD
	DEBUG(100, ("%s", msg));
	if (data != NULL && len > 0)
	{
		dump_data(100, data, len);
	}
#endif
}

char *tab_depth(int depth)
{
	static pstring spaces;
	memset(spaces, ' ', depth * 4);
	spaces[depth * 4] = 0;
	return spaces;
}

/*****************************************************************************
 * Provide a checksum on a string
 *
 *  Input:  s - the null-terminated character string for which the checksum
 *              will be calculated.
 *
 *  Output: The checksum value calculated for s.
 *
 * ****************************************************************************
 */
int str_checksum(const char *s)
{
	int res = 0;
	int c;
	int i = 0;

	while (*s)
	{
		c = *s;
		res ^= (c << (i % 15)) ^ (c >> (15 - (i % 15)));
		s++;
		i++;
	}
	return (res);
}				/* str_checksum */



/*****************************************************************
zero a memory area then free it. Used to catch bugs faster
*****************************************************************/
void zero_free(void *p, size_t size)
{
	memset(p, 0, size);
	free(p);
}


/*****************************************************************
set our open file limit to a requested max and return the limit
*****************************************************************/
int set_maxfiles(int requested_max)
{
#if (defined(HAVE_GETRLIMIT) && defined(RLIMIT_NOFILE))
	struct rlimit rlp;
	int saved_current_limit;

	if (getrlimit(RLIMIT_NOFILE, &rlp))
	{
		DEBUG(0,
		      ("set_maxfiles: getrlimit (1) for RLIMIT_NOFILE failed with error %s\n",
		       strerror(errno)));
		/* just guess... */
		return requested_max;
	}

	/* 
	   * Set the fd limit to be real_max_open_files + MAX_OPEN_FUDGEFACTOR to
	   * account for the extra fd we need 
	   * as well as the log files and standard
	   * handles etc. Save the limit we want to set in case
	   * we are running on an OS that doesn't support this limit (AIX)
	   * which always returns RLIM_INFINITY for rlp.rlim_max.
	 */

	saved_current_limit = rlp.rlim_cur = MIN(requested_max, rlp.rlim_max);

	if (setrlimit(RLIMIT_NOFILE, &rlp))
	{
		DEBUG(0,
		      ("set_maxfiles: setrlimit for RLIMIT_NOFILE for %d files failed with error %s\n",
		       (int)rlp.rlim_cur, strerror(errno)));
		/* just guess... */
		return saved_current_limit;
	}

	if (getrlimit(RLIMIT_NOFILE, &rlp))
	{
		DEBUG(0,
		      ("set_maxfiles: getrlimit (2) for RLIMIT_NOFILE failed with error %s\n",
		       strerror(errno)));
		/* just guess... */
		return saved_current_limit;
	}

#if defined(RLIM_INFINITY)
	if (rlp.rlim_cur == RLIM_INFINITY)
		return saved_current_limit;
#endif

	if ((int)rlp.rlim_cur > saved_current_limit)
		return saved_current_limit;

	return rlp.rlim_cur;
#else /* !defined(HAVE_GETRLIMIT) || !defined(RLIMIT_NOFILE) */
	/*
	 * No way to know - just guess...
	 */
	return requested_max;
#endif
}


/*****************************************************************
 splits out the start of the key (HKLM or HKU) and the rest of the key
 *****************************************************************/
BOOL reg_split_key(const char *full_keyname, uint32 *reg_type, char *key_name)
{
	/* STRIP_CONST is safe because next_token treats its argument
	   as 'const char **' - see comment in util_str.c */
	char *full_key_tmp = STRIP_CONST(full_keyname);
	pstring tmp;
	static struct field_info well_known_hk[] =
	{
		{ HKEY_CLASSES_ROOT, "HKCR", NULL },
		{ HKEY_CLASSES_ROOT, "HKEY_CLASSES_ROOT", NULL },
		{ HKEY_CURRENT_USER, "HKCU", NULL },
		{ HKEY_CURRENT_USER, "HKEY_CURRENT_USER", NULL },
		{ HKEY_LOCAL_MACHINE, "HKLM", NULL },
		{ HKEY_LOCAL_MACHINE, "HKEY_LOCAL_MACHINE", NULL },
		{ HKEY_USERS,         "HKU", NULL },
		{ HKEY_USERS,         "HKEY_USERS", NULL },
		{ HKEY_PERFORMANCE_DATA,  "HKEY_PERFORMANCE", NULL },
		{ HKEY_PERFORMANCE_DATA,  "HKP", NULL },
		{ HKEY_PERFORMANCE_DATA,  "HKPD", NULL },
		{ HKEY_PERFORMANCE_DATA,  "HKEY_PERFORMANCE_DATA", NULL },
		{ 0, NULL, NULL }
	};

	if (!next_token(&full_key_tmp, tmp, "\\", sizeof(tmp)))
	{
		return False;
	}

	DEBUG(10, ("reg_split_key: hive %s\n", tmp));

	if (! (*reg_type = str_to_enum_field (tmp, well_known_hk, 0)))
	{
		DEBUG(10, ("reg_split_key: unrecognised hive key %s\n", tmp));
		return False;
	}

	if (!next_token(NULL, key_name, "\n\r", sizeof(fstring)))
	{
		key_name[0] = 0;
	}

	DEBUG(10, ("reg_split_key: name %s\n", key_name));

	return True;
}

/**********************************************************
 Map the Active Directory userAccountControl attribute to
 SMB account control bits. The attribute is stored in the
 directory as an integer.
 **********************************************************/
NTDS_USER_FLAG_ENUM pwdb_acct_ctrl_to_ad(uint32 smbac)
{
	NTDS_USER_FLAG_ENUM adac;

	adac = 0;

	adac |= NTDS_UF_TRUSTED_FOR_DELEGATION;	/* | NTDS_UF_USE_DES_KEY_ONLY; */

	if (smbac & ACB_DISABLED)
		adac |= NTDS_UF_ACCOUNTDISABLE;
	if (smbac & ACB_HOMDIRREQ)
		adac |= NTDS_UF_HOMEDIR_REQUIRED;
	if (smbac & ACB_PWNOTREQ)
		adac |= NTDS_UF_PASSWD_NOTREQD;
	if (smbac & ACB_TEMPDUP)
		adac |= NTDS_UF_TEMP_DUPLICATE_ACCOUNT;
	if (smbac & ACB_NORMAL)
		adac |= NTDS_UF_NORMAL_ACCOUNT;
	if (smbac & ACB_MNS)
		adac |= NTDS_UF_MNS_LOGON_ACCOUNT;
	if (smbac & ACB_DOMTRUST)
		adac |= NTDS_UF_INTERDOMAIN_TRUST_ACCOUNT;
	if (smbac & ACB_WSTRUST)
		adac |= NTDS_UF_WORKSTATION_TRUST_ACCOUNT;
	if (smbac & ACB_SVRTRUST)
		adac |= NTDS_UF_SERVER_TRUST_ACCOUNT;
	if (smbac & ACB_PWNOEXP)
		adac |= NTDS_UF_DONT_EXPIRE_PASSWD;
	if (smbac & ACB_AUTOLOCK)
		adac |= NTDS_UF_LOCKOUT;
	if (smbac & ACB_PWLOCK)
		adac |= NTDS_UF_PASSWD_CANT_CHANGE;

	return adac;
}

/**********************************************************
 Map the Active Directory userAccountControl attribute to
 SMB account control bits. The attribute is stored in the
 directory as an integer.
 **********************************************************/
uint32 pwdb_acct_ctrl_from_ad(NTDS_USER_FLAG_ENUM adac)
{
	int smbac;

	smbac = 0;

	if (adac & NTDS_UF_ACCOUNTDISABLE)
		adac |= ACB_DISABLED;
	if (adac & NTDS_UF_HOMEDIR_REQUIRED)
		adac |= ACB_HOMDIRREQ;;
	if (adac & NTDS_UF_PASSWD_NOTREQD)
		adac |= ACB_PWNOTREQ;
	if (adac & NTDS_UF_TEMP_DUPLICATE_ACCOUNT)
		adac |= ACB_TEMPDUP;
	if (adac & NTDS_UF_NORMAL_ACCOUNT)
		adac |= ACB_NORMAL;
	if (adac & NTDS_UF_MNS_LOGON_ACCOUNT)
		adac |= ACB_MNS;
	if (adac & NTDS_UF_INTERDOMAIN_TRUST_ACCOUNT)
		adac |= ACB_DOMTRUST;
	if (adac & NTDS_UF_WORKSTATION_TRUST_ACCOUNT)
		adac |= ACB_WSTRUST;
	if (adac & NTDS_UF_SERVER_TRUST_ACCOUNT)
		adac |= ACB_SVRTRUST;
	if (adac & NTDS_UF_DONT_EXPIRE_PASSWD)
		adac |= ACB_PWNOEXP;
	if (adac & NTDS_UF_LOCKOUT)
		adac |= ACB_AUTOLOCK;
	if (adac & NTDS_UF_PASSWD_CANT_CHANGE)
		adac |= ACB_PWLOCK;

	return smbac;
}

/*******************************************************************
 gets password-database-format time from a string.
 ********************************************************************/

static time_t get_time_from_string(const char *p)
{
	int i;

	for (i = 0; i < 8; i++)
	{
		if (p[i] == '\0' || !isxdigit((int)(p[i] & 0xFF)))
		{
			break;
		}
	}
	if (i == 8)
	{
		/*
		 * p points at 8 characters of hex digits - 
		 * read into a time_t as the seconds since
		 * 1970 that the password was last changed.
		 */
		return (time_t) strtol(p, NULL, 16);
	}
	return (time_t) - 1;
}

/*******************************************************************
 gets password time last changed time
 ********************************************************************/

time_t pwdb_get_time_last_changed(const char *p)
{
	if (*p && !StrnCaseCmp(p, "TLC-", 4))
	{
		return get_time_from_string(p + 4);
	}
	return (time_t) - 1;
}

/*******************************************************************
 gets password last set time
 ********************************************************************/

time_t pwdb_get_last_set_time(const char *p)
{
	if (*p && !StrnCaseCmp(p, "LCT-", 4))
	{
		return get_time_from_string(p + 4);
	}
	return (time_t) - 1;
}


/*******************************************************************
 sets password-database-format time in a string.
 ********************************************************************/
static void set_time_in_string(char *p, int max_len, char *type, time_t t)
{
	slprintf(p, max_len, ":%s-%08X", type, (uint32)t);
}

/*******************************************************************
 sets logon time
 ********************************************************************/
void pwdb_set_logon_time(char *p, int max_len, time_t t)
{
	set_time_in_string(p, max_len, "LNT", t);
}

/*******************************************************************
 sets logoff time
 ********************************************************************/
void pwdb_set_logoff_time(char *p, int max_len, time_t t)
{
	set_time_in_string(p, max_len, "LOT", t);
}

/*******************************************************************
 sets kickoff time
 ********************************************************************/
void pwdb_set_kickoff_time(char *p, int max_len, time_t t)
{
	set_time_in_string(p, max_len, "KOT", t);
}

/*******************************************************************
 sets password can change time
 ********************************************************************/
void pwdb_set_can_change_time(char *p, int max_len, time_t t)
{
	set_time_in_string(p, max_len, "CCT", t);
}

/*******************************************************************
 sets password last change time
 ********************************************************************/
void pwdb_set_time_last_changed(char *p, int max_len, time_t t)
{
	set_time_in_string(p, max_len, "TLC", t);
}

/*******************************************************************
 sets password last set time
 ********************************************************************/
void pwdb_set_must_change_time(char *p, int max_len, time_t t)
{
	set_time_in_string(p, max_len, "MCT", t);
}

/*******************************************************************
 sets password last set time
 ********************************************************************/
void pwdb_set_last_set_time(char *p, int max_len, time_t t)
{
	set_time_in_string(p, max_len, "LCT", t);
}

/*****************************************************************
possibly replace mkstemp if it is broken
*****************************************************************/  
int smb_mkstemp(char *template)
{
#if HAVE_SECURE_MKSTEMP
	return mkstemp(template);
#else
	/* have a reasonable go at emulating it. Hope that
	   the system mktemp() isn't completly hopeless */
	char *p = mktemp(template);
	if (!p) return -1;
	return open(p, O_CREAT|O_EXCL|O_RDWR, 0600);
#endif
}




/*****************************************************************
like strdup but for memory
 *****************************************************************/
void *tng_memdup(const void *p, size_t size)
{
	void *p2;
	if (!p)
		return NULL;
	if (size == 0)
		return NULL;
	p2 = g_malloc(size);
	if (!p2)
		return NULL;
	memcpy(p2, p, size);
	return p2;
}

/*****************************************************************
get local hostname and cache result
 *****************************************************************/
char *myhostname(void)
{
	static pstring ret;
	if (ret[0] == 0)
	{
		get_myname(ret);
	}
	return ret;
}

/*****************************************************************
return pathname in directory, making sure directory exists
NOTE: returns static storage - not mt-safe
******************************************************************/
static char *get_path (const char *name, const char *dir)
{
	static pstring fname;

	pstrcpy(fname, dir);
	trim_char(fname, 0, '/');

	if (!directory_exist(fname, NULL))
	{
		sys_mkdir(fname, 0755);
	}

	pstrcat(fname, "/");
	pstrcat(fname, name);

	return fname;
}

/*****************************************************************
return a path in the runtime temp directory
 *****************************************************************/
char *tmp_path(char *name)
{
	return get_path(name, lp_tmpdir());
}

/*****************************************************************
return a path in the state config directory
 *****************************************************************/
char *state_path(char *name)
{
	return get_path(name, lp_statedir());
}

/*****************************************************************
return a path in the sam directory
 *****************************************************************/
char *passdb_path(char *name)
{
	return get_path(name, lp_sam_directory());
}

/*******************************************************************
 Given a filename - get its directory name
 NB: Returned in static storage.  Caveats:
 o  Not safe in thread environment.
 o  Caller must not free.
 o  If caller wishes to preserve, they should copy.
********************************************************************/

char *parent_dirname(const char *path)
{
	static pstring dirpath;
	char *p;

	if (!path)
		return (NULL);

	pstrcpy(dirpath, path);
	p = strrchr(dirpath, '/');	/* Find final '/', if any */
	if (!p)
	{
		pstrcpy(dirpath, ".");	/* No final "/", so dir is "." */
	}
	else
	{
		if (p == dirpath)
			++p;	/* For root "/", leave "/" in place */
		*p = '\0';
	}
	return dirpath;
}


/*******************************************************************
determine if a pattern contains any Microsoft wildcard characters
 *******************************************************************/
BOOL ms_has_wild(char *s)
{
	char c;
	while ((c = *s++))
	{
		switch (c)
		{
			case '*':
			case '?':
			case '<':
			case '>':
			case '"':
				return True;
		}
	}
	return False;
}


/*******************************************************************
 a wrapper that handles case sensitivity and the special handling
   of the ".." name
 *******************************************************************/
BOOL mask_match(const char *string, const char *pattern, BOOL is_case_sensitive)
{
	if (strcmp(string, "..") == 0)
		string = ".";
	if (strcmp(pattern, ".") == 0)
		return False;

	return ms_fnmatch(pattern, string, Protocol, is_case_sensitive) == 0;
}

#ifdef __INSURE__

/*******************************************************************
This routine is a trick to immediately catch errors when debugging
with insure. A xterm with a gdb is popped up when insure catches
a error. It is Linux specific.
********************************************************************/
int _Insure_trap_error(int a1, int a2, int a3, int a4, int a5, int a6)
{
	static int (*fn) ();
	int ret;
	char pidstr[10];
	pstring cmd = "/usr/X11R6/bin/xterm -display :0 -T Panic -n Panic -e /bin/sh -c 'cat /tmp/ierrs.*.%d ; gdb /proc/%d/exe %d'";

	slprintf(pidstr, sizeof(pidstr)-1, "%d", sys_getpid());
	pstring_sub(cmd, "%d", pidstr);

	if (!fn)
	{
		static void *h;
		h =
			dlopen
			("/usr/local/parasoft/insure++lite/lib.linux2/libinsure.so",
			 RTLD_LAZY);
		fn = dlsym(h, "_Insure_trap_error");
	}

	ret = fn(a1, a2, a3, a4, a5, a6);

	system(cmd);

	return ret;
}
#endif
