/* 
   Copyright (C) Elrond <elrond@samba-tng.org>  2001.
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#include "includes.h"
#include "libsamba.h"
#include "tng_misc.h"
#include "rpc_parse.h"
#include "sids.h"
#include "others.h"


static NT_USER_TOKEN *nt_token_new(const NET_USER_INFO_3 *user3, BOOL guest)
{
	size_t num_sids;
	size_t i;
	uint32 group_rid;
	DOM_SID domain_sid;
	DOM_SID *sid;
	NT_USER_TOKEN *token;

	if (!user3)
		return NULL;

	token = g_new(NT_USER_TOKEN, 1);
	if (!token)
		return NULL;

	sid_copy(&domain_sid, &user3->dom_sid.sid);

	/*
	 * 2: user, group
	 * 1: "Everyone"
	 * !guest: "Authenticated Users"
	 */
	num_sids = 2 + user3->num_groups + user3->num_other_sids + 1
		+ ((!guest) ? 1 : 0);

	group_rid = user3->group_id;

	if (user3->user_id == group_rid)
		num_sids--;

	for (i = 0; i < user3->num_groups; i++)
	{
		if (user3->gids[i].g_rid == group_rid)
			num_sids--;
	}

	token->num_sids = num_sids;
	token->user_sids = g_new(DOM_SID, num_sids);
	if (!token->user_sids)
	{
		free(token);
		return NULL;
	}

	sid = &token->user_sids[0];

	sid_copy(sid, &domain_sid);
	sid_append_rid(sid, user3->user_id);
	sid++;
	if (user3->user_id != group_rid)
	{
		sid_copy(sid, &domain_sid);
		sid_append_rid(sid, group_rid);
		sid++;
	}

	for (i = 0; i < user3->num_groups; i++)
	{
		if (user3->gids[i].g_rid == group_rid)
			continue;
		sid_copy(sid, &domain_sid);
		sid_append_rid(sid, user3->gids[i].g_rid);
		sid++;
	}

	for (i = 0; i < user3->num_other_sids; i++)
	{
		sid_copy(sid, &user3->other_sids[i].sid);
		sid++;
	}

	sid_copy(sid, global_sid_everyone);
	sid++;

	if (!guest)
	{
		/* should use global_sid_something */
		string_to_sid(sid, "S-1-5-11");
		sid++;
	}

	return token;
}

static void nt_token_show(const NT_USER_TOKEN *token)
{
	size_t i;

	if (!DEBUGLVL(2))
		return;

	DEBUGADD(2, ("nt_token:"));

	if (!token)
	{
		DEBUGADD(2, (" NULL\n"));
		return;
	}

	DEBUGADD(2, (" %u SIDs\n", token->num_sids));

	for (i = 0; i < token->num_sids; i++)
	{
		fstring sid_str;
		DEBUGADD(2, ("  %s\n",
			     sid_to_string(sid_str, &token->user_sids[i])));
	}
}

static void nt_token_free(NT_USER_TOKEN *token)
{
	if (!token)
		return;
	safe_free(token->user_sids);
	free(token);
}

const NT_USER_TOKEN *get_current_nt_token(void)
{
	static uint16 cached_vuid = UID_FIELD_INVALID;
	static NT_USER_TOKEN *cached_token = NULL;
	vuser_struct *vuser;
	BOOL guest = True;

	if (cached_vuid == get_sec_ctx())
		return cached_token;

	cached_vuid = get_sec_ctx();
	vuser = get_valid_vuser_struct(cached_vuid);

	nt_token_free(cached_token);
	if (vuser->usr)
		guest = vuser->usr->guest;
	cached_token = nt_token_new(vuser->ntusr, guest);

	if (vuser->usr)
	{
		DEBUG(2, ("get_current_nt_token: VUID=0x%x `%s' guest:%d uid:%d gid:%d\n",
			  cached_vuid, vuser->usr->requested_name, vuser->usr->guest,
			  vuser->usr->uid, vuser->usr->gid));
	}
	else
	{
		DEBUG(2, ("get_current_nt_token: VUID=0x%x\n",
		  cached_vuid));
	}

	vuid_free_vuser_struct(vuser);

	nt_token_show(cached_token);

	return cached_token;
}
