/* 
   Unix SMB/Netbios implementation.
   Version 1.9.
   Password and authentication handling
   Copyright (C) Luke Kenneth Casson Leighton 1996-2001,
   Copyright (C) Andrew Tridgell 1992-1998
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#include "includes.h"
#ifdef HAVE_PWD_H
#include <pwd.h>
#endif
#include "libsamba.h"
#include "rpc_parse.h"
#include "sids.h"
#include "others.h"
#include "tng_misc.h"

extern BOOL in_client;

/* this holds info on user ids that are already validated for this VC */
static int num_validated_users = 0;


static void show_vuser(int level, const vuser_struct *vuser)
{
	if (vuser->usr != NULL)
	{
		DEBUGADD(level, ("vuid 0x%x uid %d registered to unix name %s\n",
				 vuser->vuid, (int)vuser->usr->uid, vuser->usr->unix_name));
	}
	if (vuser->ntusr != NULL)
	{
	dump_data_pw("vuid usr sess key:\n", vuser->ntusr->user_sess_key,
		     sizeof(vuser->ntusr->user_sess_key));
	}
}


uint32 vuid_get_flags(uint16 vuid)
{
	vuser_struct *vuser = get_valid_vuser_struct(vuid);
	uint32 flags;

	if (! vuser)
	{
		DEBUG(1, ("vuid_get_flags: Invalid vuid:0x%x\n\n", vuid));
		return VUID_NONE;
	}

	flags = vuser->flags;
	vuid_free_vuser_struct(vuser);

	return flags;
}

/****************************************************************************
check if a uid has been validated, and return an pointer to the user_struct
if it has. NULL if not. vuid is biased by an offset. This allows us to
tell random client vuid's (normally zero) from valid vuids.
****************************************************************************/
vuser_struct *get_valid_vuser_struct(uint16 vuid)
{
	vuser_struct *usr = NULL;

	if (vuid == UID_FIELD_INVALID)
	{
		DEBUG(2, ("get_valid_vuser_struct: UID_FIELD_INVALID\n"));
		return NULL;
	}
	if (!tdb_lookup_vuid(vuid, &usr))
	{
		DEBUG(2, ("Warning: get_valid_vuser_struct got invalid "
			  "VUID 0x%x\n", vuid));
		return NULL;
	}
	if (usr != NULL)
	{
		if (((int)usr->usr->uid) == -1 || ((int)usr->usr->gid) == -1)
		{
			vuid_free_vuser_struct(usr);
			return NULL;
		}
	}
	return usr;
}

vuser_struct *get_public_user_struct(uint16 vuid)
{
	vuser_struct *vuser = get_valid_vuser_struct(vuid);

	if (! vuser)
		return NULL;

	if (vuser->flags & VUID_PUBLIC)
		return vuser;

	DEBUG(0, ("ATTACK?! Attempt to use a non public vuid in a place "
		  "that expects one\n"));
	show_vuser(0, vuser);

	vuid_free_vuser_struct(vuser);
	return NULL;
}

/****************************************************************************
invalidate a uid
****************************************************************************/
void invalidate_vuid(uint16 vuid)
{
	tdb_delete_vuid(vuid);
}


/****************************************************************************
return a validated username
****************************************************************************/
BOOL validated_username(uint16 vuid, char *name, size_t len)
{
	vuser_struct *vuser = get_public_user_struct(vuid);
	if (vuser == NULL)
	{
		return False;
	}
	if (vuser->usr == NULL)
	{
		vuid_free_vuser_struct(vuser);
		return False;
	}
	safe_strcpy(name, vuser->usr->unix_name, len);
	vuid_free_vuser_struct(vuser);
	return True;
}


/****************************************************************************
register a uid/name pair as being valid and that a valid password
has been given. vuid is biased by an offset. This allows us to
tell random client vuid's (normally zero) from valid vuids.
****************************************************************************/
uint16 make_vuid(vuser_struct *vuser)
{
	uint16 vuid;

	num_validated_users++;
	vuid = (uint16) ((num_validated_users - 1) + VUID_OFFSET);

	if (vuid < VUID_OFFSET)
	{
		DEBUG(0, ("WARNING: make_vuid: Someone is overflowing our vuid table\n"));
		return UID_FIELD_INVALID;
	}

	vuser->vuid = vuid;

	show_vuser(3, vuser);

	if (!tdb_store_vuid(vuid, vuser, False))
	{
		return UID_FIELD_INVALID;
	}

	return vuid;
}

/****************************************************************************
register a uid/name pair as being valid and that a valid password
has been given. vuid is biased by an offset. This allows us to
tell random client vuid's (normally zero) from valid vuids.
****************************************************************************/
static uint16 create_vuid(
		   uid_t uid, gid_t gid,
		   int n_groups, gid_t * groups,
		   const char *unix_name,
		   const char *requested_name,
		   const char *real_name,
		   BOOL guest, uint32 flags, const NET_USER_INFO_3 * info3)
{
	vuser_struct vuser;
	user_struct usr;

	usr.uid = uid;
	usr.gid = gid;
	usr.guest = guest;
	fstrcpy(usr.unix_name, unix_name);
	fstrcpy(usr.requested_name, requested_name);
	fstrcpy(usr.real_name, real_name);

	usr.n_groups = n_groups;
	usr.groups = groups;


	/* inconvenient but necessary to have a pointer to the
	 * NET_USER_INFO_3 stuff - gets stored in a tdb anyway
	 * and isn't modified.
	 */
	vuser.ntusr = STRIP_CONST(info3);
	vuser.flags = flags;
	vuser.usr = &usr;

	return make_vuid(&vuser);
}

/****************************************************************************
register a uid/name pair as being valid and that a valid password
has been given. vuid is biased by an offset. This allows us to
tell random client vuid's (normally zero) from valid vuids.
****************************************************************************/
uint16 register_vuid(uid_t uid, gid_t gid,
		     const char *unix_name,
		     const char *requested_name,
		     BOOL guest, uint32 flags, const NET_USER_INFO_3 * info3)
{
	int n_groups = 0;
	gid_t *groups = NULL;
	fstring real_name;
	uint16 vuid;

	if (!in_client)
	{
		if (lp_security() == SEC_SHARE)
			return UID_FIELD_INVALID;

#ifdef USE_RENEWABLE_AFS_TICKET
		/* The new pag needs to be initialized before the forking */
		/* This should only be done if we use encrypted passwords */
		if(lp_encrypted_passwords() && uid != 0)
			if (k_hasafs()) {
				DEBUG(3, ("Setting new pag for uid %d\n", (int)uid));
				k_setpag();
			}
#endif /* USE_RENEWABLE_AFS_TICKET */
		/* Find all the groups this uid is in and store them. 
		   Used by become_user() */
		get_unixgroups(unix_name, uid, gid, &n_groups, &groups);
	}

	/* Clearing default real name */
	fstrcpy(real_name, "<Full Name>");
#if defined(HAVE_GETPWNAM) || defined(HAVE_PWD_H)
	if (lp_unix_realname())
	{
		struct passwd *pwfile;
		if ((pwfile = getpwnam(unix_name)) != NULL)
		{
			DEBUG(3,
			      ("User name: %s\tReal name: %s\n", unix_name,
			       pwfile->pw_gecos));
			fstrcpy(real_name, pwfile->pw_gecos);
		}
	}
#endif

	DEBUG(3, ("register_vuid(uid=%d, gid=%d, %d groups, unix_name=%s, "
		  "requested_name=%s, guest=%s, ...)\n",
		  uid, gid, n_groups,
		  unix_name, requested_name, BOOLSTR(guest)));

	vuid = create_vuid(uid, gid, n_groups, groups,
			   unix_name, requested_name,
			   real_name, guest, flags, info3);

	safe_free(groups);
	return vuid;
}

/*******************************************************************
check if a username is OK
********************************************************************/
BOOL check_vuser_ok(struct uid_cache *cache, user_struct * vuser, int snum)
{
	int i;
	for (i = 0; i < cache->entries; i++)
		if (cache->list[i] == vuser->uid)
			return (True);

	if (!user_ok(vuser->unix_name, snum))
		return (False);

	i = cache->entries % UID_CACHE_SIZE;
	cache->list[i] = vuser->uid;

	if (cache->entries < UID_CACHE_SIZE)
		cache->entries++;

	return (True);
}

/****************************************************************************
  get user session key for a vuid
****************************************************************************/
BOOL vuid_get_usr_sesskey(uint16 vuid, uchar usr_sess_key[16])
{
	vuser_struct *vuser;

	if (vuid == UID_FIELD_INVALID)
	{
		memset(usr_sess_key, 0, 16);
		return True;
	}
	vuser = get_valid_vuser_struct(vuid);
	if (vuser == NULL)
	{
		DEBUG(2, ("vuid_get_usr_sesskey: no vuser struct for 0x%x\n",
			  vuid));
		return False;
	}
	if (vuser->ntusr == NULL)
	{
		memset(usr_sess_key, 0, 16);
		vuid_free_vuser_struct(vuser);
		return True;
	}

	memcpy(usr_sess_key, vuser->ntusr->user_sess_key, 16);
	vuid_free_vuser_struct(vuser);
	return True;
}


uint16 vuid_create_system(void)
{
	DOM_SID sid;
	NET_USER_INFO_3 usr;
	uint16 system_vuid;

	ZERO_STRUCT(usr);

	sid_copy(&sid, global_sid_system);
	sid_split_rid(&sid, &usr.user_id);
	usr.group_id = usr.user_id;
	make_dom_sid2(&usr.dom_sid, &sid);

	sid_copy(&sid, global_sid_builtin);
	sid_append_rid(&sid, BUILTIN_ALIAS_RID_ADMINS);
	/* XXX - This will give us problems with dynamic other_sids */
	make_dom_sid2(&usr.other_sids[0], &sid);
	usr.num_other_sids = 1;

	system_vuid = create_vuid(0, 0,
				  0, NULL,
				  "root", "SYSTEM", "The SYSTEM user",
				  False, 0,
				  &usr);

	return system_vuid;
}
