/* 
   Unix SMB/Netbios implementation.
   Version 1.9.
   SMB client generic functions
   Copyright (C) Andrew Tridgell 1994-1999
   Copyright (C) Luke Kenneth Casson Leighton 1996-2001
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/
/* MKPROTO: lib_smb_proto.h		(do not remove this line) */

#include "includes.h"
#include <errno.h>
#include "libsamba.h"
#include "dlinklist.h"
#include "byteorder.h"
#include "libsmb.h"
#include "nterr.h"
#include "tng_misc.h"


typedef struct _SmbPendingReq
{
	struct _SmbPendingReq *next, *prev;
	uint32 pid;
	uint16 mid;
	BOOL keep_pend_item;
	SmbSignReqInfo sign_req_info;
	NTSTATUS (* callback) (struct _SmbPendingReq *pi,
			       void *response_packet,
			       void *cb_data);
	void *cb_data;
} SmbPendingReq;
#ifndef NOONS_SMALL_BUILD
static SmbPendingReq *cli_pend_find(struct cli_state *cli, char *packet)
{
	SmbPendingReq *p;
	uint32 pid = SVAL(packet, smb_pid);
	uint16 mid = SVAL(packet, smb_mid);

	if (CVAL(packet, 0) != 0)
		return NULL;

	for (p = cli->pending_req_list; p; p = p->next)
	{
		if (pid == p->pid && mid == p->mid)
			break;
	}

	return p;
}
#endif

#ifndef NOONS_SMALL_BUILD
static void cli_add_pending(struct cli_state *cli, char *packet,
			    const SmbSignReqInfo *sign_req_info)
{
	SmbPendingReq *p;

	if (CVAL(packet, 0) != 0)
		return;

	p = cli_pend_find(cli, packet);

	if (p)
	{
		DEBUG(1, ("cli_add_pending: Have that one already...\n"));
		return;
	}

	p = g_new0(SmbPendingReq, 1);
	p->pid = SVAL(packet, smb_pid);
	p->mid = SVAL(packet, smb_mid);
	p->sign_req_info = *sign_req_info;

	DEBUG(2, ("SMB Pending Stack: Adding %d:%d\n", p->pid, p->mid));

	DLIST_ADD(cli->pending_req_list, p);
}
#endif
#ifndef NOONS_SMALL_BUILD
static void cli_pending_remove(struct cli_state *cli, char *packet)
{
	SmbPendingReq *p;

	if (CVAL(packet, 0) != 0)
		return;

	p = cli_pend_find(cli, packet);

	if (! p)
	{
		DEBUG(1, ("Unexpected packet: %d:%d\n",
			  SVAL(packet, smb_pid), SVAL(packet, smb_mid)));
		return;
	}

	DEBUG(2, ("SMB Pending Stack: Removing %d:%d\n", p->pid, p->mid));

	DLIST_REMOVE(cli->pending_req_list, p);

	safe_free(p);
}
#endif
/*
 * Change the port number used to call on 
 */
int cli_set_port(struct cli_state *cli, int port)
{
	if (port > 0)
	  cli->port = port;

	return cli->port;
}

/****************************************************************************
  read an smb from a fd ignoring all keepalive packets. Note that the buffer 
  *MUST* be of size BUFFER_SIZE+SAFETY_MARGIN.
  The timeout is in milliseconds

  This is exactly the same as receive_smb except that it never returns
  a session keepalive packet (just as receive_smb used to do).
  receive_smb was changed to return keepalives as the oplock processing means this call
  should never go into a blocking read.
****************************************************************************/

static BOOL client_receive_smb(Socket fd, char *buffer, unsigned int timeout)
{
	BOOL ret;

	for (;;)
	{
		ret = receive_smb(fd, buffer, timeout);

		if (!ret)
		{
			DEBUG(10, ("client_receive_smb failed\n"));
			show_msg(buffer);
			return ret;
		}

		/* Ignore session keepalive packets. */
		if (CVAL(buffer, 0) != SMBkeepalive)
			break;
	}
	show_msg(buffer);
	return ret;
}


/****************************************************************************
recv an smb
****************************************************************************/
BOOL cli_receive_smb(struct cli_state *cli)
{
	BOOL ret;
 again:
	/* there might have been an error on the socket */
	if (cli->fd == -1) return False;

	ret = client_receive_smb(cli->fd,cli->inbuf,cli->timeout);
	
	if (ret) {
		/* it might be an oplock break request */
		if (!(CVAL(cli->inbuf, smb_flg) & FLAG_REPLY) &&
		    CVAL(cli->inbuf,smb_com) == SMBlockingX &&
		    SVAL(cli->inbuf,smb_vwv6) == 0 &&
		    SVAL(cli->inbuf,smb_vwv7) == 0) {
			if (cli->use_oplocks) {
				int fnum = SVAL(cli->inbuf,smb_vwv2);
				unsigned char level = CVAL(cli->inbuf,smb_vwv3+1);
				if (!cli_oplock_ack(cli, fnum, level))
					return False;
			}
			/* try to prevent loops */
			CVAL(cli->inbuf,smb_com) = 0xFF;
			goto again;
		}
	}

	/* if the server is not responding, then note that now */
	if (!ret) {
		tng_socket_close(cli->fd);
		cli->fd = -1;
	}

	return ret;
}

/****************************************************************************
  send an smb to a fd and re-establish if necessary
****************************************************************************/

BOOL cli_send_smb(struct cli_state *cli)
{
	size_t len;
	size_t nwritten=0;
	ssize_t ret;
	BOOL reestablished=False;

	/* there might have been an error on the socket */
	if (cli->fd == -1) return False;

	len = smb_len(cli->outbuf) + 4;

	show_msg(cli->outbuf);

	while (nwritten < len) {
		ret = write_socket(cli->fd,cli->outbuf+nwritten,len - nwritten);
		if (ret <= 0 && errno == EPIPE && !reestablished) {
			if (cli_reestablish_connection(cli)) {
				reestablished = True;
				nwritten=0;
				continue;
			}
		}
		if (ret <= 0) {
			tng_socket_close(cli->fd);
			cli->fd = -1;
			DEBUG(0,("Error writing %d bytes to client. %d\n",
				 (int)len,(int)ret));
			return False;
		}
		nwritten += ret;
	}
	
	return True;
}

/****************************************************************************
setup basics in a outgoing packet
****************************************************************************/
void cli_setup_packet(struct cli_state *cli)
{
        cli->rap_error = 0;
	SSVAL(cli->outbuf,smb_pid,cli->pid);
	SSVAL(cli->outbuf,smb_uid,cli->vuid);
	SSVAL(cli->outbuf,smb_mid,cli->mid);

	if (cli->protocol > PROTOCOL_CORE)
	{
		uint16 flags2;
		SCVAL(cli->outbuf,smb_flg,0x8);
		flags2 = FLAGS2_LONG_PATH_COMPONENTS;
		if (cli->capabilities & CAP_UNICODE) {
			flags2 |= FLAGS2_UNICODE_STRINGS;
		}
		if (cli->capabilities & CAP_STATUS32) {
			flags2 |= FLAGS2_32_BIT_ERROR_CODES;
		}

		if (lp_smb_ntlmssp())
			flags2 |= FLAGS2_EXTENDED_SECURITY;

		SSVAL(cli->outbuf,smb_flg2,flags2);
	}
}

/****************************************************************************
setup the bcc length of the packet from a pointer to the end of the data
****************************************************************************/
void cli_setup_bcc(struct cli_state *cli, void *p)
{
	set_message_bcc(cli->outbuf, PTR_DIFF(p, smb_buf(cli->outbuf)));
}



/****************************************************************************
initialise a client structure
****************************************************************************/
void cli_init_creds(struct cli_state *cli, const struct ntuser_creds *usr)
{
	copy_nt_creds(&cli->usr, usr);

	cli->nt.ntlmssp_cli_flgs = usr != NULL ? usr->ntlmssp_flags : 0;

	DEBUG(10,("cli_init_creds: ntlmssp_flgs: %x user: %s domain: %s\n", 
	           cli->nt.ntlmssp_cli_flgs,
		   cli->usr.user_name, cli->usr.domain));
}


/****************************************************************************
initialise a client structure
****************************************************************************/
struct cli_state *cli_initialise(struct cli_state *cli)
{
        BOOL alloced_cli = False;

	if (!cli) {
		cli = g_new1(struct cli_state);
		if (!cli)
			return NULL;
                alloced_cli = True;
	}
	else if (cli->initialised)
		cli_shutdown(cli);

	ZERO_STRUCTPN(cli);

	cli->port = 0;
	cli->fd = -1;
	cli->cnum = -1;
	cli->pid = (uint16)sys_getpid();
	cli->mid = 1;
	cli->vuid = UID_FIELD_INVALID;
	cli->protocol = PROTOCOL_NT1;
	cli->timeout = 20000; /* Timeout is in milliseconds. */
	cli->bufsize = CLI_BUFFER_SIZE+4;
	cli->max_xmit = cli->bufsize;
	cli->outbuf = g_new(char, cli->bufsize);
	cli->inbuf = g_new(char, cli->bufsize);

	if (!cli->outbuf || !cli->inbuf)
	{
		goto error;
	}

	cli->use_ntlmv2 = lp_client_ntlmv2();

	cli_init_creds(cli, NULL);

	cli->initialised = 1;

	return cli;

	/* Clean up after malloc() error */
    error:

	SAFE_FREE(cli->inbuf);
	SAFE_FREE(cli->outbuf);

	if (alloced_cli)
		SAFE_FREE(cli);

	return NULL;
}

/****************************************************************************
close the socket descriptor
****************************************************************************/
void cli_close_socket(struct cli_state *cli)
{
	if (cli->fd != SOCKET_NULL)
	{
		tng_socket_close(cli->fd);
	}
	cli->fd = SOCKET_NULL;
}

/****************************************************************************
shutdown a client structure
****************************************************************************/
void cli_shutdown(struct cli_state *cli)
{
	DEBUG(10,("cli_shutdown\n"));
	SAFE_FREE(cli->outbuf);
	SAFE_FREE(cli->inbuf);
	cli_close_socket(cli);
	ZERO_OBJPN(cli);
}


/****************************************************************************
set socket options on a open connection
****************************************************************************/
void cli_sockopt(struct cli_state *cli, char *options)
{
	set_socket_options(cli->fd, options);
}

/****************************************************************************
set the PID to use for smb messages. Return the old pid.
****************************************************************************/
uint16 cli_setpid(struct cli_state *cli, uint16 pid)
{
	uint16 ret = cli->pid;
	cli->pid = pid;
	return ret;
}
