/* 
   Unix SMB/Netbios implementation.
   Version 1.9.
   Samba utility functions
   Copyright (C) Andrew Tridgell 1992-2000
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/
/* MKPROTO: lib_smb_proto.h		(do not remove this line) */

#include "includes.h"
#include "byteorder.h"
#include "libsamba.h"
#include "libsmb.h"

#undef DBGC_CLASS
#define DBGC_CLASS DBGC_SMB

extern int smb_read_error;


/*******************************************************************
  show a smb message structure
********************************************************************/
void show_msg(char *buf)
{
	int i;
	int bcc = 0;

	if (! DEBUGLVL(5))
		return;

	DEBUGADD(5,
	      ("size=%d\nsmb_com=0x%x\nsmb_rcls=%d\nsmb_reh=%d\nsmb_err=%d\nsmb_flg=%d\nsmb_flg2=%d\n",
	       smb_len(buf),
	       (int)CVAL(buf, smb_com), (int)CVAL(buf, smb_rcls),
	       (int)CVAL(buf, smb_reh), (int)SVAL(buf, smb_err),
	       (int)CVAL(buf, smb_flg), (int)SVAL(buf, smb_flg2)));
	DEBUGADD(5,
	      ("smb_tid=%d\nsmb_pid=%d\nsmb_uid=%d\nsmb_mid=%d\nsmt_wct=%d\n",
	       (int)SVAL(buf, smb_tid), (int)SVAL(buf, smb_pid),
	       (int)SVAL(buf, smb_uid), (int)SVAL(buf, smb_mid),
	       (int)CVAL(buf, smb_wct)));

	for (i = 0; i < (int)CVAL(buf, smb_wct); i++)
	{
		DEBUGADD(5, ("smb_vwv[%d]=%d (0x%X)\n", i,
			  SVAL(buf, smb_vwv + 2 * i), SVAL(buf,
							   smb_vwv + 2 * i)));
	}

	bcc = (int)SVAL(buf, smb_vwv + 2 * (CVAL(buf, smb_wct)));

	DEBUGADD(5, ("smb_bcc=%d\n", bcc));

	if (bcc == 0)
		return;

	if (! DEBUGLVL(10))
		return;

	if (! DEBUGLVL(50))
	{
		bcc = MIN(bcc, 512);
	}

	dump_data(10, smb_buf(buf), bcc);
}


/*******************************************************************
  setup only the byte count for a smb message
********************************************************************/
int set_message_bcc(char *buf,int num_bytes)
{
	int num_words = CVAL(buf,smb_wct);
	SSVAL(buf,smb_vwv + num_words*SIZEOFWORD,num_bytes);  
	smb_setlen(buf,smb_size + num_words*2 + num_bytes - 4);
	return (smb_size + num_words*2 + num_bytes);
}

/*******************************************************************
  setup only the byte count for a smb message, using the end of the
  message as a marker
********************************************************************/
int set_message_end(void *outbuf,void *end_ptr)
{
	return set_message_bcc((char *)outbuf,PTR_DIFF(end_ptr,smb_buf((char *)outbuf)));
}


/****************************************************************************
  send an smb to a fd 
****************************************************************************/

BOOL send_smb(Socket fd, char *buffer)
{
	size_t len;
	size_t nwritten = 0;
	ssize_t ret;
	len = smb_len(buffer) + 4;

	while (nwritten < len)
	{
		ret = write_socket(fd, buffer + nwritten, len - nwritten);
		if (ret <= 0)
		{
			DEBUG(0,
			      ("Error writing %d bytes to client. %d. Exiting\n",
			       (int)len, (int)ret));
			exit(1);
		}
		nwritten += ret;
	}

	return True;
}

/****************************************************************************
read 4 bytes of a smb packet and return the smb length of the packet
store the result in the buffer
This version of the function will return a length of zero on receiving
a keepalive packet.
timeout is in milliseconds.
****************************************************************************/

static ssize_t read_smb_length_return_keepalive(Socket fd, char *inbuf,
						unsigned int timeout)
{
	ssize_t len = 0;
	int msg_type;
	BOOL ok = False;

	while (!ok)
	{
		ok = (read_socket_with_timeout(fd, inbuf, 4, 4, timeout) == 4);

		if (!ok)
			return (-1);

		len = smb_len(inbuf);
		msg_type = CVAL(inbuf, 0);

		if (msg_type == 0x85)
			DEBUG(5, ("Got keepalive packet\n"));
	}

	DEBUG(10, ("got smb length of %d\n", len));

	return (len);
}

/****************************************************************************
read 4 bytes of a smb packet and return the smb length of the packet
store the result in the buffer. This version of the function will
never return a session keepalive (length of zero).
timeout is in milliseconds.
****************************************************************************/
ssize_t read_smb_length(Socket fd, char *inbuf, unsigned int timeout)
{
	ssize_t len;

	for (;;)
	{
		len = read_smb_length_return_keepalive(fd, inbuf, timeout);

		if (len < 0)
			return len;

		/* Ignore session keepalives. */
		if (CVAL(inbuf, 0) != 0x85)
			break;
	}

	DEBUG(10, ("read_smb_length: got smb length of %d\n", len));

	return len;
}

/****************************************************************************
  read an smb from a fd. Note that the buffer *MUST* be of size
  BUFFER_SIZE+SAFETY_MARGIN.
  The timeout is in milliseconds. 
  This function will return on a
  receipt of a session keepalive packet.
****************************************************************************/
BOOL receive_smb(Socket fd, char *buffer, unsigned int timeout)
{
	ssize_t len, ret;

	smb_read_error = 0;

	memset(buffer, '\0', smb_size + 100);

	len = read_smb_length_return_keepalive(fd, buffer, timeout);
	if (len < 0)
	{
		DEBUG(10, ("receive_smb: length < 0!\n"));
		return (False);
	}

	if (len > BUFFER_SIZE)
	{
		DEBUG(0, ("Invalid packet length! (%d bytes).\n", len));
		if (len > BUFFER_SIZE + (SAFETY_MARGIN / 2))
		{
			exit(1);
		}
	}

	if (len > 0)
	{
		ret = read_socket_data(fd, buffer + 4, len);
		if (ret != len)
		{
			smb_read_error = READ_ERROR;
			return False;
		}
	}

	return (True);
}


/*
 * XXX - make_dir_struct: memcpy -> push_ascii
 */
void make_dir_struct(char *buf, const char *mask, const char *fname,
		     SMB_OFF_T size, int mode, time_t date, BOOL case_sensitive)
{
	char *p;
	pstring mask2;

	pstrcpy(mask2, mask);

	if ((mode & aDIR) != 0)
		size = 0;

	memset(buf + 1, ' ', 11);
	if ((p = strchr(mask2, '.')) != NULL)
	{
		*p = 0;
		memcpy(buf + 1, mask2, MIN(strlen(mask2), 8));
		memcpy(buf + 9, p + 1, MIN(strlen(p + 1), 3));
		*p = '.';
	}
	else
		memcpy(buf + 1, mask2, MIN(strlen(mask2), 11));

	memset(buf + 21, '\0', DIR_STRUCT_SIZE - 21);
	CVAL(buf, 21) = mode;
	put_dos_date(buf, 22, date);
	SSVAL(buf, 26, size & 0xFFFF);
	SSVAL(buf, 28, (size >> 16) & 0xFFFF);
	StrnCpy(buf + 30, fname, 12);
	if (!case_sensitive)
		strupper(buf + 30);
	DEBUG(8, ("put name [%s] into dir struct\n", buf + 30));
}


/****************************************************************************
send a keepalive packet (rfc1002)
****************************************************************************/

BOOL send_keepalive(Socket client)
{
	unsigned char buf[4];

	buf[0] = 0x85;
	buf[1] = buf[2] = buf[3] = 0;

	return (write_socket_data(client, (char *)buf, 4) == 4);
}


BOOL cifs_ptr_in_smbbuf(void *inbuf, const void *src)
{
	if ((const char *) src < smb_buf(inbuf)
	    || (const char *) src >= (smb_buf(inbuf) + smb_buflen(inbuf)))
	{
		DEBUG(0, ("cifs_ptr_in_smbbuf: src outside buffer!\n"));
		show_msg(inbuf);
		return False;
	}

	return True;
}


void cifs_put_dos_date(void *dest, const NTTIME *nt)
{
	put_dos_date(dest, 0, nt_time_to_unix(nt));
}
void cifs_put_dos_date2(void *dest, const NTTIME *nt)
{
	put_dos_date2(dest, 0, nt_time_to_unix(nt));
}
void cifs_put_dos_date3(void *dest, const NTTIME *nt)
{
	put_dos_date3(dest, 0, nt_time_to_unix(nt));
}

void cifs_put_nttime(void *dest, const NTTIME *nt)
{
	SIVAL(dest, 0, nt->low);
	SIVAL(dest, 4, nt->high);
}


/**
 * \brief Do string alignment for SMB
 * \note align_string() is a simpler API, that probably should die.
 */
size_t cifs_align_string(const void *packet, const void *base_ptr, const char *p, int flags)
{
	if (!(flags & STR_ASCII)
	    && ((flags & STR_UNICODE)
		|| (SVAL(packet, smb_flg2) & FLAGS2_UNICODE_STRINGS)))
	{
		return ucs2_align(base_ptr, p, flags);
	}
	return 0;
}

#ifndef NOONS_SMALL_BUILD
static
size_t cifsw_pull_ucs2(const void *packet_ptr,
	UNICODE_STRING *dest,
	const void *src, int src_len,
	int flags)
{
	size_t ret, moved;
	void *buf;

	moved = ucs2_align(packet_ptr, src, flags);
	src = (const void *)((const char *)src + moved);

	tng_read_len_fixup_ucs2(src, moved, &src_len, flags);

	if (! src_len)
		return 0;

	ret = convert_string_allocate(CH_UCS2, CH_UINT16,
				      src, src_len,
				      &buf);
	dest->Buffer = buf;
	dest->Length = ret;
	dest->MaximumLength = ret;

	return src_len + moved;
}
#endif
#ifndef NOONS_SMALL_BUILD
static
size_t cifsw_pull_ascii(UNICODE_STRING *dest,
	const void *src, int src_len,
	int flags)
{
	size_t ret;
	void *buf;

	tng_read_len_fixup_ascii(src, &src_len, flags);

	if (src_len == 0)
		return 0;

	ret = convert_string_allocate(CH_DOS, CH_UINT16,
				      src, src_len,
				      &buf);
	dest->Buffer = buf;
	dest->Length = ret;
	dest->MaximumLength = ret;

	return src_len;
}
#endif
#ifndef NOONS_SMALL_BUILD
static
size_t cifsw_pull_string(const void *packet_ptr,
	UNICODE_STRING *dest,
	const void *src, int src_len,
	int flags)
{
	ZERO_STRUCTP(dest);

	if (!(flags & STR_ASCII) && \
	    ((flags & STR_UNICODE || \
	      (SVAL(packet_ptr, smb_flg2) & FLAGS2_UNICODE_STRINGS))))
	{
		return cifsw_pull_ucs2(packet_ptr, dest, src, src_len, flags);
	}
	return cifsw_pull_ascii(dest, src, src_len, flags);
}
#endif
