/* 
   Unix SMB/Netbios implementation.
   Version 1.9.
   process incoming packets - main loop
   Copyright (C) Luke Kenneth Casson Leighton 1996-2001
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#include "includes.h"
#include <errno.h>
#include "libsamba.h"
#include "byteorder.h"
#include "tng_misc.h"
#include "rpc_parse.h"
#include "rpc_client.h"
#include "others.h"


extern int smb_read_error;
extern BOOL reload_after_sighup;


/****************************************************************************
 select on a timeout on the dce/rpc socket.
 ****************************************************************************/

static BOOL receive_message_or_msrpc(int c, prs_struct *ps,
				     int timeout, BOOL *got_msrpc)
{
	int selrtn;

	DEBUG(10, ("receive_message_or_msrpc: timeout %d fd %d\n",
		   timeout, c));

	smb_read_error = 0;

	*got_msrpc = False;

	/*
	 * Check to see if we already have a message on the smb queue.
	 * If so - copy and return it.
	 */

	selrtn = read_data_outstanding(c, timeout);

	/* Check if error */
	if (selrtn == -1)
	{
		/* something is wrong. Maybe the socket is dead? */
		smb_read_error = READ_ERROR;
		DEBUG(2, ("read error on loop-back socket\n"));
		return False;
	}

	/* Did we timeout ? */
	if (selrtn == 0)
	{
		smb_read_error = READ_TIMEOUT;
		DEBUG(5, ("timeout on loop-back socket\n"));
		return False;
	}

	*got_msrpc = True;
	return receive_msrpc(c, ps, 0);
}


/*
These flags determine some of the permissions required to do an operation 

Note that I don't set NEED_WRITE on some write operations because they
are used by some brain-dead clients when printing, and I don't want to
force write permissions on print services.
*/
#define AS_USER (1<<0)
#define NEED_WRITE (1<<1)
#define TIME_INIT (1<<2)
#define CAN_IPC (1<<3)
#define AS_GUEST (1<<5)
#define QUEUE_IN_OPLOCK (1<<6)

/* 
   define a list of possible SMB messages and their corresponding
   functions. Any message that has a NULL function is unimplemented -
   please feel free to contribute implementations!
*/


/****************************************************************************
  process an smb from the client - split out from the process() code so
  it can be used by the oplock break code.
****************************************************************************/
static void process_msrpc(rpcsrv_struct * l, const char *name,
			  prs_struct *pdu)
{
	int32 len = prs_buf_len(pdu);
	BOOL more;
	DEBUG(6, ("got message of len 0x%x\n", len));

	dump_data(10, pdu->data, len);

	more = rpc_local(l, pdu->data, len, name);
	while (more)
	{
		more = msrpc_send(l->c, &l->rsmb_pdu);
		prs_free_data(&l->rsmb_pdu);
		if (more)
		{
			more = rpc_local(l, NULL, 0, name);
		}
	}
}


static void free_srv_auth_fns_array(uint32 num_entries,
				    srv_auth_fns ** entries)
{
	free_void_array(num_entries, (void **)entries, NULL);
}

static srv_auth_fns *add_srv_auth_fns_to_array(uint32 *len,
					       srv_auth_fns *** array,
					       srv_auth_fns * name)
{
	return (srv_auth_fns *) add_item_to_array(len,
						  (void ***)array,
						  (void *)name);
}

void close_srv_auth_array(rpcsrv_struct * l)
{
	if (! l)
		return;
	free_srv_auth_fns_array(l->num_auths, l->auth_fns);
}

void add_srv_auth_fn(rpcsrv_struct * l, srv_auth_fns * fn)
{
	add_srv_auth_fns_to_array(&l->num_auths, &l->auth_fns, fn);
	DEBUG(10, ("add_srv_auth_fn: %d\n", l->num_auths));
}

/****************************************************************************
  initialise from pipe
****************************************************************************/
BOOL msrpcd_init(int c, rpcsrv_struct ** l)
{
	vuser_struct *vu = NULL;
	user_struct *vuser = NULL;
	uint16 vuid;

	(*l) = NULL;

	if (!ncalrpc_srv_get_user_creds(c, &vuid))
	{
		DEBUG(0, ("authentication failed\n"));
		return False;
	}

	DEBUG(100, ("initial vuid: %x\n", vuid));

	if (vuid == UID_FIELD_INVALID)
	{
		DEBUG(0, ("msrpcd_init: initial vuid invalid\n"));
		return False;
	}

	if (!become_vuser(vuid))
	{
		return False;
	}

	vu = get_valid_vuser_struct(vuid);
	if (vu == NULL)
	{
		return False;
	}

	(*l) = g_new0(rpcsrv_struct, 1);
	if ((*l) == NULL)
	{
		return False;
	}

	(*l)->initial_pipe_vuid = vuid;
	(*l)->vuid = vuid;
	(*l)->c = c;

	vuser = vu->usr;

	if (vuser != NULL && !vuser->guest)
	{
		char *user = vuser->unix_name;
		if (!strequal(user, lp_guestaccount(-1)) &&
		    lp_servicenumber(user) < 0)
		{
			int homes = lp_servicenumber(HOMES_NAME);
			char *home = get_user_home_dir(user);
			if (homes >= 0 && home)
			{
				pstring home_dir;
				pstrcpy(home_dir, home);
				lp_add_home(user, homes, home_dir);
			}
		}
	}

	vuid_free_vuser_struct(vu);

	return True;
}

/****************************************************************************
  process commands from the client
****************************************************************************/
void msrpcd_process(msrpc_service_fns * fn, rpcsrv_struct * l,
		    const char *name)
{
	int counter = 0;
	extern fstring local_machine;

	/* re-initialise the timezone */
	TimeInit();

	set_remote_machine (name);
	fstrcpy(local_machine, global_myname);
	local_machine[15] = 0;
	strlower(local_machine);

	DEBUG(2, ("msrpc_process: client_name: %s my_name: %s\n",
		  name, local_machine));

	become_root();
	fn->reload_services(True);
	reopen_logs();
	unbecome_root();

	while (True)
	{
		BOOL got_msrpc = False;
		prs_struct pdu;

		errno = 0;

		/* free up temporary memory */
		lp_talloc_free();

		counter = (counter + 1) % 200;

		if (!receive_message_or_msrpc(l->c, &pdu,
					      SMBD_SELECT_TIMEOUT * 1000,
					      &got_msrpc))
		{
			if (smb_read_error == READ_TIMEOUT)
			{
				DEBUG(5, ("msrpcd_process: ignoring timeout\n"));
				continue;
			}
			if (smb_read_error == READ_EOF)
			{
				DEBUG(3, ("end of file from client\n"));
			}

			if (smb_read_error == READ_ERROR)
			{
				DEBUG(3, ("receive error (%s) exiting\n",
					  strerror(errno)));
			}

			unbecome_to_initial_uid();

			if (fn->idle != NULL)
			{
				DEBUG(5, ("msrpcd_process: Calling idle fn\n"));
				become_root();
				fn->idle();
				unbecome_root();
			}

			return;
		}

#if 0
		if (counter == 0)
		{
			/* reload services, if files have changed. */
			fn->reload_services(True);
		}
#endif
		/*
		 * If reload_after_sighup == True then we got a SIGHUP
		 * and are being asked to reload. Fix from <branko.cibej@hermes.si>
		 */

		if (reload_after_sighup)
		{
			DEBUG(0, ("Reloading services after SIGHUP\n"));
			become_root();
			fn->reload_services(False);
			unbecome_root();
			reload_after_sighup = False;
			/*
			 * Use this as an excuse to print some stats.
			 */
		}

		if (got_msrpc)
		{
			process_msrpc(l, name, &pdu);
		}
		prs_free_data(&pdu);
	}
}
