/* 
 *  Unix SMB/Netbios implementation.
 *  Version 2.1.
 *  RPC client routines: scheduler service
 *  Copyright (C) Matthew Chapman                   1999,
 *  Copyright (C) Luke Kenneth Casson Leighton 1996-1999,
 *  Copyright (C) Andrew Tridgell              1992-1999.
 *  
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *  
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software Foundation,
 *  Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
/* MKPROTO: rpc_client_proto.h		(do not remove this line) */

#include "includes.h"
#include "libsamba.h"
#include "rpc_parse.h"
#include "rpc_client.h"
#include "nterr.h"


#define DO_CONNECT(srv_name, error_ret_val) \
	if (!cli_connection_init(srv_name, PIPE_ATSVC, &con)) \
	{ \
		return (error_ret_val); \
	}
#define DO_MARSHALL(fn) \
	(fn("", &q_q, &buf, 0))
#define DO_CON_PIPE_REQ(opcode) \
	rpc_con_pipe_req(con, (opcode), &buf, &rbuf)
#define DO_UNMARSHALL(fn) \
	(fn("", &r_q, &rbuf, 0))


/****************************************************************************
add a job to the scheduler
****************************************************************************/
BOOL at_add_job(char *srv_name, AT_JOB_INFO *info, char *command,
		uint32 *jobid)
{
	prs_struct rbuf;
	prs_struct buf;
	AT_Q_ADD_JOB q_q;
	AT_R_ADD_JOB r_q;
	NTSTATUS status = NT_STATUS_UNSUCCESSFUL;

	struct cli_connection *con = NULL;

	DO_CONNECT(srv_name, False);

	prs_init(&buf, MARSHALL);
	prs_init(&rbuf, UNMARSHALL);

	DEBUG(4, ("Scheduler Add Job\n"));

	/* store the parameters */
	make_at_q_add_job(&q_q, srv_name, info, command);

	/* turn parameters into data stream */
	if (!DO_MARSHALL(at_io_q_add_job)
	    || !DO_CON_PIPE_REQ(AT_ADD_JOB)
	    || !DO_UNMARSHALL(at_io_r_add_job))
	{
		goto out;
	}

	status = WIN32_TO_NTSTATUS(r_q.status);
	if (! NT_STATUS_IS_OK(status))
	{
		DEBUG(0, ("AT_ADD_JOB: %s\n",
			  get_nt_error_msg(r_q.status)));
	}

	*jobid = r_q.jobid;

    out:
	rpccli_close_prs(&buf, &rbuf);

	cli_connection_unlink(con);

	return NT_STATUS_IS_OK(status);
}

/****************************************************************************
dequeue a job
****************************************************************************/
BOOL at_del_job(char *srv_name, uint32 min_jobid, uint32 max_jobid)
{
	prs_struct rbuf;
	prs_struct buf;
	AT_Q_DEL_JOB q_q;
	AT_R_DEL_JOB r_q;
	NTSTATUS status = NT_STATUS_UNSUCCESSFUL;

	struct cli_connection *con = NULL;

	DO_CONNECT(srv_name, False);

	prs_init(&buf, MARSHALL);
	prs_init(&rbuf, UNMARSHALL);

	DEBUG(4, ("Scheduler Delete Job\n"));

	/* store the parameters */
	make_at_q_del_job(&q_q, srv_name, min_jobid, max_jobid);

	/* turn parameters into data stream */
	if (!DO_MARSHALL(at_io_q_del_job)
	    || !DO_CON_PIPE_REQ(AT_DEL_JOB)
	    || !DO_UNMARSHALL(at_io_r_del_job))
	{
		goto out;
	}

	status = WIN32_TO_NTSTATUS(r_q.status);

	if (! NT_STATUS_IS_OK(status))
	{
		DEBUG(0, ("AT_DEL_JOB: %s\n",
			  get_nt_error_msg(r_q.status)));
	}

    out:
	rpccli_close_prs(&buf, &rbuf);

	cli_connection_unlink(con);

	return NT_STATUS_IS_OK(status);
}

/****************************************************************************
enumerate scheduled jobs
****************************************************************************/
BOOL at_enum_jobs(char *srv_name, uint32 *num_jobs,
		  AT_ENUM_INFO *jobs, char ***commands)
{
	prs_struct rbuf;
	prs_struct buf;
	AT_Q_ENUM_JOBS q_q;
	AT_R_ENUM_JOBS r_q;
	NTSTATUS status = NT_STATUS_UNSUCCESSFUL;
	uint32 i;

	struct cli_connection *con = NULL;

	DO_CONNECT(srv_name, False);

	prs_init(&buf, MARSHALL);
	prs_init(&rbuf, UNMARSHALL);

	DEBUG(4, ("Scheduler Enumerate Jobs\n"));

	/* store the parameters */
	make_at_q_enum_jobs(&q_q, srv_name);

	/* turn parameters into data stream */
	if (!DO_MARSHALL(at_io_q_enum_jobs)
	    || !DO_CON_PIPE_REQ(AT_ENUM_JOBS)
	    || !DO_UNMARSHALL(at_io_r_enum_jobs))
	{
		goto out;
	}

	status = WIN32_TO_NTSTATUS(r_q.status);

	if (! NT_STATUS_IS_OK(status))
	{
		DEBUG(0, ("AT_ENUM_JOBS: %s\n",
			  get_nt_error_msg(status)));
	}


	*num_jobs = 0;
	memcpy(jobs, &r_q.info,
	       r_q.num_entries * sizeof(AT_ENUM_INFO));

	for (i = 0; i < r_q.num_entries; i++)
	{
		fstring cmd;
		unistr2_to_ascii(cmd, &r_q.command[i],
				 sizeof(cmd));
		add_chars_to_array(num_jobs, commands, cmd);
	}

    out:
	rpccli_close_prs(&buf, &rbuf);

	cli_connection_unlink(con);

	return NT_STATUS_IS_OK(status);
}

/****************************************************************************
query job information
****************************************************************************/
BOOL at_query_job(char *srv_name,
		  uint32 jobid, AT_JOB_INFO *job, fstring command)
{
	prs_struct rbuf;
	prs_struct buf;
	AT_Q_QUERY_JOB q_q;
	AT_R_QUERY_JOB r_q;
	NTSTATUS status = NT_STATUS_UNSUCCESSFUL;

	struct cli_connection *con = NULL;

	DO_CONNECT(srv_name, False);

	prs_init(&buf, MARSHALL);
	prs_init(&rbuf, UNMARSHALL);

	DEBUG(4, ("Scheduler Query Job\n"));

	/* store the parameters */
	make_at_q_query_job(&q_q, srv_name, jobid);

	/* turn parameters into data stream */
	if (!DO_MARSHALL(at_io_q_query_job)
	    || !DO_CON_PIPE_REQ(AT_QUERY_JOB)
	    || !DO_UNMARSHALL(at_io_r_query_job))
	{
		goto out;
	}

	status = WIN32_TO_NTSTATUS(r_q.status);

	if (! NT_STATUS_IS_OK(status))
	{
		DEBUG(0, ("AT_QUERY_JOB: %s\n",
			  get_nt_error_msg(status)));
	}

	*job = r_q.info;
	unistr2_to_ascii(command, &r_q.command,
			 sizeof(fstring)-1);

    out:
	rpccli_close_prs(&buf, &rbuf);

	cli_connection_unlink(con);

	return NT_STATUS_IS_OK(status);
}
