/* 
   Unix SMB/Netbios implementation.
   Version 1.9.
   NT Domain Authentication SMB / MSRPC client
   Copyright (C) Andrew Tridgell              1994-2000
   Copyright (C) Luke Kenneth Casson Leighton 1996-2000
   Copyright (C) Elrond                            2000
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/
/* MKPROTO: rpc_client_proto.h		(do not remove this line) */

#include "includes.h"
#include "libsamba.h"
#include "tng_misc.h"
#include "rpc_parse.h"
#include "rpc_client.h"
#include "nterr.h"
#include "others.h"

#define DECLARE_BASIC_VARS(MARSHALL_ST, UNMARSHALL_ST) \
	prs_struct data, rdata; \
	MARSHALL_ST q_o; \
	UNMARSHALL_ST r_o
#define INIT_BASIC_VARS() \
	rpccli_prs_open(&data, &rdata)
#define report_error_code(opcodename, ntstatus) \
	DEBUG(2, (opcodename ": %s\n", \
		  get_nt_error_msg(ntstatus)))
#define clean_and_ret(val) \
	rpccli_close_prs(&data, &rdata); \
	\
	return val;


/****************************************************************************
do a SAMR change user password command
****************************************************************************/
NTSTATUS samr_chgpasswd_user(struct cli_connection *con,
		const char *srv_name, const char *user_name,
		const char nt_newpass[516], const uchar nt_oldhash[16],
		const char lm_newpass[516], const uchar lm_oldhash[16])
{
	DECLARE_BASIC_VARS(SAMR_Q_CHGPASSWD_USER, SAMR_R_CHGPASSWD_USER);
	NTSTATUS status = NT_STATUS_UNSUCCESSFUL;

	INIT_BASIC_VARS();
	ZERO_STRUCT(r_o);

	DEBUG(4,("SAMR Change User Password. server:%s username:%s\n",
	        srv_name, user_name));

	make_samr_q_chgpasswd_user(&q_o, srv_name, user_name,
	                           nt_newpass, nt_oldhash,
	                           lm_newpass, lm_oldhash);

	/* turn parameters into data stream */
	if (samr_io_q_chgpasswd_user("", &q_o, &data, 0) &&
	    rpc_con_pipe_req(con, SAMR_CHGPASSWD_USER, &data, &rdata)
	    && samr_io_r_chgpasswd_user("", &r_o, &rdata, 0))
	{
		status = r_o.status;
		if (status != 0)
		{
			report_error_code("SAMR_CHGPASSWD_USER", status);
		}
	}

	clean_and_ret(status);
}


/****************************************************************************
do a SAMR unknown 0x38 command
****************************************************************************/
BOOL samr_get_dom_pwinfo(struct cli_connection *con, const char *srv_name)
{
	DECLARE_BASIC_VARS(SAMR_Q_GET_DOM_PWINFO, SAMR_R_GET_DOM_PWINFO);
	BOOL p;
	BOOL valid_un8 = False;

	INIT_BASIC_VARS();
	ZERO_STRUCT(r_o);

	DEBUG(4,("SAMR Query Domain Password Info server:%s\n", srv_name));

	make_samr_q_get_dom_pwinfo(&q_o, srv_name);

	/* turn parameters into data stream */
	if (samr_io_q_get_dom_pwinfo("", &q_o, &data, 0) &&
	    rpc_con_pipe_req(con, SAMR_GET_DOM_PWINFO, &data, &rdata))
	{
		samr_io_r_get_dom_pwinfo("", &r_o, &rdata, 0);

		p = rdata.offset != 0;
#if 0
		if (p && r_o.status != 0)
		{
			report_error_code("SAMR_GET_DOM_PWINFO", r_o.status);
			p = False;
		}
#endif
		if (p)
		{
			valid_un8 = True;
		}
	}
	else
	{
		DEBUG(4,("samr_unknown38: rpc_con_pipe_req failed\n"));
	}

	clean_and_ret(valid_un8);
}

/****************************************************************************
do a SAMR unknown 0x8 command
****************************************************************************/
BOOL samr_query_dom_info(  POLICY_HND *domain_pol, uint16 switch_value,
				SAM_UNK_CTR *ctr)
{
	DECLARE_BASIC_VARS(SAMR_Q_QUERY_DOMAIN_INFO, SAMR_R_QUERY_DOMAIN_INFO);
	BOOL p;
	BOOL valid_un8 = False;

	DEBUG(4,("SAMR Unknown 8 switch:%d\n", switch_value));

	if (domain_pol == NULL)
		return False;

	INIT_BASIC_VARS();
	ZERO_STRUCT(r_o);

	/* store the parameters */
	make_samr_q_query_dom_info(&q_o, domain_pol, switch_value);

	/* turn parameters into data stream */
	if (samr_io_q_query_dom_info("", &q_o, &data, 0) &&
	    rpc_hnd_pipe_req(domain_pol, SAMR_QUERY_DOMAIN_INFO, &data, &rdata))
	{
		r_o.ctr = ctr;
		samr_io_r_query_dom_info("", &r_o, &rdata, 0);

		p = rdata.offset != 0;
		if (p && r_o.status != 0)
		{
			report_error_code("SAMR_QUERY_DOMAIN_INFO", r_o.status);
			p = False;
		}

		if (p)
		{
			valid_un8 = True;
		}
	}

	clean_and_ret(valid_un8);
}

/****************************************************************************
do a SAMR enumerate Domains
****************************************************************************/
uint32 samr_enum_domains(  POLICY_HND *pol,
				uint32 *start_idx, uint32 size,
				struct acct_info **sam,
				uint32 *num_sam_domains)
{
	DECLARE_BASIC_VARS(SAMR_Q_ENUM_DOMAINS, SAMR_R_ENUM_DOMAINS);
	BOOL p;
	uint32 status = 0x0;

	DEBUG(4,("SAMR Enum SAM DB max size:%x\n", size));

	if (pol == NULL || num_sam_domains == NULL || sam == NULL)
		return NT_STATUS_INVALID_PARAMETER;

	INIT_BASIC_VARS();
	ZERO_STRUCT(r_o);

	/* store the parameters */
	make_samr_q_enum_domains(&q_o, pol, *start_idx, size);

	/* turn parameters into data stream */
	if (samr_io_q_enum_domains("", &q_o, &data, 0) &&
	    rpc_hnd_pipe_req(pol, SAMR_ENUM_DOMAINS, &data, &rdata))
	{
		samr_io_r_enum_domains("", &r_o, &rdata, 0);

		status = r_o.status;
		p = rdata.offset != 0;
		if (p && r_o.status != 0)
		{
			report_error_code("SAMR_ENUM_DOMAINS", r_o.status);
			p = (r_o.status == STATUS_MORE_ENTRIES);
		}

		if (p)
		{
			uint32 i = (*num_sam_domains);
			uint32 j = 0;
			uint32 name_idx = 0;

			(*num_sam_domains) += r_o.num_entries2;
			if (!g_xrenew(struct acct_info, sam, *num_sam_domains))
			{
				(*num_sam_domains) = 0;
				i = 0;
			}

			for (j = 0; i < (*num_sam_domains) && j < r_o.num_entries2; j++, i++)
			{
				(*sam)[i].rid = r_o.sam[j].rid;
				(*sam)[i].acct_name[0] = 0;
				(*sam)[i].acct_desc[0] = 0;
				if (r_o.sam[j].hdr_name.buffer)
				{
					unistr2_to_ascii((*sam)[i].acct_name, &r_o.uni_dom_name[name_idx], sizeof((*sam)[i].acct_name)-1);
					name_idx++;
				}
				DEBUG(5,("samr_enum_domains: idx: %4d rid: %8x acct: %s\n",
				          i, (*sam)[i].rid, (*sam)[i].acct_name));
			}
			(*start_idx) = r_o.next_idx;
		}
		else if (status == 0x0)
		{
			status = NT_STATUS_INVALID_PARAMETER | 0xC0000000;
		}

		if (r_o.sam != NULL)
		{
			free(r_o.sam);
		}
		if (r_o.uni_dom_name != NULL)
		{
			free(r_o.uni_dom_name);
		}
	}

	clean_and_ret(status);
}

/****************************************************************************
do a SAMR enumerate groups
****************************************************************************/
uint32 samr_enum_dom_groups(  POLICY_HND *pol,
				uint32 *start_idx, uint32 size,
				struct acct_info **sam,
				uint32 *num_sam_groups)
{
	DECLARE_BASIC_VARS(SAMR_Q_ENUM_DOM_GROUPS, SAMR_R_ENUM_DOM_GROUPS);
	BOOL p;
	uint32 status = 0x0;

	DEBUG(4,("SAMR Enum SAM DB max size:%x\n", size));

	if (pol == NULL || num_sam_groups == NULL)
		return NT_STATUS_INVALID_PARAMETER;

	INIT_BASIC_VARS();
	ZERO_STRUCT(r_o);

	/* store the parameters */
	make_samr_q_enum_dom_groups(&q_o, pol, *start_idx, size);

	/* turn parameters into data stream */
	if (samr_io_q_enum_dom_groups("", &q_o, &data, 0) &&
	    rpc_hnd_pipe_req(pol, SAMR_ENUM_DOM_GROUPS, &data, &rdata))
	{
		samr_io_r_enum_dom_groups("", &r_o, &rdata, 0);

		status = r_o.status;
		p = rdata.offset != 0;
		if (p && r_o.status != 0)
		{
			report_error_code("SAMR_ENUM_DOM_GROUPS", r_o.status);
			p = (r_o.status == STATUS_MORE_ENTRIES);
		}

		if (p)
		{
			uint32 i = (*num_sam_groups);
			uint32 j = 0;
			uint32 name_idx = 0;

			(*num_sam_groups) += r_o.num_entries2;
			if (!g_xrenew(struct acct_info, sam, *num_sam_groups))
			{
				(*num_sam_groups) = 0;
				i = 0;
			}

			for (j = 0; i < (*num_sam_groups) && j < r_o.num_entries2; j++, i++)
			{
				(*sam)[i].rid = r_o.sam[j].rid;
				(*sam)[i].acct_name[0] = 0;
				(*sam)[i].acct_desc[0] = 0;
				if (r_o.sam[j].hdr_name.buffer)
				{
					unistr2_to_ascii((*sam)[i].acct_name, &r_o.uni_grp_name[name_idx], sizeof((*sam)[i].acct_name)-1);
					name_idx++;
				}
				DEBUG(5,("samr_enum_dom_groups: idx: %4d rid: %8x acct: %s\n",
				          i, (*sam)[i].rid, (*sam)[i].acct_name));
			}
			(*start_idx) = r_o.next_idx;
		}
		else if (status == 0x0)
		{
			status = NT_STATUS_INVALID_PARAMETER | 0xC0000000;
		}

		if (r_o.sam != NULL)
		{
			free(r_o.sam);
		}
		if (r_o.uni_grp_name != NULL)
		{
			free(r_o.uni_grp_name);
		}
	}

	clean_and_ret(status);
}

/****************************************************************************
do a SAMR enumerate aliases
****************************************************************************/
uint32 samr_enum_dom_aliases(  POLICY_HND *pol,
				uint32 *start_idx, uint32 size,
				struct acct_info **sam,
				uint32 *num_sam_aliases)
{
	DECLARE_BASIC_VARS(SAMR_Q_ENUM_DOM_ALIASES, SAMR_R_ENUM_DOM_ALIASES);
	BOOL p;
	uint32 status = 0x0;

	DEBUG(4,("SAMR Enum SAM DB max size:%x\n", size));

	if (pol == NULL || num_sam_aliases == NULL)
		return NT_STATUS_INVALID_PARAMETER;

	INIT_BASIC_VARS();
	ZERO_STRUCT(r_o);

	/* store the parameters */
	make_samr_q_enum_dom_aliases(&q_o, pol, *start_idx, size);

	/* turn parameters into data stream */
	if (samr_io_q_enum_dom_aliases("", &q_o, &data, 0) &&
	    rpc_hnd_pipe_req(pol, SAMR_ENUM_DOM_ALIASES, &data, &rdata))
	{
		samr_io_r_enum_dom_aliases("", &r_o, &rdata, 0);

		p = rdata.offset != 0;
		if (p && r_o.status != 0)
		{
			report_error_code("SAMR_ENUM_DOM_ALIASES", r_o.status);
			p = (r_o.status == STATUS_MORE_ENTRIES);
		}

		if (p)
		{
			uint32 i = (*num_sam_aliases);
			uint32 j = 0;
			uint32 name_idx = 0;

			(*num_sam_aliases) += r_o.num_entries2;
			if (!g_xrenew(struct acct_info, sam, *num_sam_aliases))
			{
				(*num_sam_aliases) = 0;
				i = 0;
			}

			for (j = 0; i < (*num_sam_aliases) && j < r_o.num_entries2; j++, i++)
			{
				(*sam)[i].rid = r_o.sam[j].rid;
				(*sam)[i].acct_name[0] = 0;
				(*sam)[i].acct_desc[0] = 0;
				if (r_o.sam[j].hdr_name.buffer)
				{
					unistr2_to_ascii((*sam)[i].acct_name, &r_o.uni_grp_name[name_idx], sizeof((*sam)[i].acct_name)-1);
					name_idx++;
				}
				DEBUG(5,("samr_enum_dom_aliases: idx: %4d rid: %8x acct: %s\n",
				          i, (*sam)[i].rid, (*sam)[i].acct_name));
			}
			(*start_idx) = r_o.next_idx;
		}
		else if (status == 0x0)
		{
			status = NT_STATUS_INVALID_PARAMETER | 0xC0000000;
		}

		if (r_o.sam != NULL)
		{
			free(r_o.sam);
		}
		if (r_o.uni_grp_name != NULL)
		{
			free(r_o.uni_grp_name);
		}
	}

	clean_and_ret(status);
}

/****************************************************************************
do a SAMR enumerate users
****************************************************************************/
uint32 samr_enum_dom_users(  POLICY_HND *pol, uint32 *start_idx, 
				uint16 acb_mask, uint16 unk_1, uint32 size,
				struct acct_info **sam,
				uint32 *num_sam_users)
{
	DECLARE_BASIC_VARS(SAMR_Q_ENUM_DOM_USERS, SAMR_R_ENUM_DOM_USERS);
	BOOL p;
	uint32 status = 0x0;

	DEBUG(4,("SAMR Enum SAM DB max size:%x\n", size));

	if (pol == NULL || num_sam_users == NULL)
		return NT_STATUS_INVALID_PARAMETER;

	INIT_BASIC_VARS();
	ZERO_STRUCT(r_o);

	/* store the parameters */
	make_samr_q_enum_dom_users(&q_o, pol, *start_idx,
	                           acb_mask, unk_1, size);

	/* turn parameters into data stream */
	if (samr_io_q_enum_dom_users("", &q_o, &data, 0) &&
	    rpc_hnd_pipe_req(pol, SAMR_ENUM_DOM_USERS, &data, &rdata))
	{
		samr_io_r_enum_dom_users("", &r_o, &rdata, 0);

		status = r_o.status;
		p = rdata.offset != 0;

		if (p && r_o.status != 0)
		{
			report_error_code("SAMR_ENUM_DOM_USERS", r_o.status);
			p = (r_o.status == STATUS_MORE_ENTRIES);
		}

		if (p)
		{
			uint32 i = (*num_sam_users);
			uint32 j = 0;
			uint32 name_idx = 0;

			(*num_sam_users) += r_o.num_entries2;
			if ((*num_sam_users) != 0)
			{
				/* FIXME - check return */
				g_xrenew(struct acct_info,sam, *num_sam_users);
			}

			if ((*sam) == NULL)
			{
				(*num_sam_users) = 0;
				i = 0;
			}

			for (j = 0; i < (*num_sam_users) && j < r_o.num_entries2; j++, i++)
			{
				(*sam)[i].rid = r_o.sam[j].rid;
				(*sam)[i].acct_name[0] = 0;
				(*sam)[i].acct_desc[0] = 0;
				if (r_o.sam[j].hdr_name.buffer)
				{
					unistr2_to_ascii((*sam)[i].acct_name, &r_o.uni_acct_name[name_idx], sizeof((*sam)[i].acct_name)-1);
					name_idx++;
				}
				DEBUG(5,("samr_enum_dom_users: idx: %4d rid: %8x acct: %s\n",
				          i, (*sam)[i].rid, (*sam)[i].acct_name));
			}
			(*start_idx) = r_o.next_idx;
		}
		else if (status == 0x0)
		{
			status = NT_STATUS_INVALID_PARAMETER | 0xC0000000;
		}

		safe_free(r_o.sam);
		safe_free(r_o.uni_acct_name);
	}
	else
	{
		status = NT_STATUS_ACCESS_DENIED | 0xC0000000;
	}

	clean_and_ret(status);
}

/****************************************************************************
do a SAMR Connect
****************************************************************************/
BOOL samr_connect(  const char *srv_name, uint32 access_mask,
				POLICY_HND *connect_pol)
{
	DECLARE_BASIC_VARS(SAMR_Q_CONNECT, SAMR_R_CONNECT);
	BOOL p;
	BOOL valid_pol = False;

	struct cli_connection *con = NULL;

	if (!cli_connection_init(srv_name, PIPE_SAMR, &con))
	{
		return False;
	}

	DEBUG(4,("SAMR Open Policy server:%s access_mask:%x\n",
				srv_name, access_mask));

	if (srv_name == NULL || connect_pol == NULL) return False;

	INIT_BASIC_VARS();

	/* store the parameters */
	make_samr_q_connect(&q_o, srv_name, access_mask);

	/* turn parameters into data stream */
	if (samr_io_q_connect("", &q_o,  &data, 0) &&
	    rpc_con_pipe_req(con, SAMR_CONNECT, &data, &rdata))
	{
		samr_io_r_connect("", &r_o, &rdata, 0);
		p = rdata.offset != 0;

		if (p && r_o.status != 0)
		{
			report_error_code("SAMR_CONNECT", r_o.status);
			p = False;
		}

		if (p)
		{
			*connect_pol = r_o.connect_pol;
			valid_pol = register_policy_hnd(get_global_hnd_cache(),
			                                cli_con_sec_ctx(con),
			                                connect_pol,
			                                access_mask) &&
			            set_policy_con(get_global_hnd_cache(),
			                                 connect_pol, con,
			                                 cli_connection_unlink);
			if (valid_pol)
			{
				policy_hnd_set_name(get_global_hnd_cache(),
						    connect_pol,
						    "SAM_CONNECT");
			}
		}
	}

	clean_and_ret(valid_pol);
}

/****************************************************************************
do a SAMR Query Security Object
****************************************************************************/
BOOL samr_query_sec_obj(  const POLICY_HND *pol,
				uint32 type,
				SEC_DESC_BUF *buf)
{
	DECLARE_BASIC_VARS(SAMR_Q_QUERY_SEC_OBJ, SAMR_R_QUERY_SEC_OBJ);
	BOOL p;
	BOOL valid_pol = False;

	DEBUG(4,("SAMR Query Sec Object: type %x\n", type));

	if (pol == NULL)
		return False;

	INIT_BASIC_VARS();
	ZERO_STRUCT(r_o);

	/* store the parameters */
	make_samr_q_query_sec_obj(&q_o, pol, type);

	/* turn parameters into data stream */
	if (samr_io_q_query_sec_obj("", &q_o,  &data, 0) &&
	    rpc_hnd_pipe_req(pol, SAMR_QUERY_SEC_OBJECT, &data, &rdata))
	{
		samr_io_r_query_sec_obj("", &r_o, &rdata, 0);
		p = rdata.offset != 0;

		if (p && r_o.status != 0)
		{
			report_error_code("SAMR_QUERY_SEC_OBJ", r_o.status);
			p = False;
		}

		if (p)
		{
			valid_pol = True;
			buf->sec = r_o.buf.sec;
		}
	}

	clean_and_ret(valid_pol);
}

/****************************************************************************
do a SAMR Open User
****************************************************************************/
BOOL samr_open_user(  const POLICY_HND *pol,
				uint32 access_mask, uint32 rid, 
				POLICY_HND *user_pol)
{
	DECLARE_BASIC_VARS(SAMR_Q_OPEN_USER, SAMR_R_OPEN_USER);
	BOOL p;
	BOOL valid_pol = False;

	DEBUG(4,("SAMR Open User.  access_mask: %08x RID:%x\n",
	          access_mask, rid));

	if (pol == NULL || user_pol == NULL) return False;

	INIT_BASIC_VARS();

	/* store the parameters */
	make_samr_q_open_user(&q_o, pol, access_mask, rid);

	/* turn parameters into data stream */
	if (samr_io_q_open_user("", &q_o,  &data, 0) &&
	    rpc_hnd_pipe_req(pol, SAMR_OPEN_USER, &data, &rdata))
	{
		samr_io_r_open_user("", &r_o, &rdata, 0);
		p = rdata.offset != 0;

		if (p && r_o.status != 0)
		{
			report_error_code("SAMR_OPEN_USER", r_o.status);
			p = False;
		}

		if (p)
		{
			*user_pol = r_o.user_pol;
			valid_pol = cli_pol_link(user_pol, pol);
		}
	}

	clean_and_ret(valid_pol);
}

/****************************************************************************
do a SAMR Open Alias
****************************************************************************/
BOOL samr_open_alias(  const POLICY_HND *domain_pol,
				uint32 flags, uint32 rid,
				POLICY_HND *alias_pol)
{
	DECLARE_BASIC_VARS(SAMR_Q_OPEN_ALIAS, SAMR_R_OPEN_ALIAS);
	BOOL p;
	BOOL valid_pol = False;

	DEBUG(4,("SAMR Open Alias. RID:%x\n", rid));

	if (alias_pol == NULL || domain_pol == NULL) return False;

	INIT_BASIC_VARS();

	/* store the parameters */
	make_samr_q_open_alias(&q_o, domain_pol, flags, rid);

	/* turn parameters into data stream */
	if (samr_io_q_open_alias("", &q_o,  &data, 0) &&
	    rpc_hnd_pipe_req(domain_pol, SAMR_OPEN_ALIAS, &data, &rdata))
	{
		samr_io_r_open_alias("", &r_o, &rdata, 0);
		p = rdata.offset != 0;

		if (p && r_o.status != 0)
		{
			report_error_code("SAMR_OPEN_ALIAS", r_o.status);
			p = False;
		}

		if (p)
		{
			*alias_pol = r_o.pol;
			valid_pol = cli_pol_link(alias_pol, domain_pol);

			if (valid_pol)
			{
				policy_hnd_set_name(get_global_hnd_cache(),
						    alias_pol,
						    "SAM_ALIAS");
			}
		}
	}

	clean_and_ret(valid_pol);
}

/****************************************************************************
do a SAMR Delete Alias Member
****************************************************************************/
BOOL samr_del_aliasmem(  POLICY_HND *alias_pol, DOM_SID *sid)
{
	DECLARE_BASIC_VARS(SAMR_Q_DEL_ALIASMEM, SAMR_R_DEL_ALIASMEM);
	BOOL p;
	BOOL valid_pol = False;

	if (alias_pol == NULL || sid == NULL) return False;

	INIT_BASIC_VARS();

	DEBUG(4,("SAMR Delete Alias Member.\n"));

	/* store the parameters */
	make_samr_q_del_aliasmem(&q_o, alias_pol, sid);

	/* turn parameters into data stream */
	if (samr_io_q_del_aliasmem("", &q_o,  &data, 0) &&
	    rpc_hnd_pipe_req(alias_pol, SAMR_DEL_ALIASMEM, &data, &rdata))
	{
		samr_io_r_del_aliasmem("", &r_o, &rdata, 0);
		p = rdata.offset != 0;

		if (p && r_o.status != 0)
		{
			report_error_code("SAMR_DEL_ALIASMEM", r_o.status);
			p = False;
		}

		if (p)
		{
			valid_pol = True;
		}
	}

	clean_and_ret(valid_pol);
}

/****************************************************************************
do a SAMR Add Alias Member
****************************************************************************/
BOOL samr_add_aliasmem(  POLICY_HND *alias_pol, DOM_SID *sid)
{
	DECLARE_BASIC_VARS(SAMR_Q_ADD_ALIASMEM, SAMR_R_ADD_ALIASMEM);
	BOOL p;
	BOOL valid_pol = False;

	if (alias_pol == NULL || sid == NULL) return False;

	INIT_BASIC_VARS();

	DEBUG(4,("SAMR Add Alias Member.\n"));

	/* store the parameters */
	make_samr_q_add_aliasmem(&q_o, alias_pol, sid);

	/* turn parameters into data stream */
	if (samr_io_q_add_aliasmem("", &q_o,  &data, 0) &&
	    rpc_hnd_pipe_req(alias_pol, SAMR_ADD_ALIASMEM, &data, &rdata))
	{
		samr_io_r_add_aliasmem("", &r_o, &rdata, 0);
		p = rdata.offset != 0;

		if (p && r_o.status != 0)
		{
			report_error_code("SAMR_ADD_ALIASMEM", r_o.status);
			p = False;
		}

		if (p)
		{
			valid_pol = True;
		}
	}

	clean_and_ret(valid_pol);
}

/****************************************************************************
do a SAMR Delete Domain Alias
****************************************************************************/
BOOL samr_delete_dom_alias(  POLICY_HND *alias_pol)
{
	DECLARE_BASIC_VARS(SAMR_Q_DELETE_DOM_ALIAS, SAMR_R_DELETE_DOM_ALIAS);
	BOOL p;
	BOOL valid_pol = False;

	if (alias_pol == NULL) return False;

	INIT_BASIC_VARS();

	DEBUG(4,("SAMR Delete Domain Alias.\n"));

	/* store the parameters */
	make_samr_q_delete_dom_alias(&q_o, alias_pol);

	/* turn parameters into data stream */
	if (samr_io_q_delete_dom_alias("", &q_o,  &data, 0) &&
	    rpc_hnd_pipe_req(alias_pol, SAMR_DELETE_DOM_ALIAS, &data, &rdata))
	{
		samr_io_r_delete_dom_alias("", &r_o, &rdata, 0);
		p = rdata.offset != 0;

		if (p && r_o.status != 0)
		{
			report_error_code("SAMR_DELETE_DOM_ALIAS", r_o.status);
			p = False;
		}

		if (p)
		{
			valid_pol = True;
		}
	}

	clean_and_ret(valid_pol);
}

/****************************************************************************
do a SAMR Create Domain User
****************************************************************************/
uint32 samr_create_dom_user(  POLICY_HND *domain_pol, const char *acct_name,
				uint32 unk_0, uint32 unk_1,
				POLICY_HND *user_pol, uint32 *rid)
{
	DECLARE_BASIC_VARS(SAMR_Q_CREATE_USER, SAMR_R_CREATE_USER);
	BOOL p;
	uint32 status = NT_STATUS_INVALID_PARAMETER;

	if (user_pol == NULL || domain_pol == NULL || acct_name == NULL || rid == NULL) return False;

	INIT_BASIC_VARS();

	DEBUG(4,("SAMR Create Domain User. Name:%s\n", acct_name));

	/* store the parameters */
	make_samr_q_create_user(&q_o, domain_pol, acct_name, unk_0, unk_1);

	/* turn parameters into data stream */
	if (samr_io_q_create_user("", &q_o,  &data, 0) &&
	    rpc_hnd_pipe_req(domain_pol, SAMR_CREATE_USER, &data, &rdata))
	{
		samr_io_r_create_user("", &r_o, &rdata, 0);
		p = rdata.offset != 0;
		status = r_o.status;

		if (p && r_o.status != 0)
		{
			report_error_code("SAMR_CREATE_USER", r_o.status);
			p = r_o.status != NT_STATUS_USER_EXISTS;
		}

		if (p)
		{
			*user_pol = r_o.user_pol;
			*rid = r_o.user_rid;
			if (!cli_pol_link(user_pol, domain_pol))
			{
				status = NT_STATUS_INVALID_HANDLE | 0xC0000000;
			}
		}
	}

	clean_and_ret(status);
}

/****************************************************************************
do a SAMR Create Domain Alias
****************************************************************************/
BOOL samr_create_dom_alias(  POLICY_HND *domain_pol, const char *acct_name,
				POLICY_HND *alias_pol, uint32 *rid)
{
	DECLARE_BASIC_VARS(SAMR_Q_CREATE_DOM_ALIAS, SAMR_R_CREATE_DOM_ALIAS);
	BOOL p;
	BOOL valid_pol = False;

	if (alias_pol == NULL || domain_pol == NULL || acct_name == NULL || rid == NULL) return False;

	INIT_BASIC_VARS();

	DEBUG(4,("SAMR Create Domain Alias. Name:%s\n", acct_name));

	/* store the parameters */
	make_samr_q_create_dom_alias(&q_o, domain_pol, acct_name);

	/* turn parameters into data stream */
	if (samr_io_q_create_dom_alias("", &q_o,  &data, 0) &&
	    rpc_hnd_pipe_req(domain_pol, SAMR_CREATE_DOM_ALIAS, &data, &rdata))
	{
		samr_io_r_create_dom_alias("", &r_o, &rdata, 0);
		p = rdata.offset != 0;

		if (p && r_o.status != 0)
		{
			report_error_code("SAMR_CREATE_DOM_ALIAS", r_o.status);
			p = False;
		}

		if (p)
		{
			*alias_pol = r_o.alias_pol;
			*rid = r_o.rid;
			valid_pol = cli_pol_link(alias_pol, domain_pol);
		}
	}

	clean_and_ret(valid_pol);
}

/****************************************************************************
do a SAMR Get Alias Info
****************************************************************************/
BOOL samr_query_aliasinfo(  POLICY_HND *alias_pol, uint16 switch_value,
				ALIAS_INFO_CTR *ctr)
{
	DECLARE_BASIC_VARS(SAMR_Q_QUERY_ALIASINFO, SAMR_R_QUERY_ALIASINFO);
	BOOL p;
	BOOL valid_pol = False;

	if (alias_pol == NULL || ctr == NULL) return False;

	INIT_BASIC_VARS();

	DEBUG(4,("SAMR Get Alias Info\n"));

	/* store the parameters */
	make_samr_q_query_aliasinfo(&q_o, alias_pol, switch_value);

	/* turn parameters into data stream */
	if (samr_io_q_query_aliasinfo("", &q_o,  &data, 0) &&
	    rpc_hnd_pipe_req(alias_pol, SAMR_QUERY_ALIASINFO, &data, &rdata))
	{
		/* get alias info */
		r_o.ctr = ctr;

		samr_io_r_query_aliasinfo("", &r_o, &rdata, 0);
		p = rdata.offset != 0;

		if (p && r_o.status != 0)
		{
			report_error_code("SAMR_QUERY_ALIASINFO", r_o.status);
			p = False;
		}

		if (p)
		{
			valid_pol = True;
		}
	}

	clean_and_ret(valid_pol);
}

/****************************************************************************
do a SAMR Set Alias Info
****************************************************************************/
BOOL samr_set_aliasinfo(  POLICY_HND *alias_pol, ALIAS_INFO_CTR *ctr)
{
	DECLARE_BASIC_VARS(SAMR_Q_SET_ALIASINFO, SAMR_R_SET_ALIASINFO);
	BOOL p;
	BOOL valid_pol = False;

	if (alias_pol == NULL || ctr == NULL) return False;

	INIT_BASIC_VARS();

	DEBUG(4,("SAMR Set Alias Info\n"));

	/* store the parameters */
	make_samr_q_set_aliasinfo(&q_o, alias_pol, ctr);

	/* turn parameters into data stream */
	if (samr_io_q_set_aliasinfo("", &q_o,  &data, 0) &&
	    rpc_hnd_pipe_req(alias_pol, SAMR_SET_ALIASINFO, &data, &rdata))
	{
		samr_io_r_set_aliasinfo("", &r_o, &rdata, 0);
		p = rdata.offset != 0;

		if (p && r_o.status != 0)
		{
			report_error_code("SAMR_SET_ALIASINFO", r_o.status);
			p = False;
		}

		if (p)
		{
			valid_pol = True;
		}
	}

	clean_and_ret(valid_pol);
}

/****************************************************************************
do a SAMR Open Group
****************************************************************************/
BOOL samr_open_group(  const POLICY_HND *domain_pol,
				uint32 flags, uint32 rid,
				POLICY_HND *group_pol)
{
	DECLARE_BASIC_VARS(SAMR_Q_OPEN_GROUP, SAMR_R_OPEN_GROUP);
	BOOL p;
	BOOL valid_pol = False;

	DEBUG(4,("SAMR Open Group. RID:%x\n", rid));

	if (group_pol == NULL || domain_pol == NULL) return False;

	INIT_BASIC_VARS();

	/* store the parameters */
	make_samr_q_open_group(&q_o, domain_pol, flags, rid);

	/* turn parameters into data stream */
	if (samr_io_q_open_group("", &q_o,  &data, 0) &&
	    rpc_hnd_pipe_req(domain_pol, SAMR_OPEN_GROUP, &data, &rdata))
	{
		samr_io_r_open_group("", &r_o, &rdata, 0);
		p = rdata.offset != 0;

		if (p && r_o.status != 0)
		{
			report_error_code("SAMR_OPEN_GROUP", r_o.status);
			p = False;
		}

		if (p)
		{
			*group_pol = r_o.pol;
			valid_pol = cli_pol_link(group_pol, domain_pol);
		}
	}

	clean_and_ret(valid_pol);
}

/****************************************************************************
do a SAMR Delete Group Member
****************************************************************************/
BOOL samr_del_groupmem(  POLICY_HND *group_pol, uint32 rid)
{
	DECLARE_BASIC_VARS(SAMR_Q_DEL_GROUPMEM, SAMR_R_DEL_GROUPMEM);
	BOOL p;
	BOOL valid_pol = False;

	if (group_pol == NULL) return False;

	INIT_BASIC_VARS();

	DEBUG(4,("SAMR Delete Group Member.\n"));

	/* store the parameters */
	make_samr_q_del_groupmem(&q_o, group_pol, rid);

	/* turn parameters into data stream */
	if (samr_io_q_del_groupmem("", &q_o,  &data, 0) &&
	    rpc_hnd_pipe_req(group_pol, SAMR_DEL_GROUPMEM, &data, &rdata))
	{
		samr_io_r_del_groupmem("", &r_o, &rdata, 0);
		p = rdata.offset != 0;

		if (p && r_o.status != 0)
		{
			report_error_code("SAMR_DEL_GROUPMEM", r_o.status);
			p = False;
		}

		if (p)
		{
			valid_pol = True;
		}
	}

	clean_and_ret(valid_pol);
}

/****************************************************************************
do a SAMR Add Group Member
****************************************************************************/
BOOL samr_add_groupmem(  POLICY_HND *group_pol, uint32 rid)
{
	DECLARE_BASIC_VARS(SAMR_Q_ADD_GROUPMEM, SAMR_R_ADD_GROUPMEM);
	BOOL p;
	BOOL valid_pol = False;

	if (group_pol == NULL) return False;

	INIT_BASIC_VARS();

	DEBUG(4,("SAMR Add Group Member.\n"));

	/* store the parameters */
	make_samr_q_add_groupmem(&q_o, group_pol, rid);

	/* turn parameters into data stream */
	if (samr_io_q_add_groupmem("", &q_o,  &data, 0) &&
	    rpc_hnd_pipe_req(group_pol, SAMR_ADD_GROUPMEM, &data, &rdata))
	{
		samr_io_r_add_groupmem("", &r_o, &rdata, 0);
		p = rdata.offset != 0;

		if (p && r_o.status != 0)
		{
			report_error_code("SAMR_ADD_GROUPMEM", r_o.status);
			p = False;
		}

		if (p)
		{
			valid_pol = True;
		}
	}

	clean_and_ret(valid_pol);
}

/****************************************************************************
do a SAMR Delete Domain User
****************************************************************************/
BOOL samr_delete_dom_user(  POLICY_HND *user_pol)
{
	DECLARE_BASIC_VARS(SAMR_Q_DELETE_DOM_USER, SAMR_R_DELETE_DOM_USER);
	BOOL p;
	BOOL valid_pol = False;

	if (user_pol == NULL) return False;

	INIT_BASIC_VARS();

	DEBUG(4,("SAMR Delete Domain User.\n"));

	/* store the parameters */
	make_samr_q_delete_dom_user(&q_o, user_pol);

	/* turn parameters into data stream */
	if (samr_io_q_delete_dom_user("", &q_o,  &data, 0) &&
	    rpc_hnd_pipe_req(user_pol, SAMR_DELETE_DOM_USER, &data, &rdata))
	{
		samr_io_r_delete_dom_user("", &r_o, &rdata, 0);
		p = rdata.offset != 0;

		if (p && r_o.status != 0)
		{
			report_error_code("SAMR_DELETE_DOM_USER", r_o.status);
			p = False;
		}

		if (p)
		{
			valid_pol = True;
		}
	}

	clean_and_ret(valid_pol);
}

/****************************************************************************
do a SAMR Delete Domain Group
****************************************************************************/
BOOL samr_delete_dom_group(  POLICY_HND *group_pol)
{
	DECLARE_BASIC_VARS(SAMR_Q_DELETE_DOM_GROUP, SAMR_R_DELETE_DOM_GROUP);
	BOOL p;
	BOOL valid_pol = False;

	if (group_pol == NULL) return False;

	INIT_BASIC_VARS();

	DEBUG(4,("SAMR Delete Domain Group.\n"));

	/* store the parameters */
	make_samr_q_delete_dom_group(&q_o, group_pol);

	/* turn parameters into data stream */
	if (samr_io_q_delete_dom_group("", &q_o,  &data, 0) &&
	    rpc_hnd_pipe_req(group_pol, SAMR_DELETE_DOM_GROUP, &data, &rdata))
	{
		samr_io_r_delete_dom_group("", &r_o, &rdata, 0);
		p = rdata.offset != 0;

		if (p && r_o.status != 0)
		{
			report_error_code("SAMR_DELETE_DOM_GROUP", r_o.status);
			p = False;
		}

		if (p)
		{
			valid_pol = True;
		}
	}

	clean_and_ret(valid_pol);
}

/****************************************************************************
do a SAMR Create Domain Group
****************************************************************************/
BOOL samr_create_dom_group(  POLICY_HND *domain_pol, const char *acct_name,
				uint32 access_mask,
				POLICY_HND *group_pol, uint32 *rid)
{
	DECLARE_BASIC_VARS(SAMR_Q_CREATE_DOM_GROUP, SAMR_R_CREATE_DOM_GROUP);
	BOOL p;
	BOOL valid_pol = False;

	if (group_pol == NULL || domain_pol == NULL || acct_name == NULL || rid == NULL) return False;

	INIT_BASIC_VARS();

	DEBUG(4,("SAMR Create Domain Group. Name:%s\n", acct_name));

	/* store the parameters */
	make_samr_q_create_dom_group(&q_o, domain_pol, acct_name, access_mask);

	/* turn parameters into data stream */
	if (samr_io_q_create_dom_group("", &q_o,  &data, 0) &&
	    rpc_hnd_pipe_req(domain_pol, SAMR_CREATE_DOM_GROUP, &data, &rdata))
	{
		samr_io_r_create_dom_group("", &r_o, &rdata, 0);
		p = rdata.offset != 0;

		if (p && r_o.status != 0)
		{
			report_error_code("SAMR_CREATE_DOM_GROUP", r_o.status);
			p = False;
		}

		if (p)
		{
			*group_pol = r_o.pol;
			*rid = r_o.rid;
			valid_pol = cli_pol_link(group_pol, domain_pol);
		}
	}

	clean_and_ret(valid_pol);
}

/****************************************************************************
do a SAMR Set Group Info
****************************************************************************/
BOOL samr_set_groupinfo(  POLICY_HND *group_pol, GROUP_INFO_CTR *ctr)
{
	DECLARE_BASIC_VARS(SAMR_Q_SET_GROUPINFO, SAMR_R_SET_GROUPINFO);
	BOOL p;
	BOOL valid_pol = False;

	if (group_pol == NULL || ctr == NULL) return False;

	INIT_BASIC_VARS();

	DEBUG(4,("SAMR Set Group Info\n"));

	/* store the parameters */
	make_samr_q_set_groupinfo(&q_o, group_pol, ctr);

	/* turn parameters into data stream */
	if (samr_io_q_set_groupinfo("", &q_o,  &data, 0) &&
	    rpc_hnd_pipe_req(group_pol, SAMR_SET_GROUPINFO, &data, &rdata))
	{
		samr_io_r_set_groupinfo("", &r_o, &rdata, 0);
		p = rdata.offset != 0;

		if (p && r_o.status != 0)
		{
			report_error_code("SAMR_SET_GROUPINFO", r_o.status);
			p = False;
		}

		if (p)
		{
			valid_pol = True;
		}
	}

	clean_and_ret(valid_pol);
}

/****************************************************************************
do a SAMR Unknown 2d
****************************************************************************/
BOOL samr_unknown_2d(  const POLICY_HND *domain_pol,
				const DOM_SID *sid)
{
	DECLARE_BASIC_VARS(SAMR_Q_UNKNOWN_2D, SAMR_R_UNKNOWN_2D);
	BOOL p;
	BOOL valid_pol = False;
	pstring sid_str;

	if (DEBUGLVL(4))
	{
		sid_to_string(sid_str, sid);
		DEBUG(4,("SAMR Unknown 0x2d.  SID:%s\n", sid_str));
	}

	if (sid == NULL || domain_pol == NULL) return False;

	INIT_BASIC_VARS();

	/* store the parameters */
	make_samr_q_unknown_2d(&q_o, domain_pol, sid);

	/* turn parameters into data stream */
	if (samr_io_q_unknown_2d("", &q_o,  &data, 0) &&
	    rpc_hnd_pipe_req(domain_pol, SAMR_UNKNOWN_2D, &data, &rdata))
	{
		samr_io_r_unknown_2d("", &r_o, &rdata, 0);
		p = rdata.offset != 0;

		if (p && r_o.status != 0)
		{
			report_error_code("SAMR_UNKNOWN_2D", r_o.status);
			p = False;
		}

		if (p)
		{
			valid_pol = True;
		}
	}

	clean_and_ret(valid_pol);
}

BOOL samr_open_domain(  const POLICY_HND *connect_pol,
			  uint32 ace_perms,
			  const DOM_SID *sid,
			  POLICY_HND *domain_pol)
{
	return isamr_open_domain(connect_pol, ace_perms, sid, domain_pol) ==
		NT_STATUS_NOPROBLEMO;
}

/****************************************************************************
do a SAMR Open Domain
****************************************************************************/
uint32 isamr_open_domain(  const POLICY_HND *connect_pol,
			  uint32 ace_perms,
			  const DOM_SID *sid,
			  POLICY_HND *domain_pol)
{
	DECLARE_BASIC_VARS(SAMR_Q_OPEN_DOMAIN, SAMR_R_OPEN_DOMAIN);
	BOOL p;
	pstring sid_str;

	r_o.status = NT_STATUS_UNSUCCESSFUL;

	if (DEBUGLVL(4))
	{
		sid_to_string(sid_str, sid);
		DEBUG(4,("SAMR Open Domain.  SID:%s Permissions:%x\n",
					sid_str, ace_perms));
	}

	if (connect_pol == NULL || sid == NULL || domain_pol == NULL) 
		return NT_STATUS_INVALID_PARAMETER;

	INIT_BASIC_VARS();

	/* store the parameters */
	make_samr_q_open_domain(&q_o, connect_pol, ace_perms, sid);

	/* turn parameters into data stream */
	if (samr_io_q_open_domain("", &q_o,  &data, 0) &&
	    rpc_hnd_pipe_req(connect_pol, SAMR_OPEN_DOMAIN, &data, &rdata))
	{
		samr_io_r_open_domain("", &r_o, &rdata, 0);
		p = rdata.offset != 0;

		if (p && r_o.status != 0)
		{
			report_error_code("SAMR_OPEN_DOMAIN", r_o.status);
			p = False;
		}

		if (p)
		{
			*domain_pol = r_o.domain_pol;
			if (!cli_pol_link(domain_pol, connect_pol)) {
				r_o.status = NT_STATUS_INTERNAL_DB_CORRUPTION;
			}
			policy_hnd_set_name(get_global_hnd_cache(),
					    domain_pol, "SAM_DOMAIN");
		}
	}

	clean_and_ret(r_o.status);
}

/****************************************************************************
do a SAMR Query Lookup Domain
****************************************************************************/
BOOL samr_query_lookup_domain(  POLICY_HND *pol, const char *dom_name,
			      DOM_SID *dom_sid)
{
	DECLARE_BASIC_VARS(SAMR_Q_LOOKUP_DOMAIN, SAMR_R_LOOKUP_DOMAIN);
	BOOL p;
	BOOL valid_query = False;

	if (pol == NULL || dom_name == NULL || dom_sid == NULL) return False;

	INIT_BASIC_VARS();

	DEBUG(4,("SAMR Query Lookup Domain.\n"));

	/* store the parameters */
	make_samr_q_lookup_domain(&q_o, pol, dom_name);

	/* turn parameters into data stream */
	if (samr_io_q_lookup_domain("", &q_o, &data, 0) &&
	    rpc_hnd_pipe_req(pol, SAMR_LOOKUP_DOMAIN, &data, &rdata))
	{
		samr_io_r_lookup_domain("", &r_o, &rdata, 0);
		p = rdata.offset != 0;

		if (p && r_o.status != 0)
		{
			report_error_code("SAMR_LOOKUP_DOMAIN", r_o.status);
			p = False;
		}

		if (p && r_o.ptr_sid != 0)
		{
			sid_copy(dom_sid, &r_o.dom_sid.sid);
			valid_query = True;
		}
	}

	clean_and_ret(valid_query);
}

/****************************************************************************
do a SAMR Query Lookup Names
****************************************************************************/
BOOL samr_query_lookup_names(const POLICY_HND *pol, uint32 flags,
			     uint32 num_names, char **names,
			     uint32 *num_rids, uint32 **rids, uint32 **types)
{
	DECLARE_BASIC_VARS(SAMR_Q_LOOKUP_NAMES, SAMR_R_LOOKUP_NAMES);
	BOOL p;
	BOOL valid_query = False;

	if (pol == NULL || flags == 0 || num_names == 0 || names == NULL ||
	    num_rids == NULL || rids == NULL || types == NULL ) return False;

	*num_rids = 0;
	*types = NULL;
	*rids  = NULL;

	INIT_BASIC_VARS();
	ZERO_STRUCT(r_o);

	DEBUG(4,("SAMR Query Lookup NAMES.\n"));

	/* store the parameters */
	make_samr_q_lookup_names(&q_o, pol, flags, num_names, names);

	/* turn parameters into data stream */
	if (samr_io_q_lookup_names("", &q_o, &data, 0) &&
	    rpc_hnd_pipe_req(pol, SAMR_LOOKUP_NAMES, &data, &rdata))
	{
		samr_io_r_lookup_names("", &r_o, &rdata, 0);
		p = rdata.offset != 0;

		if (p && r_o.status != 0)
		{
			report_error_code("SAMR_LOOKUP_NAMES", r_o.status);
			p = r_o.status == 0x107;
		}

		if (p)
		{
			if (r_o.ptr_rids != 0 && r_o.ptr_types != 0 &&
			    r_o.num_types1 == r_o.num_rids1)
			{
				uint32 i;

				valid_query = True;
				*num_rids = r_o.num_rids1;
				*types = g_new(uint32, *num_rids);
				*rids  = g_new(uint32, *num_rids);

				for (i = 0; i < r_o.num_rids1; i++)
				{
					(*rids)[i] = r_o.rids[i];
				}
				for (i = 0; i < r_o.num_types1; i++)
				{
					(*types)[i] = r_o.types[i];
				}
			}
			else if (r_o.ptr_rids == 0 && r_o.ptr_types == 0)
			{
				valid_query = True;
				*num_rids = 0;
			}
			else
			{
				p = False;
			}
		}

		samr_free_r_lookup_names(&r_o);
	}

	clean_and_ret(valid_query);
}

/****************************************************************************
do a SAMR Query Lookup RIDS
****************************************************************************/
BOOL samr_query_lookup_rids(  const POLICY_HND *pol, uint32 flags,
				uint32 num_rids, const uint32 *rids,
				uint32 *num_names,
				char   ***names,
				uint32 **type)
{
	DECLARE_BASIC_VARS(SAMR_Q_LOOKUP_RIDS, SAMR_R_LOOKUP_RIDS);
	BOOL p;
	BOOL valid_query = False;

	if (pol == NULL || flags == 0 || num_rids == 0 || rids == NULL ||
	    num_names == NULL || names == NULL || type == NULL)
		return False;

	INIT_BASIC_VARS();
	ZERO_STRUCT(r_o);

	DEBUG(4,("SAMR Query Lookup RIDs.\n"));

	/* store the parameters */
	make_samr_q_lookup_rids(&q_o, pol, flags, num_rids, rids);

	/* turn parameters into data stream */
	if (samr_io_q_lookup_rids("", &q_o,  &data, 0) &&
	    rpc_hnd_pipe_req(pol, SAMR_LOOKUP_RIDS, &data, &rdata))
	{
		samr_io_r_lookup_rids("", &r_o, &rdata, 0);
		p = rdata.offset != 0;

		if (p && r_o.status != 0)
		{
			report_error_code("SAMR_LOOKUP_RIDS", r_o.status);
			p = False;
		}

		if (p)
		{
			if (r_o.ptr_names != 0 && r_o.ptr_types != 0 &&
			    r_o.num_types1 == r_o.num_names1)
			{
				uint32 i;
				valid_query = True;

				(*num_names) = 0;
				(*names) = NULL;

				for (i = 0; i < r_o.num_names1; i++)
				{
					fstring tmp;
					unistr2_to_ascii(tmp, &r_o.uni_name[i], sizeof(tmp)-1);
					add_chars_to_array(num_names, names, tmp);
				}

				if ((*num_names) != 0)
				{
					(*type) = g_new(uint32, *num_names);
				}

				for (i = 0; (*type) != NULL && i < r_o.num_types1; i++)
				{
					(*type)[i] = r_o.type[i];
				}
			}
			else if (r_o.ptr_names == 0 && r_o.ptr_types == 0)
			{
				valid_query = True;
				*num_names = 0;
				*names = NULL;
				*type = NULL;
			}
			else
			{
				p = False;
			}
		}

		samr_free_r_lookup_rids(&r_o);
	}

	clean_and_ret(valid_query);
}

/****************************************************************************
do a SAMR Query Alias Members
****************************************************************************/
BOOL samr_query_aliasmem(  const POLICY_HND *alias_pol, 
				uint32 *num_mem, DOM_SID2 *sid)
{
	DECLARE_BASIC_VARS(SAMR_Q_QUERY_ALIASMEM, SAMR_R_QUERY_ALIASMEM);
	BOOL p;
	BOOL valid_query = False;

	DEBUG(4,("SAMR Query Alias Members.\n"));

	if (alias_pol == NULL || sid == NULL || num_mem == NULL) return False;

	INIT_BASIC_VARS();

	/* store the parameters */
	make_samr_q_query_aliasmem(&q_o, alias_pol);

	/* turn parameters into data stream */
	if (samr_io_q_query_aliasmem("", &q_o,  &data, 0) &&
	    rpc_hnd_pipe_req(alias_pol, SAMR_QUERY_ALIASMEM, &data, &rdata))
	{
		/* get user info */
		r_o.sid = sid;

		samr_io_r_query_aliasmem("", &r_o, &rdata, 0);
		p = rdata.offset != 0;

		if (p && r_o.status != 0)
		{
			report_error_code("SAMR_QUERY_ALIASMEM", r_o.status);
			p = False;
		}

		if (p)
		{
			valid_query = True;
			*num_mem = r_o.num_sids;
		}

	}

	clean_and_ret(valid_query);
}

/****************************************************************************
do a SAMR Query User Aliases
****************************************************************************/
BOOL samr_query_useraliases(  const POLICY_HND *pol,
				uint32 *ptr_sid, DOM_SID2 *sid,
				uint32 *num_aliases, uint32 **rid)
{
	DECLARE_BASIC_VARS(SAMR_Q_QUERY_USERALIASES, SAMR_R_QUERY_USERALIASES);
	BOOL p;
	BOOL valid_query = False;

	DEBUG(4,("SAMR Query User Aliases.\n"));

	if (pol == NULL || sid == NULL || rid == NULL || num_aliases == 0)
		return False;

	INIT_BASIC_VARS();
	ZERO_STRUCT(q_o);

	/* store the parameters */
	make_samr_q_query_useraliases(&q_o, pol, 1, ptr_sid, sid);

	/* turn parameters into data stream */
	if (samr_io_q_query_useraliases("", &q_o,  &data, 0) &&
	    rpc_hnd_pipe_req(pol, SAMR_QUERY_USERALIASES, &data, &rdata))
	{
		r_o.rid = NULL;

		samr_io_r_query_useraliases("", &r_o, &rdata, 0);
		*rid = r_o.rid;
		p = rdata.offset != 0;

		if (p && r_o.status != 0)
		{
			report_error_code("SAMR_QUERY_USERALIASES", r_o.status);
			p = False;
		}

		if (p)
		{
			valid_query = True;
			*num_aliases = r_o.num_entries;
		}

	}

	clean_and_ret(valid_query);
}

/****************************************************************************
do a SAMR Query Group Members
****************************************************************************/
BOOL samr_query_groupmem(  POLICY_HND *group_pol, 
				uint32 *num_mem, uint32 **rid, uint32 **attr)
{
	DECLARE_BASIC_VARS(SAMR_Q_QUERY_GROUPMEM, SAMR_R_QUERY_GROUPMEM);
	BOOL p;
	BOOL valid_query = False;

	DEBUG(4,("SAMR Query Group Members.\n"));

	if (group_pol == NULL || rid == NULL || attr == NULL || num_mem == NULL)
		return False;

	INIT_BASIC_VARS();
	ZERO_STRUCT(r_o);

	/* store the parameters */
	make_samr_q_query_groupmem(&q_o, group_pol);

	/* turn parameters into data stream */
	if (samr_io_q_query_groupmem("", &q_o,  &data, 0) &&
	    rpc_hnd_pipe_req(group_pol, SAMR_QUERY_GROUPMEM, &data, &rdata))
	{
		samr_io_r_query_groupmem("", &r_o, &rdata, 0);
		*rid  = r_o.rid ;
		*attr = r_o.attr;
		p = rdata.offset != 0;

		if (p && r_o.status != 0)
		{
			report_error_code("SAMR_QUERY_GROUPMEM", r_o.status);
			p = False;
		}

		if (p && 
		    ((r_o.ptr_rids != 0 && r_o.ptr_attrs != 0) ||
		     (r_o.ptr_rids == 0 && r_o.ptr_attrs == 0)) &&
		      r_o.num_rids == r_o.num_attrs)
		{
			valid_query = True;
			*num_mem = r_o.num_rids;
		}

	}

	clean_and_ret(valid_query);
}

/****************************************************************************
do a SAMR Query User Groups
****************************************************************************/
BOOL samr_query_usergroups(  POLICY_HND *pol, uint32 *num_groups,
				DOM_GID **gid)
{
	DECLARE_BASIC_VARS(SAMR_Q_QUERY_USERGROUPS, SAMR_R_QUERY_USERGROUPS);
	BOOL p;
	BOOL valid_query = False;

	DEBUG(4,("SAMR Query User Groups.\n"));

	if (pol == NULL || gid == NULL || num_groups == 0)
		return False;

	INIT_BASIC_VARS();
	ZERO_STRUCT(r_o);

	/* store the parameters */
	make_samr_q_query_usergroups(&q_o, pol);

	/* turn parameters into data stream */
	if (samr_io_q_query_usergroups("", &q_o,  &data, 0) &&
	    rpc_hnd_pipe_req(pol, SAMR_QUERY_USERGROUPS, &data, &rdata))
	{
		/* get user info */
		r_o.gid = NULL;

		samr_io_r_query_usergroups("", &r_o, &rdata, 0);
		*gid = r_o.gid;
		p = rdata.offset != 0;

		if (p && r_o.status != 0)
		{
			report_error_code("SAMR_QUERY_USERGROUPS", r_o.status);
			p = False;
		}

		if (p && r_o.ptr_0 != 0)
		{
			valid_query = True;
			*num_groups = r_o.num_entries;
		}
	}

	clean_and_ret(valid_query);
}

/****************************************************************************
do a SAMR Query Group Info
****************************************************************************/
BOOL samr_query_groupinfo(  POLICY_HND *pol,
				uint16 switch_value, GROUP_INFO_CTR* ctr)
{
	DECLARE_BASIC_VARS(SAMR_Q_QUERY_GROUPINFO, SAMR_R_QUERY_GROUPINFO);
	BOOL p;
	BOOL valid_query = False;

	DEBUG(4,("SAMR Query Group Info.  level: %d\n", switch_value));

	if (pol == NULL || ctr == NULL || switch_value == 0)
		return False;

	INIT_BASIC_VARS();
	ZERO_STRUCT(r_o);
	/* get group info */
	r_o.ctr = ctr;

	/* store the parameters */
	make_samr_q_query_groupinfo(&q_o, pol, switch_value);

	/* turn parameters into data stream */
	if (samr_io_q_query_groupinfo("", &q_o,  &data, 0)
	    && rpc_hnd_pipe_req(pol, SAMR_QUERY_GROUPINFO, &data, &rdata)
	    && samr_io_r_query_groupinfo("", &r_o, &rdata, 0))
	{
		p = rdata.offset != 0;

		if (p && r_o.status != 0)
		{
			report_error_code("SAMR_QUERY_GROUPINFO", r_o.status);
			p = False;
		}

		if (p && r_o.ctr->switch_value1 != switch_value)
		{
			DEBUG(4,("SAMR_QUERY_GROUPINFO: received incorrect level %d\n",
			          r_o.ctr->switch_value1));
		}

		if (p && r_o.ptr != 0)
		{
			valid_query = True;
		}
	}

	clean_and_ret(valid_query);
}

/****************************************************************************
do a SAMR Set User Info
****************************************************************************/
BOOL samr_set_userinfo2(POLICY_HND *pol, uint16 switch_value, void *usr)
{
	DECLARE_BASIC_VARS(SAMR_Q_SET_USERINFO2, SAMR_R_SET_USERINFO2);
	BOOL p;
	SAM_USERINFO_CTR ctr;
	BOOL valid_query = False;

	ctr.info.id = usr;

	DEBUG(4,("SAMR Set User Info 2.  level: %d\n", switch_value));

	if (pol == NULL || usr == NULL || switch_value == 0)
		return False;

	INIT_BASIC_VARS();
	ZERO_STRUCT(r_o);

	/* store the parameters */
	make_samr_q_set_userinfo2(&q_o, pol, switch_value, &ctr);

	/* turn parameters into data stream */
	if (samr_io_q_set_userinfo2("", &q_o,  &data, 0) &&
	    rpc_hnd_pipe_req(pol, SAMR_SET_USERINFO2, &data, &rdata))
	{
		samr_io_r_set_userinfo2("", &r_o, &rdata, 0);
		p = rdata.offset != 0;

		if (p && r_o.status != 0)
		{
			report_error_code("SAMR_SET_USERINFO2", r_o.status);
			p = False;
		}

		if (p)
		{
			valid_query = True;
		}
	}

	free_samr_q_set_userinfo2(&q_o);
	clean_and_ret(valid_query);
}

/****************************************************************************
do a SAMR Set User Info
****************************************************************************/
BOOL samr_set_userinfo(POLICY_HND *pol, uint16 switch_value, void *usr)
{
	DECLARE_BASIC_VARS(SAMR_Q_SET_USERINFO, SAMR_R_SET_USERINFO);
	BOOL p;
	SAM_USERINFO_CTR ctr;
	BOOL valid_query = False;

	DEBUG(4,("SAMR Set User Info.  level: %d\n", switch_value));

	if (pol == NULL || usr == NULL || switch_value == 0)
		return False;

	INIT_BASIC_VARS();
	ZERO_STRUCT(r_o);

	q_o.ctr = &ctr;

	/* store the parameters */
	make_samr_q_set_userinfo(&q_o, pol, switch_value, usr);

	/* turn parameters into data stream */
	if (samr_io_q_set_userinfo("", &q_o,  &data, 0) &&
	    rpc_hnd_pipe_req(pol, SAMR_SET_USERINFO, &data, &rdata))
	{
		samr_io_r_set_userinfo("", &r_o, &rdata, 0);
		p = rdata.offset != 0;

		if (p && r_o.status != 0)
		{
			report_error_code("SAMR_SET_USERINFO", r_o.status);
			p = False;
		}

		if (p)
		{
			valid_query = True;
		}
	}

	clean_and_ret(valid_query);
}

/****************************************************************************
do a SAMR Query User Info
****************************************************************************/
BOOL samr_query_userinfo(  POLICY_HND *pol, uint16 switch_value,
				SAM_USERINFO_CTR *ctr)
{
	DECLARE_BASIC_VARS(SAMR_Q_QUERY_USERINFO, SAMR_R_QUERY_USERINFO);
	BOOL p;
	BOOL valid_query = False;

	DEBUG(4,("SAMR Query User Info.  level: %d\n", switch_value));

	if (pol == NULL || ctr == NULL || switch_value == 0)
		return False;

	INIT_BASIC_VARS();
	ZERO_STRUCT(r_o);

	/* store the parameters */
	make_samr_q_query_userinfo(&q_o, pol, switch_value);

	/* turn parameters into data stream */
	if (samr_io_q_query_userinfo("", &q_o,  &data, 0) &&
	    rpc_hnd_pipe_req(pol, SAMR_QUERY_USERINFO, &data, &rdata))
	{
		r_o.ctr = ctr;

		p = samr_io_r_query_userinfo("", &r_o, &rdata, 0)
			&& rdata.offset != 0;

		if (p && r_o.status != 0)
		{
			report_error_code("SAMR_QUERY_USERINFO", r_o.status);
			p = False;
		}

		if (p && r_o.ptr == 0)
		{
			p = False;
		}

		if (p && r_o.ctr->switch_value != switch_value)
		{
			DEBUG(4,("SAMR_QUERY_USERINFO: received incorrect level %d\n",
			          r_o.ctr->switch_value));
		}

		if (p && r_o.ptr != 0)
		{
			valid_query = True;
		}
	}

	clean_and_ret(valid_query);
}

/****************************************************************************
do a SAMR Close
****************************************************************************/
BOOL samr_close(  POLICY_HND *hnd)
{
	DECLARE_BASIC_VARS(SAMR_Q_CLOSE_HND, SAMR_R_CLOSE_HND);
	BOOL p;
	BOOL valid_close = False;

	DEBUG(4,("SAMR Close\n"));

	if (hnd == NULL)
		return False;

	INIT_BASIC_VARS();
	ZERO_STRUCT(r_o);

	/* store the parameters */
	make_samr_q_close_hnd(&q_o, hnd);

	/* turn parameters into data stream */
	if (samr_io_q_close_hnd("", &q_o,  &data, 0) &&
	    rpc_hnd_pipe_req(hnd, SAMR_CLOSE_HND, &data, &rdata))
	{
		samr_io_r_close_hnd("", &r_o, &rdata, 0);
		p = rdata.offset != 0;

		if (p && r_o.status != 0)
		{
			report_error_code("SAMR_CLOSE_HND", r_o.status);
			p = False;
		}

		if (p)
		{
			valid_close = True;
		}
	}

	close_policy_hnd(get_global_hnd_cache(), hnd);

	clean_and_ret(valid_close);
}

/****************************************************************************
do a SAMR query display info
****************************************************************************/
uint32 samr_query_dispinfo(POLICY_HND *pol_domain, uint32 *start_idx,
			   uint16 level, uint32 *num_entries, 
			   SAM_DISPINFO_CTR *ctr)
{
	DECLARE_BASIC_VARS(SAMR_Q_QUERY_DISPINFO, SAMR_R_QUERY_DISPINFO);
	BOOL p;
	uint32 status = 0;

	DEBUG(4,("SAMR Query Display Info.  level: %d\n", level));

	if (pol_domain == NULL || num_entries == NULL || ctr == NULL ||
	    level == 0 || start_idx == 0)
	{
		return NT_STATUS_INVALID_PARAMETER;
	}

	INIT_BASIC_VARS();
	ZERO_STRUCT(r_o);

	/* store the parameters */
	make_samr_q_query_dispinfo(&q_o, pol_domain, level, *start_idx, 
				   MAX_SAM_ENTRIES);

	/* turn parameters into data stream */
	if (samr_io_q_query_dispinfo("", &q_o,  &data, 0) &&
	    rpc_hnd_pipe_req(pol_domain, SAMR_QUERY_DISPINFO, &data, &rdata))
	{
		/* get user info */
		r_o.ctr = ctr;

		samr_io_r_query_dispinfo("", &r_o, &rdata, 0);
		p = rdata.offset != 0;
		status = r_o.status;

		if (p && r_o.status != 0)
		{
			report_error_code("SAMR_QUERY_DISPINFO", r_o.status);
			p = (r_o.status == STATUS_MORE_ENTRIES);
		}

		if (p && r_o.switch_level != level)
		{
			DEBUG(4,("SAMR_QUERY_DISPINFO: received incorrect level %d\n",
			          r_o.switch_level));
		}

		if (p && r_o.ptr_entries != 0)
		{
			(*num_entries) = r_o.num_entries;
			(*start_idx) += r_o.num_entries; /* No next_idx */
		}
	}

	clean_and_ret(status);
}
