/* 
   Unix SMB/Netbios implementation.
   Version 1.9.
   SMB client generic functions
   Copyright (C) Luke Kenneth Casson Leighton 1996-2001
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/
/* MKPROTO: rpc_client_proto.h		(do not remove this line) */

#include "includes.h"
#include "libsamba.h"
#include "tng_misc.h"
#include "libsmb.h"
#include "rpc_parse.h"
#include "rpc_client.h"
#include "others.h"

#undef  DBGC_CLASS
#define DBGC_CLASS DBGC_RPC


struct ncacn_np_use
{
	struct ncacn_np *cli;
	uint32 num_users;
};

static struct ncacn_np_use **msrpcs = NULL;
static uint32 num_msrpcs = 0;

/****************************************************************************
terminate client connection
****************************************************************************/
static void ncacn_np_shutdown(struct ncacn_np *cli)
{
	if (cli != NULL)
	{
		if (cli->smb != NULL)
		{
			if (cli->smb->initialised)
			{
				cli_nt_session_close(cli->smb, cli->fnum);
			}
			cli_net_use_del(cli->smb->desthost,
					&cli->smb->usr, False, False);
		}
	}
}

static BOOL ncacn_np_establish_connection(struct ncacn_np *cli,
				   const char *srv_name,
				   const struct ntuser_creds *ntc,
				   const char *pipe_name, 
				   BOOL reuse)
{
	BOOL new_smb_conn;
	cli->smb = cli_net_use_add(srv_name, ntc,
				   True, &new_smb_conn, 0);
	if (cli->smb == NULL)
	{
		DEBUG(0, ("ncacn_np_establish_connection(%s, %s, %s): "
			  "cli_net_use_add failed\n",
			  srv_name, pipe_name, BOOLSTR(reuse)));
		return False;
	}
	if (!cli_nt_session_open(cli->smb, pipe_name, &cli->fnum))
	{
		cli_net_use_del(srv_name, ntc, False, NULL);
		cli->smb = NULL;
		return False;
	}
	fstrcpy(cli->pipe_name, pipe_name);
	dump_data_pw("sess key:", cli->smb->nt.usr_sess_key, 16);
	return True;
}

#if 0

/* Useful for debugging rpc caching */

static void dump_msrpcs(void)
{
	int i;

	for (i = 0; i < num_msrpcs; i++) {
		if (msrpcs[i]) {
			DEBUG(3, ("[%d]: %s %s num_users=%d np_use=0x%08x "
				  " np=0x%08x smb=0x%08x\n",
				  i, msrpcs[i]->cli->smb->desthost,
				  msrpcs[i]->cli->pipe_name,
				  msrpcs[i]->num_users, 
				  msrpcs[i],
				  msrpcs[i]->cli,
				  msrpcs[i]->cli->smb));
		} else {
			DEBUG(3, ("[%d]: NULL\n", i));
		}
	}
}

#endif

/****************************************************************************
terminate client connection
****************************************************************************/
static void ncacn_np_use_free(struct ncacn_np_use *cli)
{
	if (cli->cli != NULL)
	{
		int i;

		/* Check that another connection doesn't have this same smb
		   open. */

		for (i = 0; i < num_msrpcs; i++) 
		{
			if (msrpcs[i] && msrpcs[i] != cli &&
			    msrpcs[i]->cli && 
			    msrpcs[i]->cli->smb == cli->cli->smb) {
				goto no_shutdown;
			}
		}

		if (cli->cli->initialised)
		{
			ncacn_np_shutdown(cli->cli);
		}
	no_shutdown:

		ZERO_STRUCTP(cli->cli);
		free(cli->cli);
	}

	ZERO_STRUCTP(cli);
	free(cli);
}

/****************************************************************************
free a client array
****************************************************************************/
#ifndef NOONS_SMALL_BUILD
static void free_ncacn_np_array(uint32 num_entries,
				struct ncacn_np_use **entries)
{
	void (*fn) (void *) = (void (*)(void *))&ncacn_np_use_free;
	free_void_array(num_entries, (void **)entries, *fn);
}
#endif
/****************************************************************************
add a client state to the array
****************************************************************************/
static struct ncacn_np_use *add_ncacn_np_to_array(uint32 * len,
						  struct ncacn_np_use
						  ***array,
						  struct ncacn_np_use *cli)
{
	int i;
	for (i = 0; i < num_msrpcs; i++)
	{
		if (msrpcs[i] == NULL)
		{
			msrpcs[i] = cli;
			return cli;
		}
	}

	return (struct ncacn_np_use *)add_item_to_array(len,
							(void ***)array,
							(void *)cli);

}

/****************************************************************************
initiate client array
****************************************************************************/
#ifndef NOONS_SMALL_BUILD
static void init_ncacn_np_use(void)
{
	msrpcs = NULL;
	num_msrpcs = 0;
}
#endif
/****************************************************************************
terminate client array
****************************************************************************/
#ifndef NOONS_SMALL_BUILD
static void free_ncacn_np_use(void)
{
	free_ncacn_np_array(num_msrpcs, msrpcs);
	init_ncacn_np_use();
}
#endif
/****************************************************************************
find client state.  server name, user name, domain name and password must all
match.
****************************************************************************/
static struct ncacn_np_use *ncacn_np_find(const char *srv_name,
					  const char *pipe_name,
					  uint16 vuid,
					  const struct ntuser_creds
					  *usr_creds, BOOL reuse)
{
	int i;
	const char *sv_name = srv_name;

	if (strnequal("\\PIPE\\", pipe_name, 6))
	{
		pipe_name = &pipe_name[6];
	}

	if (strnequal("\\\\", sv_name, 2))
	{
		sv_name = &sv_name[2];
	}

	if (usr_creds != NULL)
	{
		DEBUG(10, ("ncacn_np_find: %s %s %s",
			   srv_name, usr_creds->user_name, usr_creds->domain));
	}
	else
	{
		DEBUG(10,("ncacn_np_find: %s (no creds)\n", srv_name));
	}

	DEBUG(10, ("[%x]\n", vuid));

	for (i = 0; i < num_msrpcs; i++)
	{
		char *ncacn_np_srv_name = NULL;
		struct ncacn_np_use *c = msrpcs[i];
		uint16 k;

		char *ncacn_np_name = NULL;

		if (c == NULL || c->cli == NULL || c->cli->smb == NULL ||
		    c->cli->smb->fd == -1 ||
		    !c->cli->initialised)
		{
			continue;
		}

		ncacn_np_name = c->cli->pipe_name;
		ncacn_np_srv_name = c->cli->smb->desthost;

		k = c->cli->smb->nt.vuid;

		DEBUG(10, ("ncacn_np_find[%d]: %s %s %s %s [%x]\n",
			   i, ncacn_np_name, ncacn_np_srv_name,
			   c->cli->smb->usr.user_name,
			   c->cli->smb->usr.domain, k));

		if (strnequal("\\\\", ncacn_np_srv_name, 2))
		{
			ncacn_np_srv_name = &ncacn_np_srv_name[2];
		}

		if (strnequal("\\PIPE\\", ncacn_np_name, 6))
		{
			ncacn_np_name = &ncacn_np_name[6];
		}

		if (!strequal(ncacn_np_name, pipe_name))
		{
			continue;
		}
		if (!strequal(ncacn_np_srv_name, sv_name))
		{
			continue;
		}
		if (k != vuid)
		{
			continue;
		}
		if (usr_creds == NULL)
		{
			if (reuse)
			{
				return c;
			}
			else
			{
				continue;
			}
		}
		if (!strequal
		    (usr_creds->user_name, c->cli->smb->usr.user_name))
		{
			continue;
		}
		if (!reuse
		    && !pwd_compare(&usr_creds->pwd, &c->cli->smb->usr.pwd))
		{
			DEBUG(100, ("password doesn't match\n"));
			continue;
		}
		if (usr_creds->domain[0] == 0)
		{
			return c;
		}
		if (strequal(usr_creds->domain, c->cli->smb->usr.domain))
		{
			return c;
		}
	}

	return NULL;
}

/****************************************************************************
initialise a msrpcent structure
****************************************************************************/
static struct ncacn_np *ncacn_np_initialise(uint16 vuid)
{
	struct ncacn_np *msrpc;

	msrpc = g_new(struct ncacn_np, 1);
	if (!msrpc)
		return NULL;

	ZERO_STRUCTP(msrpc);

	msrpc->fnum = -1;
	msrpc->initialised = 1;

	return msrpc;
}

/****************************************************************************
create a new client state from user credentials
****************************************************************************/
static struct ncacn_np_use *ncacn_np_use_get(const char *pipe_name,
					     uint16 vuid)
{
	struct ncacn_np_use *cli = g_new0(struct ncacn_np_use, 1);

	if (cli == NULL)
	{
		return NULL;
	}

	cli->cli = ncacn_np_initialise(vuid);

	if (cli->cli == NULL)
	{
		free(cli);
		return NULL;
	}

	return cli;
}

/****************************************************************************
init client state
****************************************************************************/
struct ncacn_np *ncacn_np_use_add(const char *pipe_name,
				  uint16 vuid,
				  const char *srv_name,
				  const struct ntuser_creds *ntc,
				  BOOL reuse, BOOL *is_new_connection)
{
	struct ncacn_np_use *cli;

	DEBUG(10, ("ncacn_np_use_add: %s:%s\n", srv_name, pipe_name));

	(*is_new_connection) = False;
	cli = ncacn_np_find(srv_name, pipe_name, vuid, ntc, reuse);

	if (cli != NULL)
	{
		cli->num_users++;
		DEBUG(10, ("ncacn_np_use_add: num_users: %d\n",
			   cli->num_users));
		return cli->cli;
	}

	(*is_new_connection) = True;

	cli = ncacn_np_use_get(pipe_name, vuid);

	if (!ncacn_np_establish_connection(cli->cli, srv_name,
					   ntc, pipe_name, True))
	{
		DEBUG(1, ("ncacn_np_use_add: connection to %s:%s failed\n",
			  srv_name, pipe_name));
		ncacn_np_use_free(cli);
		return NULL;
	}

	cli->cli->smb->nt.vuid = vuid;

	if (vuid == UID_FIELD_INVALID)
	{
		NET_USER_INFO_3 usr;
		uid_t uid = getuid();
		gid_t gid = getgid();
		char *name = uidtoname(uid);

		DEBUG(3, ("ncacn_np_use_add: No vuid, allocating one "
			  "for the current user\n"));

		ZERO_STRUCT(usr);

		cli->cli->smb->nt.vuid =
			register_vuid(uid, gid,
				      name, name,
				      False, 0,
				      &usr);
	}

	add_ncacn_np_to_array(&num_msrpcs, &msrpcs, cli);
	cli->num_users++;

	return cli->cli;
}

/****************************************************************************
delete a client state
****************************************************************************/
BOOL ncacn_np_use_del(const char *srv_name, const char *pipe_name,
		      uint16 vuid,
		      BOOL force_close, BOOL *connection_closed)
{
	int i;
	DEBUG(10, ("ncacn_np_net_use_del: %s, %s. force close: %s ",
		   srv_name, pipe_name, BOOLSTR(force_close)));
	DEBUG(3, ("[%x]\n", vuid));

	if (connection_closed != NULL)
	{
		*connection_closed = False;
	}

	if (strnequal("\\PIPE\\", pipe_name, 6))
	{
		pipe_name = &pipe_name[6];
	}

	if (strnequal("\\\\", srv_name, 2))
	{
		srv_name = &srv_name[2];
	}

	for (i = 0; i < num_msrpcs; i++)
	{
		char *ncacn_np_name = NULL;
		char *ncacn_np_srv_name = NULL;
		struct ncacn_np_use *c = msrpcs[i];
		uint16 k;

		if (c == NULL || c->cli == NULL || c->cli->smb == NULL)
			continue;

		ncacn_np_name = c->cli->pipe_name;
		ncacn_np_srv_name = c->cli->smb->desthost;

		k = c->cli->smb->nt.vuid;

		DEBUG(10, ("use_del[%d]: %s %s %s %s [%x]\n",
			   i, ncacn_np_name, ncacn_np_srv_name,
			   c->cli->smb->usr.user_name,
			   c->cli->smb->usr.domain, k));

		if (strnequal("\\PIPE\\", ncacn_np_name, 6))
		{
			ncacn_np_name = &ncacn_np_name[6];
		}
		if (!strequal(ncacn_np_name, pipe_name))
		{
			continue;
		}
		if (strnequal("\\\\", ncacn_np_srv_name, 2))
		{
			ncacn_np_srv_name = &ncacn_np_srv_name[2];
		}
		if (!strequal(ncacn_np_srv_name, srv_name))
		{
			continue;
		}
		if (vuid != k)
		{
			continue;
		}
		/* decrement number of users */
		c->num_users--;
		DEBUG(3, ("idx: %i num_users now: %d\n",
			   i, c->num_users));
		if (force_close || c->num_users == 0)
		{
			ncacn_np_use_free(c);
			msrpcs[i] = NULL;
			if (connection_closed != NULL)
			{
				*connection_closed = True;
			}
		}
		return True;
	}

	return False;
}


void ncacn_np_use_close(struct ncacn_np *np_con)
{
	unsigned int i;

	for (i = 0; i < num_msrpcs; i++)
	{
		struct ncacn_np_use *c = msrpcs[i];

		if ((! c) || (c->cli != np_con))
			continue;
		/* decrement number of users */
		c->num_users--;

		DEBUG(5, ("ncacn_np_use_close: Connection to %s:%s has "
			  "%d users now, if zero going to close\n",
			  np_con->smb->desthost, np_con->pipe_name, c->num_users));
		if (c->num_users == 0)
		{
			ncacn_np_use_free(c);
			msrpcs[i] = NULL;
		}
		return;
	}
}

/****************************************************************************
enumerate client states
****************************************************************************/
#ifndef NOONS_SMALL_BUILD
static void ncacn_np_use_enum(uint32 * num_cons, struct use_info ***use)
{
	int i;
	*num_cons = 0;
	*use = NULL;
	for (i = 0; i < num_msrpcs; i++)
	{
		struct use_info item;
		ZERO_STRUCT(item);
		if (msrpcs[i] == NULL)
			continue;
		item.connected = msrpcs[i]->cli != NULL ? True : False;
		if (item.connected)
		{
			item.srv_name = msrpcs[i]->cli->pipe_name;
			item.vuid = msrpcs[i]->cli->smb->nt.vuid;
		}

		add_use_info_to_array(num_cons, use, &item);
	}
}
#endif
