/* 
 *  Unix SMB/Netbios implementation.
 *  Version 1.9.
 *  RPC Pipe server routines
 *  Copyright (C) Luke Kenneth Casson Leighton 1996-2001
 *  
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *  
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software Foundation,
 *  Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "includes.h"
#include "libsamba.h"
#include "tng_misc.h"
#include "byteorder.h"
#include "rpc_parse.h"
#include "others.h"


/*******************************************************************
turns a DCE/RPC request into a DCE/RPC reply

this is where the data really should be split up into an array of
headers and data sections.

********************************************************************/
static BOOL api_netsec_create_pdu(rpcsrv_struct * l, uint32 data_start,
				  prs_struct *resp)
{
	netsec_auth_struct *a = (netsec_auth_struct *) l->auth_info;

	BOOL ret;
	uint32 data_len;
	uint32 frag_len;
	uint32 auth_len;
	uint32 data_end = l->rdata.offset + 8 + 0x20;
	char *data;
	uint8 flags;
	uint32 call_id = l->hdr.call_id;

	prs_struct rhdr;
	prs_struct rauth;
	prs_struct rverf;

	RPC_HDR_RESP hdr_resp;
	RPC_HDR_AUTH auth_info;
	RPC_AUTH_NETSEC_CHK verf;
	uchar sign[8];
	static const uchar netsec_sig[8] = NETSEC_SIGNATURE;

	DEBUG(5,
	      ("api_netsec_create_pdu: data_start: %d data_end: %d max_tsize: %d\n",
	       data_start, data_end, l->hdr_ba.bba.max_tsize));

	auth_len = l->hdr.auth_len;

	DEBUG(10, ("api_netsec_create_pdu: auth\n"));

	if (auth_len != 0x20)
	{
		return False;
	}

	prs_init(&rhdr, MARSHALL);

	/* set up rpc header (fragmentation issues) */
	if (data_start == 0)
	{
		flags = RPC_FLG_FIRST;
	}
	else
	{
		flags = 0;
	}

	hdr_resp.alloc_hint = data_end - data_start;	/* calculate remaining data to be sent */
	hdr_resp.cancel_count = 0x0;
	hdr_resp.context_id = l->hdr_req.context_id;
	hdr_resp.reserved = 0x0;

	DEBUG(10, ("alloc_hint: %d\n", hdr_resp.alloc_hint));

	if (hdr_resp.alloc_hint + 0x18 <= l->hdr_ba.bba.max_tsize)
	{
		flags |= RPC_FLG_LAST;
		frag_len = hdr_resp.alloc_hint + 0x18;
	}
	else
	{
		frag_len = l->hdr_ba.bba.max_tsize;
	}

	hdr_resp.alloc_hint -= auth_len + 8;

	data_len = frag_len - auth_len - 8 - 0x18;

	prs_set_start(&rhdr, 0);
	rhdr.end = 0x18;

	DEBUG(10, ("hdr flags: %x\n", flags));

	if (!make_rpc_hdr(&l->hdr, RPC_RESPONSE, flags, call_id,
			  frag_len, auth_len))
	{
		return False;
	}

	/* store the header in the data stream */
	smb_io_rpc_hdr("rhdr", &(l->hdr), &rhdr, 0);
	smb_io_rpc_hdr_resp("resp", &(hdr_resp), &rhdr, 0);

	/* find location in data where we start to create the
	 * pdu from (prepend header, append authenticator)
	 */

	data = prs_data(&l->rdata, data_start);
	l->rdata_offset += data_len;

	prs_debug_out(&l->rdata, "rdata", 200);

	/* happen to know that NETSEC authentication verifier is 32 bytes */
	frag_len = data_len + auth_len + 8 + 0x18;

	prs_init(&rauth, MARSHALL);
	prs_init(&rverf, MARSHALL);
	prs_realloc_data(&rauth, 8);
	prs_realloc_data(&rverf, auth_len);

	make_rpc_hdr_auth(&auth_info, 0x44, l->auth_level, 0x0, l->key_id);
	smb_io_rpc_hdr_auth("hdr_auth", &auth_info, &rauth, 0);

	ZERO_OBJ(sign);
	RSIVAL(sign, 0, a->seq_num2);

	make_rpc_auth_netsec_chk(&verf, netsec_sig, NULL, sign, NULL);
	ret = netsec_encode(a, &verf, data, data_len);

	if (ret)
	{
		smb_io_rpc_auth_netsec_chk("verf", &verf, &rverf, 0);
		a->seq_num++;
	}

	if (ret)
	{
		/* create response PDU: prepend header, then data, then auth */
		prs_init(resp, MARSHALL);
		prs_append_prs(resp, &rhdr);
		prs_append_data(resp, data, data_len);
		prs_append_prs(resp, &rauth);
		ret = prs_append_prs(resp, &rverf);
	}

	prs_free_data(&rauth);
	prs_free_data(&rverf);
	prs_free_data(&rhdr);

	if (IS_BITS_SET_ALL(flags, RPC_FLG_LAST))
	{
		DEBUG(10, ("create_netsec_reply: finished sending\n"));
		prs_free_data(&l->rdata);
	}

	return ret;
}

static BOOL api_netsec_verify(rpcsrv_struct * l)
{
	uint16 vuid = get_sec_ctx();

	netsec_auth_struct *a = (netsec_auth_struct *) l->auth_info;
	struct dcinfo dc;

	if (vuid == UID_FIELD_INVALID)
	{
		DEBUG(1, ("api_netsec_verify: no vuid\n"));
		return False;
	}

	DEBUG(3, ("api_netsec_verify: checking credentials for %s\\%s\n",
		  a->netsec_neg.domain, a->netsec_neg.myname));

	/*
	 * obtain the session key
	 */
	if (!cred_get( a->netsec_neg.domain, a->netsec_neg.myname, &dc))
	{
		DEBUG(1, ("api_netsec_verify: No creds for %s\\%s\n",
			  a->netsec_neg.domain, a->netsec_neg.myname));
		return False;
	}

	l->auth_validated = True;

	ZERO_OBJ(a->sess_key);
	memcpy(a->sess_key, dc.sess_key, sizeof(dc.sess_key));

	dump_data_pw("sess_key:\n", a->sess_key, sizeof(a->sess_key));

	if (l->auth_validated)
	{
		a->seq_num = 0;
		a->seq_num2 = 0;
	}
	else
	{
		l->auth_validated = False;
	}

	return l->auth_validated;
}

static BOOL api_netsec(rpcsrv_struct * l, uint32 msg_type)
{
	switch (msg_type)
	{
		case 0x3:
		case 0x13:
		{
			netsec_auth_struct *a;
			a = (netsec_auth_struct *) l->auth_info;

			smb_io_rpc_auth_netsec_neg("", &a->netsec_neg,
						   &l->data_i, 0);
			if (l->data_i.offset == 0)
				return False;

			return api_netsec_verify(l);
		}
		default:
		{
			/* NEGSEC expected: unexpected message type */
			DEBUG(1, ("unexpected message type in NEGSEC %d\n",
				  msg_type));
			return False;
		}
	}

	return False;
}

static BOOL api_netsec_auth_chk(rpcsrv_struct * l, enum RPC_PKT_TYPE pkt_type)
{
	switch (pkt_type)
	{
		case RPC_BINDACK:
		{
			RPC_AUTH_VERIFIER auth_verifier;
			smb_io_rpc_auth_verifier("", &auth_verifier,
						 &l->data_i, 0);
			if (l->data_i.offset == 0)
				return False;

			if (strequal(auth_verifier.signature, ""))
			{
				return api_netsec(l, auth_verifier.msg_type);
			}
			break;
		}
		default:
		{
			DEBUG(1,
			      ("api_netsec_auth_chk: unknown pkt_type %x\n",
			       pkt_type));
			return False;
		}
	}
	return False;
}

static BOOL api_netsec_auth_gen(rpcsrv_struct * l, prs_struct *resp,
				enum RPC_PKT_TYPE pkt_type)
{
	BOOL ret;
	RPC_HDR_AUTH auth_info;
	RPC_AUTH_VERIFIER auth_verifier;
	RPC_AUTH_NETSEC_RESP auth_resp;
	prs_struct rauth;
	prs_struct rverf;
	prs_struct rresp;

	prs_init(resp, MARSHALL);
	prs_init(&(rauth), MARSHALL);
	prs_init(&(rverf), MARSHALL);
	prs_init(&(rresp), MARSHALL);

	/*** authentication info ***/

	make_rpc_hdr_auth(&auth_info, 0x44, l->auth_level, 0, 1);
	smb_io_rpc_hdr_auth("", &auth_info, &rverf, 0);
	prs_realloc_data(&rverf, rverf.offset);

	/*** NETSEC verifier ***/

	make_rpc_auth_verifier(&auth_verifier, "\001", 0x0);
	smb_io_rpc_auth_verifier("", &auth_verifier, &rauth, 0);
	prs_realloc_data(&rauth, rauth.offset);

	/* NETSEC challenge ** */

	make_rpc_auth_netsec_resp(&auth_resp, 0x05);
	smb_io_rpc_auth_netsec_resp("", &auth_resp, &rresp, 0);
	prs_realloc_data(&rresp, rresp.offset);

	/***/
	/*** then do the header, now we know the length ***/
	/***/

	make_rpc_hdr(&l->hdr, pkt_type, RPC_FLG_FIRST | RPC_FLG_LAST,
		     l->hdr.call_id,
		     l->rdata.offset + rverf.offset + rauth.offset +
		     rresp.offset + 0x10, rauth.offset + rresp.offset);

	smb_io_rpc_hdr("", &l->hdr, resp, 0);

	ret = prs_append_prs(resp, &l->rdata)
		&& prs_append_prs(resp, &rverf)
		&& prs_append_prs(resp, &rauth)
		&& prs_append_prs(resp, &rresp);

	prs_free_data(&l->rdata);
	prs_free_data(&rauth);
	prs_free_data(&rverf);
	prs_free_data(&rresp);

	return ret;
}

static BOOL api_netsec_decode_pdu(rpcsrv_struct * l)
{
	netsec_auth_struct *a = (netsec_auth_struct *) l->auth_info;
	int data_len;
	int auth_len;
	uint32 old_offset;
	/* int old_align; */
	RPC_HDR_AUTH auth_info;
	RPC_AUTH_NETSEC_CHK netsec_chk;

	auth_len = l->hdr.auth_len;

	if (auth_len != 0x20)
	{
		return False;
	}

	data_len = l->hdr.frag_len - auth_len - 8 - 0x18;

	DEBUG(5, ("api_pipe_auth_process: data %d auth %d\n",
		  data_len, auth_len));

	/*** skip the data, record the offset so we can restore it again */
	old_offset = l->data_i.offset;

	l->data_i.offset += data_len;
	/* old_align = prs_set_align(&l->data_i.align, 0); */
	smb_io_rpc_hdr_auth("hdr_auth", &auth_info, &l->data_i, 0);
	if (!rpc_hdr_netsec_auth_chk(&(auth_info)))
		return False;

	g_return_val_if_fail(smb_io_rpc_auth_netsec_chk("netsec_chk", &netsec_chk,
							&l->data_i, 0), False);
	/* prs_set_align(&l->data_i, old_align); */

	if (!netsec_decode(a, &netsec_chk,
			   prs_data(&l->data_i, old_offset), data_len))
	{
		DEBUG(1, ("api_netsec_decode_pdu: netsec_decode failed\n"));
		return False;
	}

	/* restore the [data, now decoded] offset */
	l->data_i.offset = old_offset;

	rpc_decode_auth_eat(&l->data_i, &l->hdr, &auth_info, 0, 0, 0, 0);

	return True;
}

static BOOL api_netsec_hdr_chk(RPC_HDR_AUTH * auth_info, void **auth_struct)
{
	DEBUG(10, ("api_netsec_hdr_chk:\n"));
	if (!rpc_hdr_netsec_auth_chk(auth_info))
	{
		return False;
	}
	(*auth_struct) = g_new(netsec_auth_struct, 1);
	return (*auth_struct) != NULL;
}

srv_auth_fns netsec_fns = {
	api_netsec_hdr_chk,
	api_netsec_auth_chk,
	api_netsec_auth_gen,
	api_netsec_decode_pdu,
	api_netsec_create_pdu,
};
