
/* 
 *  Unix SMB/Netbios implementation.
 *  Version 1.9.
 *  RPC Pipe client / server routines
 *  Copyright (C) Luke Kenneth Casson Leighton 1996-2001
 *  Copyright (C) Paul Ashton                  1997.
 *  
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *  
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software Foundation,
 *  Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "includes.h"
#include "libsamba.h"
#include "tng_misc.h"
#include "rpc_parse.h"
#include "others.h"


/*******************************************************************
turns a DCE/RPC response stream into a DCE/RPC reply
********************************************************************/
static BOOL create_rpc_reply(rpcsrv_struct * l, uint32 data_start,
			     prs_struct *resp)
{
	BOOL ret;

	if (l->auth != NULL)
	{
		SMB_ASSERT(l->auth->api_create_pdu != NULL);
		ret = l->auth->api_create_pdu(l, data_start, resp);
		if (!ret)
		{
			DEBUG(1, ("auth->api_create_pdu failed\n"));
		}
	}
	else
	{
		DEBUG(2, ("create_rpc_reply: No auth-system to create pdu\n"));
		ret = False;
	}
	if ((!ret) || IS_BITS_SET_ALL(l->hdr.flags, RPC_FLG_LAST))
	{
		DEBUG(10, ("create_rpc_reply: finished sending\n"));
		prs_free_data(&l->rdata);
	}
	return ret;
}


struct api_cmd
{
	char *pipe_clnt_name;
	char *pipe_srv_name;
	RPC_IFACE *abstract;
	BOOL (*fn) (rpcsrv_struct *);
};

static struct api_cmd **api_fd_commands = NULL;
static size_t num_cmds = 0;

static void api_cmd_free(struct api_cmd *item)
{
	if (!item)
		return;
	safe_free(item->pipe_clnt_name);
	safe_free(item->pipe_srv_name);
	safe_free(item->abstract);
	free(item);
}

static struct api_cmd *api_cmd_dup(const struct api_cmd *from)
{
	struct api_cmd *copy = NULL;
	if (!from)
		return NULL;

	copy = g_new0(struct api_cmd, 1);
	if (!copy)
		return NULL;

	if (from->pipe_clnt_name) {
		copy->pipe_clnt_name = strdup(from->pipe_clnt_name);
		if (!copy->pipe_clnt_name)
			goto oom;
	}
	if (from->pipe_srv_name) {
		copy->pipe_srv_name = strdup(from->pipe_srv_name);
		if (!copy->pipe_srv_name)
			goto oom;
	}
	if (from->fn)
		copy->fn = from->fn;
	if (from->abstract) {
		copy->abstract = g_new1(RPC_IFACE);
		if (!copy->abstract)
			goto oom;
		*copy->abstract = *from->abstract;
	}
	return copy;
    oom:
	DEBUG(0,("api_cmd_dup: out of memory\n"));
	api_cmd_free(copy);
	return NULL;
}

static void free_api_cmd_array(uint32 num_entries, struct api_cmd **entries)
{
	void (*fn) (void *) = (void (*)(void *))&api_cmd_free;
	free_void_array(num_entries, (void **)entries, *fn);
}

static struct api_cmd *add_api_cmd_to_array(uint32 *len,
					    struct api_cmd ***array,
					    const struct api_cmd *name)
{
	void *(*fn) (const void *) = (void *(*)(const void *))&api_cmd_dup;
	return (struct api_cmd *)add_copy_to_array(len,
						   (void ***)array,
						   (const void *)name, *fn,
						   False);
}


void close_msrpc_command_processor(void)
{
	free_api_cmd_array(num_cmds, api_fd_commands);
}

void add_msrpc_command_processor(char *pipe_name,
				 char *process_name,
				 BOOL (*fn) (rpcsrv_struct *),
				 const RPC_IFACE *abstract)
{
	struct api_cmd cmd;
	RPC_IFACE abstract_copy;

	cmd.pipe_clnt_name = pipe_name;
	cmd.pipe_srv_name = process_name;
	cmd.fn = fn;
	if (abstract)
	{
		abstract_copy = *abstract;
		cmd.abstract = &abstract_copy;
	}
	else
		cmd.abstract = NULL;

	add_api_cmd_to_array(&num_cmds, &api_fd_commands, &cmd);
}

static BOOL api_pipe_nack_resp(rpcsrv_struct * l, uint16 rej_code,
			       prs_struct *resp)
{
	prs_struct rhdr;
	prs_struct rnack;
	RPC_HDR_NACK hdr_nack;

	DEBUG(5, ("api_pipe_nack_resp: make response\n"));

#if 0
	l->faulted_once_before = True;
#endif

	prs_init(&rhdr, MARSHALL);
	prs_init(&rnack, MARSHALL);

	/***/
	/*** set up the header and nack status ***/
	/***/

	hdr_nack.rej_code = rej_code;

	make_rpc_hdr(&l->hdr, RPC_BINDNACK,
		     RPC_FLG_FIRST | RPC_FLG_LAST, l->hdr.call_id, 0x12, 0);

	smb_io_rpc_hdr("hdr", &l->hdr, &rhdr, 0);
	smb_io_rpc_hdr_nack("nack", &hdr_nack, &rnack, 0);
	prs_realloc_data(&rhdr, rhdr.offset);
	prs_realloc_data(&rnack, rnack.offset);

	prs_init(resp, MARSHALL);
	prs_append_prs(resp, &rhdr);
	prs_append_prs(resp, &rnack);

	prs_free_data(&rnack);
	prs_free_data(&rhdr);

	return True;
}

static BOOL api_pipe_fault_resp(rpcsrv_struct * l, uint32 status,
				prs_struct *resp)
{
	prs_struct rhdr;
	prs_struct rfault;
	RPC_HDR_FAULT hdr_fault;
	RPC_HDR_RESP hdr_resp;

	DEBUG(5, ("api_pipe_fault_resp: make response\n"));

	if (l->auth_validated == False)
	{
		l->faulted_once_before = True;
	}

	prs_init(&rhdr, MARSHALL);
	prs_init(&rfault, MARSHALL);

	/***/
	/*** set up the header, response header and fault status ***/
	/***/

	hdr_fault.status = status;
	hdr_fault.reserved = 0x0;

	hdr_resp.alloc_hint = 0x0;
	hdr_resp.cancel_count = 0x0;
	hdr_resp.context_id = 0x0;
	hdr_resp.reserved = 0x0;

	make_rpc_hdr(&l->hdr, RPC_FAULT,
		     RPC_FLG_NOCALL | RPC_FLG_FIRST | RPC_FLG_LAST,
		     l->hdr.call_id, 0x20, 0);

	smb_io_rpc_hdr("hdr", &(l->hdr), &(rhdr), 0);
	smb_io_rpc_hdr_resp("resp", &(hdr_resp), &(rhdr), 0);
	smb_io_rpc_hdr_fault("fault", &(hdr_fault), &(rfault), 0);
	prs_realloc_data(&rhdr, rhdr.offset);
	prs_realloc_data(&rfault, rfault.offset);

	prs_init(resp, MARSHALL);
	prs_append_prs(resp, &rhdr);
	prs_append_prs(resp, &rfault);

	prs_free_data(&rfault);
	prs_free_data(&rhdr);

	return True;
}

static BOOL srv_pipe_bind_and_alt_req(const struct api_cmd *api_cmd,
				      rpcsrv_struct * l,
				      const char *ack_pipe_name,
				      prs_struct *resp,
				      enum RPC_PKT_TYPE pkt_type)
{
	BOOL ret;

	prs_struct rhdr;
	uint32 assoc_gid;

	RPC_HDR_RB *hdr_rb;
	char *uuid_transfer, *uuid_abstract;
	RPC_IFACE ret_abstract;

	l->auth = NULL;

	/* decode the bind request */
	if (!smb_io_rpc_hdr_rb("", &l->hdr_rb, &l->data_i, 0))
	{
		DEBUG(1, ("srv_pipe_bind_and_alt_req: smb_io_rpc_hdr_rb failed\n"));
	}

	if (l->data_i.offset == 0)
	{
		DEBUG(1, ("srv_pipe_bind_and_alt_req: decode hdr_rb failed\n"));
		return False;
	}

	hdr_rb = &l->hdr_rb;

	uuid_transfer = rpc_uuid_to_str(&hdr_rb->transfer.uuid);
	uuid_abstract = rpc_uuid_to_str(&hdr_rb->abstract.uuid);

	DEBUG(5, ("srv_pipe_bind_and_alt_req: transfer: %s, 0x%x, "
		  "abstract: %s, 0x%x\n",
		  uuid_transfer, hdr_rb->transfer.version,
		  uuid_abstract, hdr_rb->abstract.version));

	if (api_cmd->abstract
	    && (memcmp(&hdr_rb->abstract, api_cmd->abstract,
		       sizeof(hdr_rb->abstract)) != 0))
	{
		char *uuid_server = rpc_uuid_to_str(&api_cmd->abstract->uuid);

		DEBUG(1, ("srv_pipe_bind_and_alt_req: Client requests "
			  "abstract %s, 0x%x, we only have %s, 0x%x\n",
			  uuid_abstract, hdr_rb->abstract.version,
			  uuid_server, api_cmd->abstract->version));
		safe_free(uuid_server);

		ZERO_STRUCT(ret_abstract);
	}
	else
	{
		ret_abstract = hdr_rb->transfer;
	}

	safe_free(uuid_transfer);
	safe_free(uuid_abstract);

	if (l->hdr.auth_len != 0)
	{
		RPC_HDR_AUTH auth_info;
		BOOL found = False;
		size_t i;
		size_t old_offset, new_offset;

		old_offset = prs_offset(&l->data_i);
		new_offset = prs_buf_len(&l->data_i) - (l->hdr.auth_len + 8);

		if (old_offset != new_offset)
			DEBUG(4, ("Offset mismatch: 0x%x != 0x%x\n",
				  old_offset, new_offset));

		if (! prs_set_offset(&l->data_i, new_offset))
			return False;

		if (! smb_io_rpc_hdr_auth("", &auth_info, &l->data_i, 0))
			return False;

		for (i = 0; i < l->num_auths && !found; i++)
		{
			if (l->auth_fns[i]->api_is_auth(&auth_info,
							&l->auth_info))
			{
				l->auth = l->auth_fns[i];
				found = True;
			}
		}
		if (!found)
		{
			DEBUG(1, ("srv_pipe_bind_and_alt_req: Couldn't find "
				  "module for auth 0x%x/0x%x (0x%x)\n",
				  auth_info.auth_type, auth_info.auth_level,
				  l->hdr.auth_len));
			return False;
		}

		assoc_gid = sys_getpid();

		l->key_id = auth_info.key_id;
		l->auth_level = auth_info.auth_level;
	}
	else
	{
		extern srv_auth_fns noauth_fns;

		l->auth = &noauth_fns;
		l->auth_info = NULL;

		assoc_gid = l->hdr_rb.bba.assoc_gid;
#if 0
		if (assoc_gid != 0)
		{
			l->key.pid = assoc_gid;
		}
#endif
	}

#if 0
	if (assoc_gid == 0)
	{
		assoc_gid = sys_getpid();
	}
#endif

	if (l->auth != NULL)
	{
		SMB_ASSERT(l->auth->api_auth_chk != NULL);
		if (!l->auth->api_auth_chk(l, pkt_type))
		{
			safe_free(l->auth_info);
			l->auth_info = NULL;
			DEBUG(1, ("srv_pipe_bind_and_alt_req: auth failed\n"));
			return False;
		}
	}
	DEBUG(5, ("api_pipe_bind_req: make response. %d\n", __LINE__));

	prs_init(&l->rdata, MARSHALL);
	prs_init(&rhdr, MARSHALL);

	/***/
	/*** do the bind ack first ***/
	/***/

	make_rpc_hdr_ba(&l->hdr_ba,
			l->hdr_rb.bba.max_tsize,
			l->hdr_rb.bba.max_rsize,
			assoc_gid,
			ack_pipe_name, 0x1, 0x0, 0x0, &ret_abstract);

	smb_io_rpc_hdr_ba("", &l->hdr_ba, &l->rdata, 0);
	prs_realloc_data(&l->rdata, l->rdata.offset);

	if (l->auth != NULL)
	{
		/***/
		/*** now the authentication ***/
		/***/

		SMB_ASSERT(l->auth->api_auth_gen != NULL);

		ret = l->auth->api_auth_gen(l, resp, pkt_type);

		if (!ret)
		{
			free(l->auth_info);
			l->auth_info = NULL;
			prs_free_data(&l->rdata);
			return False;
		}
	}
	else
	{
		return False;
	}

	return ret;
}

static BOOL api_pipe_bind_and_alt_req(rpcsrv_struct * l,
				      const char *name,
				      prs_struct *resp,
				      enum RPC_PKT_TYPE pkt_type)
{
	fstring ack_pipe_name;
	fstring pipe_srv_name;
	size_t i = 0;

	DEBUG(5, ("api_pipe_bind_req: decode request. %d\n", __LINE__));

	for (i = 0; i < num_cmds; i++)
	{
		if (strequal(api_fd_commands[i]->pipe_clnt_name, name) &&
		    api_fd_commands[i]->fn != NULL)
		{
			DEBUG(3,
			      ("api_pipe_bind_req: \\PIPE\\%s -> \\PIPE\\%s\n",
			       api_fd_commands[i]->pipe_clnt_name,
			       api_fd_commands[i]->pipe_srv_name));
			fstrcpy(pipe_srv_name,
				api_fd_commands[i]->pipe_srv_name);
			break;
		}
	}

	if (api_fd_commands[i]->fn == NULL)
		return False;

	switch (pkt_type)
	{
		case RPC_BINDACK:
		{
			/* name has to be \PIPE\xxxxx */
			fstrcpy(ack_pipe_name, "\\PIPE\\");
			fstrcat(ack_pipe_name, pipe_srv_name);
			break;
		}
		case RPC_ALTCONTRESP:
		{
			/* secondary address CAN be NULL
			 * as the specs says it's ignored.
			 * It MUST NULL to have the spoolss working.
			 */
			fstrcpy(ack_pipe_name, "");
			break;
		}
		default:
		{
			DEBUG(10,("api_pipe_bind_and_alt_req: unknown pkt_type %d\n", pkt_type));
			return False;
		}
	}

	return srv_pipe_bind_and_alt_req(api_fd_commands[i], l,
					 ack_pipe_name, resp, pkt_type);
}

/*
 * The RPC Alter-Context call is used only by the spoolss pipe
 * simply because there is a bug (?) in the MS unmarshalling code
 * or in the marshalling code. If it's in the later, then Samba
 * have the same bug.
 */
static BOOL api_pipe_bind_req(rpcsrv_struct * l, const char *name,
			      prs_struct *resp)
{
	return api_pipe_bind_and_alt_req(l, name, resp, RPC_BINDACK);
}

static BOOL api_pipe_alt_req(rpcsrv_struct * l, const char *name,
			     prs_struct *resp)
{
	return api_pipe_bind_and_alt_req(l, name, resp, RPC_ALTCONTRESP);
}

static uint32 api_pipe_request(rpcsrv_struct * l, const char *name,
			     prs_struct *resp)
{
	size_t i = 0;

	if ((l->auth != NULL) && l->auth_validated)
	{
		DEBUG(10, ("api_pipe_request: validated auth\n"));

		if (l->hdr.auth_len && !l->auth->api_decode_pdu)
		{
			DEBUG(2, ("api_pipe_request: Got packet with auth, "
				  "but none was negotiated\n"));
			return nca_op_rng_error;
		}

		if (l->auth->api_decode_pdu && !l->auth->api_decode_pdu(l))
			return nca_op_rng_error;
	}

	if (num_cmds == 0)
	{
		DEBUG(0, ("no commands to process!\n"));
	}
	else
	{
		DEBUG(10, ("pipe name: %s\n", name));
	}

	for (i = 0; i < num_cmds; i++)
	{
		DEBUG(10,
		      ("search name: %s\n",
		       api_fd_commands[i]->pipe_clnt_name));
		if (strequal(api_fd_commands[i]->pipe_clnt_name, name)
		    && api_fd_commands[i]->fn != NULL)
		{
			DEBUG(3,
			      ("Doing \\PIPE\\%s\n",
			       api_fd_commands[i]->pipe_clnt_name));
			if (!api_fd_commands[i]->fn(l))
			{
				return nca_op_rng_error;
			}
			l->rdata_offset = 0;

			/* create the rpc pdu */
			if (! create_rpc_reply(l, 0, resp))
				return nca_op_rng_error;

			return 0x0;
		}
	}

	return nca_op_rng_error;
}

static BOOL rpc_redir_local(rpcsrv_struct * l, prs_struct *req,
			    prs_struct *resp, const char *name)
{
	BOOL reply = False;
	uint32 fault_status = nca_op_rng_error;
	BOOL last;
	BOOL first;

	if (req->data == NULL || req->data_size == 0)
	{
		if (l->rdata.data == NULL)
		{
			return False;
		}
		/* hmm, must need some more data.
		 * create, flatten and return data in a single pdu
		 */
		if (!create_rpc_reply(l, l->rdata_offset, resp))
			return False;

		return True;
	}

	if (req->data == NULL)
		return False;

	/* lkclXXXX still assume that the first complete PDU is always
	   in a single request!!!
	 */
	/* process the rpc header */
	req->offset = 0x0;
	req->io = True;
	smb_io_rpc_hdr("hdr", &l->hdr, req, 0);

	if (req->offset == 0)
		return False;

	last = IS_BITS_SET_ALL(l->hdr.flags, RPC_FLG_LAST);
	first = IS_BITS_SET_ALL(l->hdr.flags, RPC_FLG_FIRST);

	if (l->hdr.pkt_type == RPC_BIND ||
	    l->hdr.pkt_type == RPC_ALTCONT || l->hdr.pkt_type == RPC_BINDRESP)
	{
		last = True;
		first = True;
	}

	if (first)
	{
		prs_init(&l->data_i, UNMARSHALL);
		prs_set_packtype(&l->data_i, l->hdr.pack_type);
	}
	if (last)
	{
		prs_append_data(&l->data_i,
				prs_data(req, req->offset),
				req->data_size - req->offset);
	}
	else
	{
		prs_init(resp, MARSHALL);
		return True;
	}

	/* previous authentication failure.  don't give a monkey's what
	 * is sent to us, we reject it, outright
	 */

	if (l->faulted_once_before)
	{
		DEBUG(10,
		      ("rpc_redir_local: faulted before (so do it again)\n"));
		prs_free_data(&l->data_i);
		return api_pipe_fault_resp(l, nca_op_rng_error, resp);
	}

	switch (l->hdr.pkt_type)
	{
		case RPC_BIND:
		{
			reply = become_vuser(l->initial_pipe_vuid);
			if (!reply)
			{
				DEBUG(10, ("become_vuser failed: %x\n", l->vuid));
				break;
			}
			reply = api_pipe_bind_req(l, name, resp);
			if (!reply)
			{
				reply = api_pipe_nack_resp(l, 0x0, resp);
			}
			break;
		}
		case RPC_ALTCONT:
		{
			reply = api_pipe_alt_req(l, name, resp);
			break;
		}
		case RPC_REQUEST:
		{
			if (l->auth != NULL && !l->auth_validated)
			{
				/* authentication _was_ requested
				   and it failed.  sorry, no deal!
				 */
				reply = False;
			}
			else
			{
				/* read the rpc header */
				reply = smb_io_rpc_hdr_req("req",
							   &(l->hdr_req),
							   &l->data_i, 0);
				if (!reply)
				{
					DEBUG(1, ("smb_io_rpc_hdr_req failed\n"));
				}
				if (reply)
				{
					reply = become_vuser(l->vuid);

				}
				if (!reply)
				{
					DEBUG(1, ("become_vuser failed: %x\n", l->vuid));
				}
				if (reply)
				{
					fault_status
						= api_pipe_request(l, name,
								   resp);
					reply = (fault_status == 0x0);
				}
				if (!reply)
				{
					DEBUG(2, ("dce/rpc request failed\n"));
				}
			}
			break;
		}
		case RPC_BINDRESP:	/* not the real name! */
		{
			if (l->auth != NULL)
			{
				SMB_ASSERT(l->auth->api_auth_chk != NULL);

				reply = l->auth->api_auth_chk(l,
							      l->hdr.
							      pkt_type);
				if (!reply)
				{
					DEBUG(1, ("rpc_redir_local: Call to l->auth->api_auth_chk failed\n"));
				}
			}
			if (!reply)
			{
				l->auth = NULL;
				if (l->auth_info != NULL)
				{
					free(l->auth_info);
					l->auth_info = NULL;
				}
				l->auth_validated = False;
			}
			break;
		}
	}

	if (!reply)
	{
		reply = api_pipe_fault_resp(l, fault_status, resp);
	}

	if (reply)
	{
		/* flatten the data into a single pdu */
		DEBUG(200, ("rpc_redir_local: %d\n", __LINE__));
		prs_debug_out(resp, "redir_local resp", 200);

		prs_free_data(&l->data_i);
		return True;
	}

	/* delete intermediate data used to set up the pdu.  leave
	   rdata alone because that's got the rest of the data in it */
	prs_free_data(&l->data_i);

	return reply;
}

/*******************************************************************
 receives a netlogon pipe and responds.
 ********************************************************************/
static BOOL api_rpc_command(rpcsrv_struct * l, const char *rpc_name,
			    const struct api_struct *api_rpc_cmds)
{
	int fn_num;
	int dbg_level;
	BOOL ret = True;
	const char *fn_name;

	for (fn_num = 0; api_rpc_cmds[fn_num].name; fn_num++)
	{
		if (api_rpc_cmds[fn_num].opnum == l->hdr_req.opnum)
		{
			break;
		}
	}

	fn_name = api_rpc_cmds[fn_num].name;

	if (fn_name == NULL || api_rpc_cmds[fn_num].fn == NULL)
	{
		dbg_level = 2;
		ret = False;
	}
	else
	{
		dbg_level = 3;
	}

	DEBUG(dbg_level, ("api_rpc_command: %s op 0x%02x - %s%s\n",
			  rpc_name, l->hdr_req.opnum,
			  (fn_name ? fn_name : "unknown"),
			  (api_rpc_cmds[fn_num].fn ? ""
			   : " (not implemented)")));

	if (! ret)
	{
		return False;
	}

	prs_init(&l->rdata, MARSHALL);
	prs_reset_ptr_token();

	/* Set the right side for the prs_structs */
	prs_set_side(&l->data_i, PRS_SERVER);
	prs_set_side(&l->rdata, PRS_SERVER);

	{
		size_t len = prs_data_size(&l->data_i);
		DEBUG(10, ("The dce/rpc body input data\n"));
		dump_data(10, prs_data_read(&l->data_i, 0, len), len);
	}

	/* do the actual command */
	/* XXX - NULL should realy be a handle for the connection (Elrond) */
	ret = api_rpc_cmds[fn_num].fn(NULL, &l->data_i, &l->rdata);

	if (!ret)
	{
		DEBUG(2, ("%s command %s failed -> fault pdu\n",
			  rpc_name, fn_name));
	}

	if (ret && (prs_offset(&l->data_i) != prs_data_size(&l->data_i)))
	{
		DEBUG(2, ("Notice: %s command %s parsed only 0x%x/0x%x "
			  "(%d) from the input data\n", rpc_name, fn_name,
			  prs_offset(&l->data_i), prs_data_size(&l->data_i),
			  prs_data_size(&l->data_i) - prs_offset(&l->data_i)));
	}

	if (l->rdata.data == NULL || l->rdata.offset == 0)
	{
		prs_free_data(&l->rdata);
		return False;
	}

	prs_realloc_data(&l->rdata, l->rdata.offset);

	DEBUG(10, ("called %s\n", rpc_name));

	return ret;
}


/*******************************************************************
 receives a netlogon pipe and responds.
 ********************************************************************/
BOOL api_rpcTNP(rpcsrv_struct * l, const char *rpc_name,
		const struct api_struct *api_rpc_cmds)
{
	if (l == NULL)
	{
		DEBUG(1, ("NULL rpcsrv_struct\n"));
		return False;
	}
	if (l->data_i.data == NULL)
	{
		DEBUG(2, ("%s: NULL data received\n", rpc_name));
		return False;
	}

	/* interpret the command */
	if (!api_rpc_command(l, rpc_name, api_rpc_cmds))
	{
		return False;
	}

	return True;
}

/*******************************************************************
 entry point from msrpc to smb.  adds data received to pdu; checks
 pdu; hands pdu off to msrpc, which gets a pdu back (except in the
 case of the RPC_BINDCONT pdu).
 ********************************************************************/
BOOL rpc_local(rpcsrv_struct * l, char *data, int len, const char *name)
{
	BOOL reply = False;

	DEBUG(10, ("rpc_local: len %d\n", len));

	if (len != 0)
	{
		reply = prs_add_data(&l->smb_pdu, data, len);

		if (reply && is_complete_pdu(&l->smb_pdu))
		{
			l->smb_pdu.offset = l->smb_pdu.data_size;
			prs_set_start(&l->smb_pdu, 0);
			reply =
				rpc_redir_local(l, &l->smb_pdu, &l->rsmb_pdu,
						name);
			prs_free_data(&l->smb_pdu);
			prs_init(&l->smb_pdu, UNMARSHALL);
		}
	}
	else
	{
		if (l->rdata.data == NULL || l->rdata.data_size == 0)
		{
			DEBUG(10, ("rpc_local: no data to send\n"));
			return True;
		}
		prs_free_data(&l->smb_pdu);
		prs_init(&l->smb_pdu, UNMARSHALL);
		reply = rpc_redir_local(l, &l->smb_pdu, &l->rsmb_pdu, name);
	}
	return reply;
}
