/* 
   Unix SMB/Netbios implementation.
   Version 1.9.
   SMB client
   Copyright (C) Luke Kenneth Casson Leighton 1996-2001
   Copyright (C) Andrew Tridgell 1994-1998
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#include "includes.h"
#include <signal.h>
#include "libsamba.h"
#include "tng_misc.h"
#include "tng_readline.h"
#include "libnmb.h"
#include "libsmb.h"
#include "rpc_parse.h"
#include "rpc_client.h"
#include "nterr.h"
#include "pathnames.h"
#include "version.h"
#include "rpcclient.h"

#ifndef REGISTER
#define REGISTER 0
#endif

extern pstring user_socket_options;
extern struct ntuser_creds *usr_creds;
extern BOOL in_client;



#define IS_HIDDEN_ALIAS(cmd) (! (cmd)->description)
#define CMD_HAS_SUBCMD(cmd) (!(cmd)->fn || (cmd)->subcommands)


static void free_cmd_set_array(struct command_set **entries);
static int find_command(int argc, char **argv,
			struct command_set **ret_cmd, int *skip_args);
static uint32 cmd_help(struct client_info *info, int argc, char *argv[]);
static uint32 cmd_quit(struct client_info *info, int argc, char *argv[]);
static uint32 cmd_set(struct client_info *info, int argc, char *argv[]);
static NTSTATUS cmd_conn_list(struct client_info *info, int argc, char *argv[]);
static uint32 cmd_use(struct client_info *info, int argc, char *argv[]);

static struct ntuser_creds usr;

struct client_info cli_info;

char **cmd_argv = NULL;
uint32 cmd_argc = 0;

FILE *out_hnd;


static void cmd_set_free(struct command_set *item)
{
	if (item != NULL)
	{
		safe_free(item->name);
	}
	free_cmd_set_array(item->subcommands);
	safe_free(item);
}

static struct command_set *cmd_set_dup(const struct command_set *from)
{
	if (from != NULL)
	{
		struct command_set *copy = g_new(struct command_set, 1);
		if (copy != NULL)
		{
			memcpy(copy, from, sizeof(struct command_set));
			if (from->name != NULL)
			{
				copy->name = strdup(from->name);
			}
		}
		return copy;
	}
	return NULL;
}

static void free_cmd_set_array(struct command_set **entries)
{
	struct command_set **cmds;

	if (! entries)
		return;

	for (cmds = entries; *cmds; cmds++)
		cmd_set_free(*cmds);

	free(entries);
}

static struct command_set *add_cmd_set_to_array(struct command_set ***array,
						const struct command_set *cmd)
{
	struct command_set *copy = cmd_set_dup(cmd);

	return (struct command_set *)add_item_to_array(NULL,
						       (void ***)array,
						       copy);

}

static struct command_set **commands = NULL;

/****************************************************************************
 add in individual command-sets.
 ****************************************************************************/
void add_command_set(const struct command_set *cmds)
{
	while (cmds->name && *cmds->name)
	{
		add_cmd_set_to_array(&commands, cmds);
		cmds++;
	}
}

/****************************************************************************
 This defines the commands supported by this client
 ****************************************************************************/
static const struct command_set general_commands[] = {
	/*
	 * maintenance
	 */

	{
		"devel", NULL,
		"Commands for developer use"
	},
	{
		"set", cmd_set,
		"run rpcclient inside rpcclient (change options etc.)"
	},
	{
		"use", cmd_use,
		"net use and net view"
	},
	{
		"conn", NULL,
		"Connection manipulation commands"
	},
	{
		"conn list", cmd_conn_list,
		"List active connections"
	},
	{
		"exit", cmd_quit,
		"logoff the server"
	},
	{ "quit", cmd_quit },
	{ "q", cmd_quit },
	{ "bye", cmd_quit },
	{
		"help", cmd_help,
		"[command] give help on a command"
	},
	{ "?", cmd_help },
	{
		"!", NULL,
		"run a shell command on the local system"
	},

	{ NULL }
};


/****************************************************************************
do a (presumably graceful) quit...
****************************************************************************/
static uint32 cmd_quit(struct client_info *info, int argc, char *argv[])
{
#ifdef MEM_MAN
	{
		smb_mem_write_status(dbf);
		smb_mem_write_errors(dbf);
		smb_mem_write_verbose(dbf);
		dbgflush();
	}
#endif
	free_connections();
	exit(0);
}

/****************************************************************************
help
****************************************************************************/
static const struct command_set *cmd_find_primary(struct command_set **table,
						  const struct command_set *cmd)
{
	if (! table)
		return NULL;
	for (; *table; table++)
	{
		const struct command_set *t = (*table);

		if ((t->fn == cmd->fn) && t->description)
			return t;
		if (! t->subcommands)
			continue;
		t = cmd_find_primary(t->subcommands, cmd);
		if (t)
			return t;
	}
	return NULL;
}

static void cmd_help_show_one(const struct command_set *cmd)
{
	if (IS_HIDDEN_ALIAS(cmd))
	{
		const struct command_set *aliased;

		aliased = cmd_find_primary(commands, cmd);
		if (! aliased)
		{
			report(out_hnd, "`%s' doesn't have a description and isn't an alias\n",
				cmd->name);
			return;
		}
		report(out_hnd, "`%s' seems to be an alias for `%s'\n\n",
			cmd->name, aliased->name);
		cmd = aliased;
	}
	report(out_hnd, "HELP for `%s':\n\t%s\n",
		cmd->name, cmd->description);
}

void cmd_help_current(struct client_info *info)
{
	if (! info->current_cmd)
	{
		report(out_hnd, "No current command active\n");
		return;
	}
	cmd_help_show_one(info->current_cmd);
}

static const char *cmd_strip_prefix(const char *p,
				    const char *prefix, int prefixlen)
{
	if (prefix)
	{
		if (strncmp(p, prefix, prefixlen) != 0)
			return NULL;

		p += prefixlen;

		if (*p == '\0')
			return NULL;

		if (*p == ' ')
			p++;
		else
			return NULL;
	}
	if (strchr(p, ' '))
		return NULL;

	return p;
}

static void cmd_help_show_list(struct command_set **table, 
			       const char *prefix)
{
	int maxlen = 0;
	int prefixlen = 0;
	int i;

	if (! table)
		return;

	if (prefix)
		prefixlen = strlen(prefix);

	for (i = 0; table[i]; i++)
	{
		int len;
		const char *p = table[i]->name;

		if (IS_HIDDEN_ALIAS(table[i]))
			continue;
		p = cmd_strip_prefix(p, prefix, prefixlen);
		if (! p)
			continue;
		len = strlen(p);
		if (CMD_HAS_SUBCMD(table[i]))
			len += 4;
		if (len > maxlen)
			maxlen = len;
	}
	maxlen += 2;
	for (i = 0; table[i]; i++)
	{
		int showlen;
		const char *p, *dots;
		const char *n = table[i]->name;

		if (IS_HIDDEN_ALIAS(table[i]))
			continue;
		n = cmd_strip_prefix(n, prefix, prefixlen);
		if (! n)
			continue;
		if (CMD_HAS_SUBCMD(table[i]))
			dots = " ...";
		else
			dots = "";
		p = strchr(table[i]->description, '\n');
		if (p)
			showlen = p - commands[i]->description;
		else
			showlen = -1;
		fprintf(out_hnd, "  %s%s%*s%.*s\n", n, dots,
			maxlen - strlen(n) - strlen(dots), "",
			showlen, commands[i]->description);
	}
}


static uint32 cmd_help(struct client_info *info, int argc, char *argv[])
{
	struct command_set *cmd;

	if (argc > 1)
	{
		int i = find_command(argc-1, argv+1, &cmd, NULL);
		if (i >= 0 || i==-3)
		{
			cmd_help_show_one(cmd);
			if (! cmd->fn)
				cmd_help_show_list(commands, cmd->name);
		}
	}
	else
	{
		cmd_help_show_list(commands, NULL);
	}
	return 0;
}

/*******************************************************************
  lookup a command string in the list of commands, including 
  abbreviations
  ******************************************************************/
static int find_command(int argc, char **argv,
			struct command_set **ret_cmd, int *skip_args)
{
	struct command_set *found_cmd = NULL;
	int tok_len;
	const char *tok;
	const char *prefix = NULL;
	int prefixlen = 0;
	struct command_set **t, **table;
	int matches;

	if (ret_cmd)
		*ret_cmd = NULL;
	if (skip_args)
		*skip_args = 0;

	if (! argc)
		return -3; /* need further command words ;) */

	table = commands;

	while (True)
	{
		struct command_set *cmd = NULL;	/* shut gcc up */

		tok = argv[0];
		tok_len = strlen(tok);
		matches = 0;

		for (t = table; *t; t++)
		{
			const char *n;

			n = cmd_strip_prefix((*t)->name, prefix, prefixlen);

			if (! n)
				continue;

			if (strequal(n, tok))
			{
				matches = 1;
				cmd = *t;
				break;
			}
			else if (! IS_HIDDEN_ALIAS(*t)
				 && strnequal(n, tok, tok_len))
			{
				matches++;
				cmd = *t;
			}
		}
		if (matches == 0)
			break;
		if (matches == 1)
		{
			found_cmd = cmd;
			if (skip_args)
				(*skip_args)++;
			if (cmd->fn && (! cmd->subcommands))
				break;
			if (cmd->subcommands)
			{
				table = cmd->subcommands;
				prefix = cmd->name;
				prefixlen = strlen(prefix);
			}
			else
			{
				prefix = cmd->name;
				prefixlen = strlen(prefix);
			}
			argc--;
			argv++;
			if (! argc)
				break;
		}
		if (matches > 1)
			break;
	}

	if (found_cmd)
		*ret_cmd = found_cmd;

	if (matches > 1)
		return -2;

	if ((! found_cmd) || (matches == 0))
		return -1;

	if (! found_cmd->fn)
		return -3;

	return 0;
}

/****************************************************************************
  process commands from the client
****************************************************************************/
static NTSTATUS do_command(struct client_info *info, char *line)
{
	NTSTATUS status = NT_STATUS_INVALID_PARAMETER;
	int i, skip_args;
	pstring exitmsg;
	struct command_set *cmd;

	exitmsg[0] = 0;

	if (!argstr_to_array(line, &cmd_argc, &cmd_argv))
		return status;

	if (cmd_argc == 0)
		return NT_STATUS_OK;

	i = find_command(cmd_argc, cmd_argv, &cmd, &skip_args);

	if (i >= 0)
	{
		int argc = ((int)cmd_argc);
		char **argv = cmd_argv;

		optind = 0;
		skip_args--;
		argc -= skip_args;
		argv += skip_args;

		DEBUG(4, ("Executing `%s'\n", cmd->name));
		info->current_cmd = cmd;
		status = cmd->fn(info, argc, argv);
		info->current_cmd = NULL;
		DEBUG(4, ("Done with `%s': %s (0x%08x)\n", cmd->name,
			  get_nt_error_msg(status), status));

		if (status == NT_STATUS_UNSUCCESSFUL)
		{
			unsafe_strcpy(exitmsg, "FAILED");
		}
		else if (NT_STATUS_IS_ERR(status))
		{
			slprintf(exitmsg, sizeof(pstring)-1,
				 "FAILED: %s (0x%08x)\n",
				 get_nt_error_msg(status), status);
		}
	}
	else if (i == -3)
	{
		unsafe_strcpy(exitmsg, "command expects further subcommand.\n"
			      "\tUse 'help' to show possibilities.\n");
	}
	else if (i == -2)
	{
		unsafe_strcpy(exitmsg, "command abbreviation ambiguous");
	}
	else
	{
		unsafe_strcpy(exitmsg, "command not found");
	}

	if (exitmsg[0])
		fprintf(out_hnd, "%s: %s\n",
			(cmd ? cmd->name : cmd_argv[0]), exitmsg);

	free_char_array(cmd_argc, cmd_argv);
	cmd_argc = 0;
	cmd_argv = NULL;

	return status;
}


/****************************************************************************
  process commands from the client
****************************************************************************/
static uint32 process(struct client_info *info, char *cmd_str)
{
	uint32 status = 0;
	pstring line;
	char *cmd = cmd_str;

	if (cmd != NULL)
	{
		while (cmd[0] != '\0')
		{
			char *p;

			if ((p = strchr(cmd, ';')) == 0)
			{
				strncpy(line, cmd, 999);
				line[1000] = '\0';
				cmd = "";
			}
			else
			{
				if (p - cmd > 999)
					p = cmd + 999;
				strncpy(line, cmd, p - cmd);
				line[p - cmd] = '\0';
				cmd = p + 1;
			}

			status = do_command(info, line);
			if (status == 0x0)
			{
				continue;
			}
		}
	}
	else
	{
		while (!feof(stdin))
		{
			char *ret_line;
			pstring pline;
			BOOL at_sym = False;
			
			memcpy(pline, "[", 2);
			if (usr.domain[0] != 0)
			{
				pstrcat(pline, usr.domain);
				pstrcat(pline, "\\");
				at_sym = True;
			}
			if (usr.user_name[0] != 0)
			{
				pstrcat(pline, usr.user_name);
				at_sym = True;
			}
			if (at_sym)
			{
				pstrcat(pline, "@");
			}

			pstrcat(pline, cli_info.dest_host);
			pstrcat(pline, "]$ ");

			ret_line = tng_micro_readline(pline);
			if (!ret_line)
				break;

			pstrcpy(line, ret_line);
			safe_free(ret_line);

			/* special case - first char is ! */
			if (*line == '!')
			{
				system(line + 1);
				continue;
			}

			status = do_command(info, line);
			if (status == 0x0)
			{
				continue;
			}
		}
	}
	return status;
}

/****************************************************************************
usage on the program
****************************************************************************/
static void usage(char *pname)
{
	tng_out_version_blurb();
	fprintf(out_hnd, "Usage: %s [options]\n", pname);

	fprintf(out_hnd,
		"\t-S server             server or IP address to connect to\n"
		"\t-S .                  connect to local socket (default)\n"
		"\t-p port               connect to the specified TCP port\n"
		"\t-n netbios name       Use this name as my NetBIOS name\n"
		"\t-U username           set the network username\n"
		"\t-U username%%pass      set the network username and password\n"
		"\t-W domain             set the domain name\n"
		"\t-A filename           file with authentication credentials\n"
		"\t-c 'command string'   execute semicolon-separated commands\n"
		"\t-d debuglevel         set the debuglevel\n"
		"\t-l log basename.      basename for log/debug files\n"
		"\t-E                    write messages to stderr instead of stdout\n"
		"\t-N                    use empty password (same as -U [username]%%)\n"
		"\t-t terminal code      terminal i/o code {sjis|euc|jis7|jis8|junet|hex}\n"
		"\t-m max protocol       (ignored for backward compatibility)\n"
		"\t-I dest IP            (same as '-S', for backward compatibility)\n"
		"\n");
}

#ifdef HAVE_READLINE

/****************************************************************************
GNU readline completion functions
****************************************************************************/

/* Complete an rpcclient command */

static struct command_set **tabber_table = NULL;
static const char *tabber_prefix = NULL;
static int tabber_prefixlen;
static BOOL tabber_hide_hidden_aliases;

static char *complete_cmd(const char *text, int state)
{
	static struct command_set **t;
	static size_t len_text;
	const char *name;

	/* Initialise */

	rl_completion_append_character = ' ';

	if (state == 0)
	{
		struct command_set **p;

		len_text = strlen(text);

		t = tabber_table;
		if (tabber_prefix)
			tabber_prefixlen = strlen(tabber_prefix);
		else
			tabber_prefixlen = 0;

		tabber_hide_hidden_aliases = False;
		for (p = t; *p; p++)
		{
			if (IS_HIDDEN_ALIAS(*p))
				continue;
			name = cmd_strip_prefix((*p)->name, tabber_prefix,
						tabber_prefixlen);
			if (name && (StrnCaseCmp(name, text, len_text) == 0))
			{
				/*
				 * Okay, we found one non-hidden alias,
				 * so don't show any hidden aliases.
				 */
				tabber_hide_hidden_aliases = True;
				break;
			}
		}
	}

	/* Return the next name which partially matches the list of commands */

	for (; *t; t++)
	{
		if (tabber_hide_hidden_aliases && IS_HIDDEN_ALIAS(*t))
			continue;
		name = cmd_strip_prefix((*t)->name, tabber_prefix,
					tabber_prefixlen);
		if (! name)
			continue;
		if (StrnCaseCmp(name, text, len_text) == 0)
		{
			/* advance pointer, normaly done by for/continue ;) */
			t++;
			return strdup(name);
		}
	}

	return NULL;
}

static char **complete_cmd_helper(struct command_set *cmd,
				  const char *text)
{
	tabber_table = commands;
	tabber_prefix = NULL;
	if (cmd)
	{
		if (cmd->subcommands)
		{
			tabber_table = cmd->subcommands;
			tabber_prefix = cmd->name;
		}
		else
		{
			if (cmd->fn)
				/* This command doesn't have subcommands */
				return NULL;
			tabber_prefix = cmd->name;
		}
	}
	return rl_completion_matches(text, complete_cmd);
}


/* Main completion function */

static char **completion_fn(const char *text, int start, int end)
{
	static uint32 my_argc = 0;
	static char **my_argv = NULL;
	struct command_set *cmd;
	int i, skip_args;
	char r;
	char* (* completer) (const char *, int);
	char **ret_value;

	/***
	   All the code until the find_command here is just for one
	   thing:
	   Get the words till but exckluding the word, that readline
	   tries to complete.
	   We need this for two things:
	   a) find the multi word command, for which we might
	      complete an arg
	   b) To know, how many words to strip of for custom
	      completers.
	      Scroll down to the end of this function to see more on
	      this.

	   From the first glance, this job seems easy:

	      word1 word2 wor
	                  ^  ^- Cursor
	                  +- "start"
	   text will contain "wor".

	   So we simply write a NUL char at start, parse the line
	   and we're done.

	   Well, it's not that simple.
	   The problems include:
	   a) start might be in the middle of a word (depending on
	      how readline sees words).
	   b) start and the cursor might be at the same place,
	      inside a word.
	   c) Make this independent of the function that parsses the
	      whole line into words.

	   We can catch a) by parsing two times. First time with the
	   normal line, second time with a NUL at start. (Note, that
	   we do it the other way round below).
	   And then comparing the last relevant word.
	   But we don't get b) with that.
	   We catch b) with the "weird technology from outer
	   space ;)" below. It simply makes the "with NUL" version
	   different for the comparison.
	   c) is simply reached by using the normal parser and not
	   trying to parse the line ourselves.
	***/

	if (! *text && start > 0)
		start--; /* weird technology from outer space ;) */
	
	r = rl_line_buffer[start];
	rl_line_buffer[start] = '\0';
	argstr_to_array(rl_line_buffer, &my_argc, &my_argv);
	rl_line_buffer[start] = r;

	/* reparse */
	argstr_to_array(rl_line_buffer, &cmd_argc, &cmd_argv);

	if (cmd_argc < my_argc)
	{
		DEBUG(0, ("completion_fn: *ARG* cmd_argc shrinks at reparse\n"));
		return NULL;
	}

	if ((my_argc > 0)
	    && (strcmp(my_argv[my_argc-1], cmd_argv[my_argc-1]) != 0))
	{
		/*
		 * readline completing a word inside an "arg"
		 * Remove the complete arg from our list here
		 */
		my_argc--;
		safe_free(my_argv[my_argc]);
		my_argv[my_argc] = NULL;
	}

	i = find_command(my_argc, my_argv, &cmd, &skip_args);

	if ((i == -1) || (i == -2))
		/*
		 * We can't complete args for an unknwon
		 * or ambiguous command
		 */
		return NULL;

	if (i == -3)
	{
		return complete_cmd_helper(cmd, text);
	}

	if (skip_args < my_argc-1)
		/* Completely beyond anything, we can complete for */
		return NULL;

	completer = cmd->compl_args[my_argc - skip_args];

	if (! completer)
	{
		if (skip_args == my_argc)
			return complete_cmd_helper(cmd, text);
		else
			return NULL;
	}

	skip_args--;

	/*
	 * The Registry completer expects the old but
	 * weird "API" on cmd_arg*:
	 * a) cmd_argv[0] being the command
	 *    we let it point to the last word of the command
	 * b) cmd_argc being one too large... This was a "buglet"
	 *    of get_cmd_args();
	 */
	cmd_argc++;
	cmd_argc -= skip_args;
	cmd_argv += skip_args;
	ret_value = rl_completion_matches(text, completer);
	cmd_argc += skip_args;
	cmd_argv -= skip_args;
	cmd_argc--;

	return ret_value;
}

/* To avoid filename completion being activated when no valid
   completions are found, we assign this stub completion function
   to the rl_completion_entry_function variable. */

static char *complete_cmd_null(const char *text, int state)
{
	return NULL;
}

#endif /* HAVE_READLINE */

static void set_user_password(struct ntuser_creds *u,
			      BOOL got_pass, char *password)
{
	/* set the password cache info */
	if (got_pass)
	{
		if (password == NULL)
		{
			DEBUG(10, ("set_user_password: NULL pwd\n"));
			pwd_set_nullpwd(&(u->pwd));
		}
		else
		{
			/* generate 16 byte hashes */
			DEBUG(10, ("set_user_password: generate\n"));
			pwd_make_lm_nt_16(&(u->pwd), password);
		}
	}
	else
	{
		DEBUG(10, ("set_user_password: read\n"));
		pwd_read(&(u->pwd), "Enter Password:", True);
	}
}


static NTSTATUS cmd_conn_list(struct client_info *info, int argc, char *argv[])
{
	uint32 i;
	uint32 num_uses;
	struct use_info **use;

	cli_net_use_enum(&num_uses, &use);

	if (num_uses == 0)
	{
		report(out_hnd, "No connections\n");
	}
	else
	{
		report(out_hnd, "Connections:\n");

		for (i = 0; i < num_uses; i++)
		{
			if (use[i] != NULL && use[i]->connected)
			{
				report(out_hnd, "Server:\t%s\t",
				       use[i]->srv_name);
				report(out_hnd, "Key:\t[%x]\t",
				       use[i]->vuid);
				report(out_hnd, "User:\t%s\t",
				       use[i]->user_name);
				report(out_hnd, "Domain:\t%s\n",
				       use[i]->domain);
			}
		}
	}

	return NT_STATUS_OK;
}




static uint32 cmd_use(struct client_info *info, int argc, char *argv[])
{
	int opt;
	BOOL net_use_add = True;
	BOOL force_close = False;
	fstring dest_host;
	fstring srv_name;
	BOOL null_pwd = False;
	BOOL got_pwd = False;
	pstring password;

	copy_nt_creds(&usr, usr_creds);

	fstrcpy(dest_host, cli_info.dest_host);
	fstrcpy(usr.user_name, optarg);
	info->reuse = False;

	if (argc <= 1)
	{
		report(out_hnd,
		       "net [\\\\Server] [-U user%%pass] [-W domain] [-d] [-f]\n");
		report(out_hnd, "    -d     Deletes a connection\n");
		report(out_hnd, "    -f     Forcibly deletes a connection\n");
	}

	if (argc > 1 && (*argv[1] != '-'))
	{
		if (strnequal("\\\\", argv[1], 2) ||
		    strnequal("//", argv[1], 2))
		{
			fstrcpy(dest_host, argv[1] + 2);
		}
		argc--;
		argv++;
	}

	while ((opt = getopt(argc, argv, "dU:W:")) != EOF)
	{
		switch (opt)
		{
			case 'U':
			{
				char *lp;
				fstrcpy(usr.user_name, optarg);
				if ((lp = strchr(usr.user_name, '%')))
				{
					*lp = 0;
					pstrcpy(password, lp + 1);
					memset(lp + 1, 'X', strlen(password));
					got_pwd = True;
				}
				if (usr.user_name[0] == 0
				    && password[0] == 0)
				{
					null_pwd = True;
				}
				break;
			}

			case 'N':
			{
				null_pwd = True;
			}
			case 'W':
			{
				fstrcpy(usr.domain, optarg);
				break;
			}

			case 'd':
			{
				net_use_add = False;
				break;
			}

			case 'f':
			{
				force_close = True;
				break;
			}

			default:
			{
				report(out_hnd,
				       "net -S \\server [-U user%%pass] [-W domain] [-d] [-f]\n");
				break;
			}
		}
	}

	if (strnequal("\\\\", dest_host, 2))
	{
		fstrcpy(srv_name, dest_host);
	}
	else
	{
		fstrcpy(srv_name, "\\\\");
		fstrcat(srv_name, dest_host);
	}
	strupper(srv_name);

	if (net_use_add)
	{
		BOOL isnew;
		if (null_pwd)
		{
			set_user_password(&usr, True, NULL);
		}
		else
		{
			set_user_password(&usr, got_pwd, password);
		}

		/* paranoia: destroy the local copy of the password */
		ZERO_STRUCT(password);

		report(out_hnd, "Server:\t%s:\tUser:\t%s\tDomain:\t%s\n",
		       srv_name, usr.user_name, usr.domain);
		report(out_hnd, "Connection:\t");

		if (cli_net_use_add(srv_name, &usr, info->reuse,
				    &isnew, cli_info.port) != NULL)
		{
			report(out_hnd, "OK\n");
		}
		else
		{
			report(out_hnd, "FAILED\n");
		}
	}
	else
	{
		BOOL closed;
		report(out_hnd, "Server:\t%s:\tUser:\t%s\tDomain:\t%s\n",
		       srv_name, usr.user_name, usr.domain);
		report(out_hnd, "Connection:\t");

		if (!cli_net_use_del(srv_name, &usr,
				     force_close, &closed))
		{
			report(out_hnd, ": Does not exist\n");
		}
		else if (force_close && closed)
		{
			report(out_hnd, ": Forcibly terminated\n");
		}
		else if (closed)
		{
			report(out_hnd, ": Terminated\n");
		}
		else
		{
			report(out_hnd, ": Unlinked\n");
		}
	}

	/* paranoia: destroy the local copy of the password */
	ZERO_STRUCT(password);

	return 0;
}

/******************************************************************
   allow or disallow automatic connections.  rpctorture, because it
   does not reestablish connections after sys_fork(), fails unless the
   connection is established AFTER the sys_fork()
 ******************************************************************/
static BOOL auto_connect = True;
void cmd_set_no_autoconnect(void)
{
	auto_connect = False;
}

#define CMD_STR 0x1
#define CMD_DBF 0x2
#define CMD_SVC 0x4
#define CMD_TERM 0x8
#define CMD_PASS 0x10
#define CMD_USER 0x20
#define CMD_NOPW 0x40
#define CMD_DBLV 0x80
#define CMD_HELP 0x100
#define CMD_SOCK 0x200
#define CMD_IFACE 0x400
#define CMD_DOM 0x800
#define CMD_PWHASH 0x1000
#define CMD_HOST 0x2000
#define CMD_NAME 0x4000
#define CMD_DBG 0x8000
#define CMD_SCOPE 0x10000
#define CMD_INTER 0x20000

/* command options bitfield */
static uint32 cmd_set_options = CMD_SVC;


static void read_authfile (char *filename, struct ntuser_creds *creds)
{
	FILE *auth;
	fstring buf;
	uint16 len = 0;
	char *ptr, *val, *param;
	int i, action;

#define READAUTH_USER 1
#define READAUTH_PASS 2
#define READAUTH_DOMAIN 3
#define READAUTH_LMHASH 4
#define READAUTH_NTHASH 5
	static struct { char *str; int option; int action; } tokens[] = {
		{ "username",	CMD_USER,	READAUTH_USER },
		{ "password",	CMD_PASS,	READAUTH_PASS },
		{ "domain",	CMD_DOM,	READAUTH_DOMAIN },
		{ "workgroup",	CMD_DOM,	READAUTH_DOMAIN },
		{ "lmhash",	CMD_PWHASH,	READAUTH_LMHASH },
		{ "lm#",	CMD_PWHASH,	READAUTH_LMHASH },
		{ "nthash",	CMD_PWHASH,	READAUTH_NTHASH },
		{ "nt#",	CMD_PWHASH,	READAUTH_NTHASH },
		{ NULL, }
	};

	if ((auth=sys_fopen(filename, "r")) == NULL)
	{
		/* fail if we can't open the credentials file */
		DEBUG(0,("ERROR: Unable to open credentials file!\n"));
		return;
	}

	while (!feof(auth))
	{
		/* get a line from the file */
		if (!fgets (buf, sizeof(buf), auth))
			continue;

		len = strlen(buf);

		/* skip empty lines */
		if ((len) && (buf[len-1]=='\n'))
		{
			buf[len-1] = '\0';
			len--;
		}
		if (len == 0)
			continue;

		/* break up the line into parameter & value.
		   will need to eat a little whitespace possibly */
		param = buf;
		if (!(ptr = strchr(buf, '=')))
			continue;
		val = ptr+1;
		*ptr = '\0';

		/* eat leading white space */
		while ((*val!='\0') && ((*val==' ') || (*val=='\t')))
			val++;

		for (action=i=0; action==0 && tokens[i].str; i++)
		{
			if (!strwicmp(tokens[i].str, param))
			{
				cmd_set_options |= tokens[i].option;
				action = tokens[i].action;
			}
		}
		switch (action)
		{
			case READAUTH_USER:
			{
				fstrcpy(creds->user_name, val);
				break;
			}
			case READAUTH_PASS:
			{
				fstrcpy(creds->pwd.password, val);
				break;
			}
			case READAUTH_DOMAIN:
			{
				fstrcpy(creds->domain, val);
				break;
			}
			case READAUTH_LMHASH:
			{
				pwdb_gethexpwd(val, creds->pwd.smb_lm_pwd, NULL);
				break;
			}
			case READAUTH_NTHASH:
			{
				pwdb_gethexpwd(val, creds->pwd.smb_nt_pwd, NULL);
				break;
			}
			default:; /* ignore */
		}

		ZERO_OBJ(buf);
	}
	fclose(auth);

	return;
}

static uint32 cmd_set(struct client_info *info, int argc, char *argv[])
{
	BOOL interactive = True;
	char *cmd_str = NULL;
	int opt;
	extern char *optarg;
	pstring term_code;
	pstring password;	/* local copy only, if one is entered */

	fstring srv_name;
	char *new_debuglevel = NULL;

	password[0] = 0;
	pwd_init (&(usr.pwd));
	usr_creds = &usr;
	info->reuse = False;
#ifdef KANJI
	pstrcpy(term_code, KANJI);
#else /* KANJI */
	*term_code = 0;
#endif /* KANJI */

	cmd_set_options |= CMD_INTER;

	if (argc > 1 && (*argv[1] != '-'))
	{
		if (strnequal("\\\\", argv[1], 2) ||
		    strnequal("//", argv[1], 2))
		{
			cmd_set_options |= CMD_HOST;
			fstrcpy(cli_info.dest_host, argv[1] + 2);
			strupper(cli_info.dest_host);
		}
		argc--;
		argv++;
	}

	if (argc > 1 && (*argv[1] != '-'))
	{
		cmd_set_options |= CMD_PASS;
		pstrcpy(password, argv[1]);
		memset(argv[1], 'X', strlen(argv[1]));
		argc--;
		argv++;
	}

	while ((opt = getopt(argc, argv,
			     "A:Rs:O:M:S:p:i:Nn:d:l:hI:EB:U:L:t:m:W:T:D:c:")) !=
	       EOF)
	{
		switch (opt)
		{
			case 'R':
			{
				info->reuse = True;
				break;
			}

			case 'p':
			{
				cli_info.port = atoi(optarg);
				break;
			}

			case 'O':
			{
				cmd_set_options |= CMD_SOCK;
				pstrcpy(user_socket_options, optarg);
				break;
			}

			case 'I':
			{
				fprintf (stderr, "warning: -I deprecated, use -S\n");
				/* no break */
			}
			case 'S':
			{
				cmd_set_options |= CMD_HOST;
				fstrcpy(cli_info.dest_host, optarg);
				strupper(cli_info.dest_host);
				if(!strcmp(optarg,"."))
				{
					cmd_set_options |= CMD_NOPW | CMD_PASS;
				}
				break;
			}

			case 'U':
			{
				char *lp;
				cmd_set_options |= CMD_USER;
				fstrcpy(usr.user_name, optarg);
				if ((lp = strchr(usr.user_name, '%')))
				{
					*lp = 0;
					pstrcpy(password, lp + 1);
					cmd_set_options |= CMD_PASS;
					memset(strchr(optarg, '%') + 1, 'X',
					       strlen(lp + 1));
				}
				if (usr.user_name[0] == 0
				    && password[0] == 0)
				{
					cmd_set_options |= CMD_NOPW;
				}
				break;
			}

			/* authfile -- get creds from file */
			case 'A':
			{
				read_authfile (optarg, &usr);
				pstrcpy(password, usr.pwd.password);
				break;
			}

			case 'W':
			{
				cmd_set_options |= CMD_DOM;
				fstrcpy(usr.domain, optarg);
				break;
			}

			case 'E':
			{
				cmd_set_options |= CMD_DBG;
				dbf = stderr;
				break;
			}

			case 'n':
			{
				cmd_set_options |= CMD_NAME;
				pstrcpy(global_myname, optarg);
				break;
			}

			case 'N':
			{
				cmd_set_options |= CMD_NOPW | CMD_PASS;
				break;
			}

			case 'd':
			{
				cmd_set_options |= CMD_DBLV;
				new_debuglevel = optarg;
				break;
			}

			case 'l':
			{
				cmd_set_options |= CMD_INTER;
				slprintf(debugf, sizeof(debugf) - 1,
					 "%s.client", optarg);
				interactive = False;
				break;
			}

			case 'c':
			{
				cmd_set_options |= CMD_STR;
				cmd_str = optarg;
				break;
			}

			case 'h':
			{
				cmd_set_options |= CMD_HELP;
				usage(argv[0]);
				break;
			}

			case 's':
			{
				cmd_set_options |= CMD_SVC;
				pathname_smb_conf = optarg;
				break;
			}

			case 't':
			{
				cmd_set_options |= CMD_TERM;
				pstrcpy(term_code, optarg);
				break;
			}

			case 'm':
			{
				fprintf (stderr, "warning: ignoring '-m %s'\n", optarg);
				break;
			}

			default:
			{
				cmd_set_options |= CMD_HELP;
				usage(argv[0]);
				break;
			}
		}
	}

	/* default to loopback */
	if (!(cmd_set_options & CMD_HOST))
	{
		fstrcpy(cli_info.dest_host, ".");
		cmd_set_options |= CMD_HOST | CMD_NOPW | CMD_PASS;
	}

	if (IS_BITS_SET_ALL(cmd_set_options, CMD_INTER))
	{
		setup_logging(debugf, interactive);
		if (!interactive)
			reopen_logs();
	}

	if (IS_BITS_SET_ALL(cmd_set_options, CMD_SVC))
	{
		lp_load_globals_only();
		cmd_set_options &= ~CMD_SVC;
	}

	init_myname();

	fstrcpy(cli_info.myhostname, global_myname);

	/* Internationialisation fixes - convert username and password to
	   dos codepage. */

	if (new_debuglevel)
	{
		debug_parse_levels(new_debuglevel);
	}

	if (IS_BITS_SET_ALL(cmd_set_options, CMD_INTER))
	{
		load_interfaces();
	}

	DEBUG(10, ("cmd_set: options: %x\n", cmd_set_options));

	if (IS_BITS_SET_ALL(cmd_set_options, CMD_HELP))
	{
		return 0;
	}

	if (IS_BITS_SET_ALL(cmd_set_options, CMD_NOPW))
	{
		set_user_password(&usr, True, NULL);
	}
	else if (IS_BITS_SET_ALL(cmd_set_options, CMD_PWHASH))
	{
		/* from pwd_make_nt_lm_16() */
		usr.pwd.null_pwd = False;
		usr.pwd.cleartext = False;
		usr.pwd.crypted = False;
	}
	else
	{
		set_user_password(&usr,
				  IS_BITS_SET_ALL(cmd_set_options, CMD_PASS),
				  password);
	}

	/* paranoia: destroy the local copy of the password */
	ZERO_STRUCT(password);

	if (strcmp(cli_info.dest_host, "*") == 0)
	{
		/* special case - we want the PDC */
		struct in_addr ip;
		if (!resolve_srv_name(cli_info.dest_host,
				      cli_info.dest_host, &ip))
		{
			report(out_hnd, "ERROR: Failed to find the PDC\n");
			return 1;
		}
	}

	fstrcpy(srv_name, "\\\\");
	fstrcat(srv_name, cli_info.dest_host);
	strupper(srv_name);

	if (auto_connect && !strequal(srv_name, "\\\\."))
	{
		BOOL isnew;
		report(out_hnd, "Server:\t%s:\tUser:\t%s\tDomain:\t%s\n",
		       srv_name, usr.user_name, usr.domain);
		report(out_hnd, "Connection:\t");

		if (cli_net_use_add(srv_name, &usr, info->reuse,
				    &isnew, cli_info.port) != NULL)
		{
			report(out_hnd, "OK\n");
		}
		else
		{
			report(out_hnd, "FAILED\n");
		}
		usr_creds = NULL;
	}
	if (cmd_str != NULL)
	{
		return process(&cli_info, cmd_str);
	}

	return 0;
}

static void read_user_env(struct ntuser_creds *u)
{
	pstring password;

	password[0] = 0;

	if (getenv("USER"))
	{
		char *p;
		fstrcpy(u->user_name, getenv("USER"));

		/* modification to support userid%passwd syntax in the USER var
		   25.Aug.97, jdblair@uab.edu */

		if ((p = strchr(u->user_name, '%')))
		{
			*p = 0;
			pstrcpy(password, p + 1);
			memset(strchr(getenv("USER"), '%') + 1, 'X',
			       strlen(password));
		}
		strupper(u->user_name);
	}

	/* modification to support PASSWD environmental var
	   25.Aug.97, jdblair@uab.edu */
	if (getenv("PASSWD"))
	{
		pstrcpy(password, getenv("PASSWD"));
	}

	if (*u->user_name == 0 && getenv("LOGNAME"))
	{
		fstrcpy(u->user_name, getenv("LOGNAME"));
		strupper(u->user_name);
	}

	set_user_password(u, True, password);

	/* paranoia: destroy the local copy of the password */
	ZERO_STRUCT(password);
}

static void readline_init(void)
{
#ifdef HAVE_READLINE
	/* Initialise GNU Readline */
	rl_readline_name = "rpcclient";
	rl_attempted_completion_function = completion_fn;
	rl_completion_entry_function = complete_cmd_null;

	rl_completer_quote_characters = "\"";

	/* Initialise history list */

	using_history();

#else
	int x;
	x = 0;			/* stop compiler warnings */
#endif /* HAVE_READLINE */
}

static void readline_shutdown(void)
{
#ifdef HAVE_READLINE
	clear_history();
#endif
}


/****************************************************************************
  main program
****************************************************************************/

static void cmd_interp_cleanup(void)
{
	readline_shutdown();

	free_cmd_set_array(commands);
	commands = NULL;

	free_connections();

	vuid_db_close();

	prs_sma_shutdown();

	lp_talloc_free();

	charconv_reset();
}

int command_main(int argc, char *argv[])
{
	uint32 status;
	char *progname;

	in_client = True;
	debug_parse_levels("2");

	add_command_set(general_commands);

	ZERO_STRUCT(usr);

	usr_creds = &usr;

	out_hnd = stdout;

	progname = strrchr(argv[0], '/');
	if (progname)
		progname++;
	else
		progname = argv[0];

	/* XXX - Setting a central log for rpcclient... weird... */
	debug_logfile_for_prog(progname);

	fstrcpy(usr.domain, "");
	fstrcpy(usr.user_name, "");

	fstrcpy(cli_info.myhostname, "");
	fstrcpy(cli_info.dest_host, "");
	cli_info.dest_ip.s_addr = 0;

	ZERO_STRUCT(cli_info.dom.level3_sid);
	ZERO_STRUCT(cli_info.dom.level5_sid);
	fstrcpy(cli_info.dom.level3_dom, "");
	fstrcpy(cli_info.dom.level5_dom, "");

	readline_init();
	tng_libsamba_init();
#ifdef HAVE_SIGPIPE
	BlockSignals(True,SIGPIPE);
#endif
	init_connections();

	generate_wellknown_sids();

	read_user_env(&usr);

	status = cmd_set(&cli_info, argc, argv);

	if (IS_BITS_SET_SOME(cmd_set_options, CMD_HELP | CMD_STR))
	{
		cmd_interp_cleanup();

		report(out_hnd, "Exit Status: %s\n",
		       get_nt_error_msg(status));
		/* unix only has 8 bit error codes - blergh */
		exit(status & 0xFF);
	}

	DEBUG(3, ("%s client started (version %s)\n",
		  timestring(False), VERSION));

	status = process(&cli_info, NULL);

	cmd_interp_cleanup();

	report(out_hnd, "Exit Status: %s\n", get_nt_error_msg(status));

	return status;
}
