/* 
   Unix SMB/Netbios implementation.
   Version 1.9.
   Samba utility functions
   Copyright (C) Andrew Tridgell 1992-1999
   Copyright (C) Luke Kenneth Casson Leighton 1996 - 1999
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful, 
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#include "includes.h"
#include "libsamba.h"
#include "rpcclient.h"
#include "rpc_reg.h"

struct field_info reg_val_types[] =
{
	/* description field unused at the moment */
	{ REGVAL_NONE, "REG_NONE", },
	{ REGVAL_SZ, "REG_SZ", "Unicode string" },
	{ REGVAL_EXPAND_SZ, "REG_EXPAND_SZ", "Unicode string to be var-expanded" },
	{ REGVAL_BINARY, "REG_BINARY", "Arbitrary binary data" },
	{ REGVAL_DWORD, "REG_DWORD", "32-bit integer" },
	{ REGVAL_SZ_MULTI, "REG_SZ_MULTI", "Multiple strings, null-separated" },
	{ REGVAL_FULL_RES_DESCR, "REG_FULL_RESOURCE_DESCRIPTOR", "binary data" },
	{ REGVAL_RES_REQ_LIST, "REG_RESOURCE_REQUIREMENTS_LIST", "binary data" },

	/* for compatibility (we previously used these names) */
	{ REGVAL_SZ, "UNISTR", },
	{ REGVAL_BINARY, "BYTES", },
	{ REGVAL_DWORD, "DWORD", },
	{ REGVAL_NONE }
};

/****************************************************************************
convert a security permissions into a string
****************************************************************************/
static const char *get_reg_val_type_str(uint32 type)
{
	const char *typestr = enum_field_to_str(type, reg_val_types, False);
	if (!typestr)
	{
		static char defaulttype[16];
		slprintf(defaulttype, 15, "hex(%x)", type);
		typestr = defaulttype;
	}
	return typestr;
}


static void print_reg_value(FILE *out_hnd, const char *val_name, 
				uint32 val_type, const BUFFER2 *value)
{
	const char *type;
	fstring valstr;

	type = get_reg_val_type_str(val_type);

	switch (val_type)
	{
		case REGVAL_NONE:
		{
			report(out_hnd, "\t%s:\t%s:\n", val_name, type);
			break;
		}
		case REGVAL_SZ:
		case REGVAL_EXPAND_SZ:
		{
			unibuf_to_ascii(valstr, value->buffer,
					MIN(value->buf_len, sizeof(valstr)-1));
			report(out_hnd, "\t%s:\t%s:\t%s\n", val_name, type, valstr);
			break;
		}

		default: /* unknown */
		case REGVAL_BINARY:
		case REGVAL_FULL_RES_DESCR:
		case REGVAL_RES_REQ_LIST:
		{
			if (value->buf_len <= 8)
			{
				report(out_hnd, "\t%s:\t%s:\t", val_name, type);
				out_data(out_hnd, (const char*)value->buffer,
				         value->buf_len, 8, "");
			}
			else
			{
				report(out_hnd, "\t%s:\t%s:\n", val_name, type);
				out_data(out_hnd, (const char*)value->buffer,
				         value->buf_len, 16, "");
			}
			break;
		}

		case REGVAL_DWORD:
		{
			report(out_hnd, "\t%s:\t%s:\t0x%08x\n", val_name, type, buffer2_to_uint32(value));
			break;
		}

		case REGVAL_SZ_MULTI:
		{
			buffer2_to_multistr(valstr, value, sizeof(valstr));
			report(out_hnd, "\t%s:\t%s:\t%s\n", val_name, type, valstr);
			break;
		}
	}
}

/****************************************************************************
 display structure
 ****************************************************************************/
void display_reg_value_info(FILE *out_hnd, enum action_type action, 
				const char *val_name, 
				uint32 val_type, const BUFFER2 *value)
{
	switch (action)
	{
		case ACTION_HEADER:
		{
			break;
		}
		case ACTION_ENUMERATE:
		{
			print_reg_value(out_hnd, val_name, val_type, value);
			break;
		}
		case ACTION_FOOTER:
		{
			break;
		}
	}
}

/****************************************************************************
 display structure
 ****************************************************************************/
void display_reg_key_info(FILE *out_hnd, enum action_type action, 
				const char *key_name, time_t key_mod_time)
{
	switch (action)
	{
		case ACTION_HEADER:
		{
			break;
		}
		case ACTION_ENUMERATE:
		{
			report(out_hnd, "\t%s\t(%s)\n", 
			        key_name, http_timestring(key_mod_time));
			break;
		}
		case ACTION_FOOTER:
		{
			break;
		}
	}
}

