/* 
 *  Unix SMB/Netbios implementation.
 *  Version 1.9.
 *  RPC Pipe client / server routines
 *  Copyright (C) Andrew Tridgell              1992-2000,
 *  Copyright (C) Luke Kenneth Casson Leighton 1996-2000,
 *  Copyright (C) Sander Striker                    2000,
 *  Copyright (C) Elrond                            2000
 *  
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *  
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software Foundation,
 *  Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */


#include "includes.h"
#include "libsamba.h"
#include "rpc_parse.h"
#include "nterr.h"
#include "sids.h"
#include "others.h"
#include "tng_misc.h"


/*******************************************************************
 samr_make_usr_obj_sd
 ********************************************************************/
uint32 samr_make_usr_obj_sd(SEC_DESC_BUF *buf, DOM_SID *usr_sid,
			    BOOL pw_can_change)
{
	DOM_SID adm_sid;
	DOM_SID act_sid;
	DOM_SID glb_sid;

	SEC_ACL *dacl = NULL;
	SEC_ACE *dace = NULL;
	SEC_ACCESS mask;
	SEC_DESC *sec = NULL;
	int len;

	DEBUG(15, ("samr_make_usr_obj_sd: %d\n", __LINE__));

	dacl = g_new1(SEC_ACL);
	dace = g_new(SEC_ACE, 4);
	sec = g_new1(SEC_DESC);

	if (dacl == NULL || dace == NULL || sec == NULL)
	{
		safe_free(dacl);
		safe_free(dace);
		safe_free(sec);
		return NT_STATUS_NO_MEMORY;
	}

	sid_copy(&adm_sid, global_sid_builtin);
	sid_append_rid(&adm_sid, BUILTIN_ALIAS_RID_ADMINS);

	sid_copy(&act_sid, global_sid_builtin);
	sid_append_rid(&act_sid, BUILTIN_ALIAS_RID_ACCOUNT_OPS);

	sid_copy(&glb_sid, global_sid_everyone);

	mask.mask = 0x2031b;
	if (pw_can_change)
		mask.mask |= USER_ACCESS_CHANGE_PASSWORD;
	make_sec_ace(&dace[0], &glb_sid, 0, mask, 0);
	mask.mask = 0xf07ff;
	make_sec_ace(&dace[1], &adm_sid, 0, mask, 0);
	make_sec_ace(&dace[2], &act_sid, 0, mask, 0);
	mask.mask = 0x20004;
	if (pw_can_change)
		mask.mask |= USER_ACCESS_CHANGE_PASSWORD;
	make_sec_ace(&dace[3], usr_sid, 0, mask, 0);

	make_sec_acl(dacl, 2, 4, dace);

	len = make_sec_desc(sec, 1,
	              SEC_DESC_DACL_PRESENT|SEC_DESC_SELF_RELATIVE,
	              NULL, NULL, NULL, dacl);

	make_sec_desc_buf(buf, len, sec);
	buf->undoc = 0x1;

	return NT_STATUS_NOPROBLEMO;
}


static int compare_gids(const void *pa, const void *pb)
{
	const DOM_GID *a = pa, *b = pb;
	return a->g_rid - b->g_rid;
}

void make_dom_gids(const DOMAIN_GRP *mem, size_t num_members,
		   uint32 users_primary_group_rid,
		   DOM_GID ** ppgids, uint32 *num_created_gids)
{
	uint32 count;
	uint32 i, j;
	DOM_GID *gids = NULL;

	g_return_if_fail(ppgids && num_created_gids);
	g_return_if_fail(num_members==0 || mem);

	*ppgids = NULL;
	*num_created_gids = 0;

	DEBUG(4, ("make_dom_gids: %u (group_rid: 0x%x)\n",
		  num_members, users_primary_group_rid));

	count = num_members;
	if (users_primary_group_rid == 0xffffffff)
		users_primary_group_rid = 0;
	if (users_primary_group_rid)
		count++;

	if (! count)
		return;

	gids = g_new(DOM_GID, count);

	if (gids == NULL)
	{
		DEBUG(0, ("make_dom_gids: malloc for %u entries failed\n",
			  count));
		return;
	}

	i = 0;
	if (users_primary_group_rid)
	{
		gids[i].g_rid = users_primary_group_rid;
		gids[i].attr = 0x7;
		i++;
	}

	for (; i < count; i++, mem++)
	{
		gids[i].g_rid = mem->rid;
		gids[i].attr = mem->attr;
	}

	qsort(gids, count, sizeof(DOM_GID), compare_gids);

	/* Remove dupes */
	for (i = 0, j = 0; j < count; i++, j++)
	{
		if (i != j)
			gids[i] = gids[j];
		while (j+1 < count && gids[i].g_rid == gids[j+1].g_rid)
			j++;
	}
	DEBUG(6, ("Reducing gid count due to dupes from %u to %u\n",
		  count, i));
	count = i;

	*ppgids = gids;
	*num_created_gids = count;
}
