/* 
   Unix SMB/Netbios implementation.
   Version 1.9.
   file closing
   Copyright (C) Andrew Tridgell 1992-1998
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/
/* MKPROTO: smbd_proto.h		(do not remove this line) */

#include "includes.h"
#include <errno.h>
#include "libsamba.h"
#include "smbd.h"
#include "vfs.h"


/****************************************************************************
run a file if it is a magic script
****************************************************************************/
static void check_magic(files_struct *fsp,connection_struct *conn)
{
  const char *fname;

  if (!*lp_magicscript(SNUM(conn)))
    return;

  fname = file_unix_name(fsp);

  DEBUG(5,("checking magic for %s\n", fname));

  {
    const char *p;
    if (!(p = strrchr(fname,'/')))
      p = fname;
    else
      p++;

    if (!strequal(lp_magicscript(SNUM(conn)),p))
      return;
  }

  {
    int ret;
    pstring magic_output;

    if (*lp_magicoutput(SNUM(conn)))
      pstrcpy(magic_output,lp_magicoutput(SNUM(conn)));
    else
      slprintf(magic_output, sizeof(magic_output)-1, "%s.out", fname);

    chmod(fname,0755);
    ret = smbrun(fname,magic_output,False);
    DEBUG(3,("Invoking magic command %s gave %d\n",fname,ret));
    unlink(fname);
  }
}

/****************************************************************************
  Common code to close a file or a directory.
****************************************************************************/

static void close_filestruct(files_struct *fsp)
{   
	connection_struct *conn = fsp->conn;
    
	flush_write_cache(fsp, CLOSE_FLUSH);

	fsp->is_directory = False; 
	fsp->stat_open = False; 
    
	conn->num_files_open--;
	if(fsp->wbmpx_ptr) {  
		free(fsp->wbmpx_ptr);
		fsp->wbmpx_ptr = NULL; 
	}  
}    

/****************************************************************************
 Close a file - possibly invalidating the read prediction.

 If normal_close is 1 then this came from a normal SMBclose (or equivalent)
 operation otherwise it came as the result of some other operation such as
 the closing of the connection. In the latter case printing and
 magic scripts are not run.
****************************************************************************/

static int close_normal_file(files_struct *fsp, BOOL normal_close)
{
	BOOL delete_on_close = fsp->delete_on_close;
	connection_struct *conn = fsp->conn;
	int err = 0;

	remove_pending_lock_requests_by_fid(fsp);

	close_filestruct(fsp);

	if (normal_close && fsp->print_file) {
		print_fsp_end(fsp);
		file_free(fsp);
		return 0;
	}

	/*
	 */

	lock_share_entry_fsp(fsp);
	del_share_mode(fsp);
	unlock_share_entry_fsp(fsp);

	if(EXCLUSIVE_OPLOCK_TYPE(fsp->oplock_type))
		release_file_oplock(fsp);

	locking_close_file(fsp);

	err = fd_close(conn, fsp);
	if (err)
		DEBUG(2, ("Error closing %s: %s\n", file_debug_name(fsp),
			  strerror(err)));

	/* check for magic scripts */
	if (normal_close) {
		check_magic(fsp,conn);
	}

	/*
	 * NT can set delete_on_close of the last open
	 * reference to a file.
	 */
	if (normal_close && delete_on_close) {
		DEBUG(5, ("close_file: file %s. Delete on close was set - deleting file.\n",
			  file_debug_name(fsp)));
		if(SMB_VFS_UNLINK(fsp->conn, file_unix_name(fsp)) != 0) {
			/*
			 * This call can potentially fail as another smbd may have
			 * had the file open with delete on close set and deleted
			 * it when its last reference to this file went away. Hence
			 * we log this but not at debug level zero.
			 */

			DEBUG(5, ("close_file: file %s. Delete on close was "
				  "set and unlink failed with error %s\n",
				  file_debug_name(fsp), strerror(errno) ));
		}
	}

	/*
	 * If there was a modify time outstanding,
	 * try and set it here.
	 */
	if (fsp->pending_modtime)
		set_filetime(conn, file_dos_name(fsp), fsp->pending_modtime);

	file_free(fsp);

	return err;
}

/****************************************************************************
 Close a directory opened by an NT SMB call. 
****************************************************************************/
  
static int close_directory(files_struct *fsp, BOOL normal_close)
{
	remove_pending_change_notify_requests_by_fid(fsp);

	/*
	 * NT can set delete_on_close of the last open
	 * reference to a directory also.
	 */

	if (normal_close && fsp->directory_delete_on_close) {
		BOOL ok = rmdir_internals(fsp->conn, fsp->dos_fsp_name);
		DEBUG(5,("close_directory: %s. Delete on close was set - deleting directory %s.\n",
			file_debug_name(fsp), ok ? "succeeded" : "failed" ));

		/*
		 * Ensure we remove any change notify requests that would
		 * now fail as the directory has been deleted.
		 */

		if(ok)
			remove_pending_change_notify_requests_by_filename(fsp);
	}

	/*
	 * Do the code common to files and directories.
	 */
	close_filestruct(fsp);
	
	file_free(fsp);

	return 0;
}

/****************************************************************************
 Close a file opened with null permissions in order to read permissions.
****************************************************************************/

static int close_statfile(files_struct *fsp, BOOL normal_close)
{
	close_filestruct(fsp);
	
	file_free(fsp);

	return 0;
}

#ifndef NOONS_SMALL_BUILD
static int close_pipe(files_struct *fsp, BOOL normal_close G_GNUC_UNUSED)
{
	close_rpc_pipe_hnd(fsp->pipe_info);
	fsp->pipe_info = NULL;

	close_filestruct(fsp);	
	file_free(fsp);

	return 0;
}
#endif


static const char *conn_service_name(connection_struct * conn)
{
	return lp_servicename(SNUM(conn));
}


/**
 * \brief Close any fsp (directory, file, stat, pipe)
 */
int close_file(files_struct *fsp, BOOL normal_close)
{
	DEBUG(3, ("Closing %s [%s]%s (number of open files before close: %d)"
		  " (fd=%d, fnum=%d)\n",
		  fsp_type_str(fsp), conn_service_name(fsp->conn),
		  file_debug_name(fsp), fsp->conn->num_files_open,
		  fsp->fd, fsp->fnum));

	if(fsp->is_directory)
		return close_directory(fsp, normal_close);
	else if(fsp->stat_open)
		return close_statfile(fsp, normal_close);
	return close_normal_file(fsp, normal_close);
}
