/* 
   Unix SMB/Netbios implementation.
   Version 1.9.
   Directory handling routines
   Copyright (C) Andrew Tridgell 1992-1998
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/
/* MKPROTO: smbd_proto.h		(do not remove this line) */

/**
 * \file smbd/dir.c
 * \brief Basic directory stuff
 */

#include "includes.h"
#include <errno.h>
#include "libsamba.h"
#include "tng_misc.h"
#include "dlinklist.h"
#include "smbd.h"
#include "vfs.h"


/****************************************************************************
 Check a filetype for being valid.
****************************************************************************/

BOOL dir_check_ftype(connection_struct *conn,int mode,SMB_STRUCT_STAT *st,int dirtype)
{
	int mask;
      
	DEBUG(10, ("dir_check_ftype: maskbits=0x%x, mode=0x%x\n",
		   dirtype, mode));

	/* Check the "may have" search bits. */
	if (((mode & ~dirtype) & (aHIDDEN | aSYSTEM | aDIR)) != 0)
		return False;

	/* Check the "must have" bits, which are the may have bits shifted eight */
	/* If must have bit is set, the file/dir can not be returned in search unless the matching
	   file attribute is set */
	mask = ((dirtype >> 8) & (aDIR|aARCH|aRONLY|aHIDDEN|aSYSTEM)); /* & 0x37 */
	if(mask) {
		if((mask & (mode & (aDIR|aARCH|aRONLY|aHIDDEN|aSYSTEM))) == mask)   /* check if matching attribute present */
			return True;
		else
			return False;
	}

	return True;
}

/****************************************************************************
 Get an 8.3 directory entry.
****************************************************************************/

BOOL get_dir_entry(connection_struct *conn,char *mask,int dirtype,char *fname,
                   SMB_OFF_T *size,int *mode,time_t *date,BOOL check_descend)
{
  BOOL found = False;
  SMB_STRUCT_STAT sbuf;
  pstring path;
  pstring pathreal;
  BOOL isrootdir;
  pstring filename;
  BOOL needslash;

  *path = *pathreal = *filename = 0;

  isrootdir = (strequal(conn->dirpath,"./") ||
	       strequal(conn->dirpath,".") ||
	       strequal(conn->dirpath,"/"));
  
  needslash = ( conn->dirpath[strlen(conn->dirpath) -1] != '/');

  if (!conn->dirptr)
    return(False);
  
  while (!found)
  {
    const char *dname = ReadDirName(conn->dirptr);

    DEBUG(6,("readdir on dirptr 0x%lx now at offset %d\n",
          (long)conn->dirptr,TellDir(conn->dirptr)));
      
    if (dname == NULL) 
      return(False);
      
    pstrcpy(filename,dname);      

    /* notice the special *.* handling. This appears to be the only difference
       between the wildcard handling in this routine and in the trans2 routines.
       see masktest for a demo
    */
    if ((strcmp(mask,"*.*") == 0) ||
	mask_match(filename,mask,False) ||
        (name_map_mangle(conn, filename, True, False) &&
         mask_match(filename,mask,False)))
    {
      if (isrootdir && (strequal(filename,"..") || strequal(filename,".")))
        continue;

      if (!is_8_3(filename, False)) {
	      name_map_mangle(conn, filename, True, False);
      }

      pstrcpy(fname,filename);
      *path = 0;
      pstrcpy(path,conn->dirpath);
      if(needslash)
        pstrcat(path,"/");
      pstrcpy(pathreal,path);
      pstrcat(path,fname);
      pstrcat(pathreal,dname);
      if (vfs_stat(conn, dos_to_unix_static(pathreal), &sbuf) != 0)
      {
        DEBUG(5,("Couldn't stat 1 [%s]. Error = %s\n",path, strerror(errno) ));
        continue;
      }
	  
      *mode = dos_mode(conn,pathreal,&sbuf);

      if (!dir_check_ftype(conn,*mode,&sbuf,dirtype)) 
      {
        DEBUG(5,("[%s] attribs didn't match %x\n",filename,dirtype));
        continue;
      }

      *size = sbuf.st_size;
      *date = sbuf.st_mtime;

      DEBUG(0,("get_dir_entry mask=[%s] found %s fname=%s\n",mask, pathreal,fname));
	  
      found = True;
    }
  }

  return(found);
}



typedef struct _Dir
{
	struct _Dir *next, *prev;
	int pos;
	size_t numentries;
	size_t mallocsize;
	char *data;
	char *current;
	connection_struct *conn;
	char *path;
	DIR *sysdirp;
	BOOL use_veto;
} Dir;


static void Dir_load(Dir *dirp)
{
	const char *n;
	size_t used=0;
	connection_struct *conn = dirp->conn;
	DIR *p = dirp->sysdirp;
	BOOL use_veto = dirp->use_veto;

	dirp->numentries = 0;
	while ((n = vfs_readdirname(conn, p)))
	{
		size_t l, oldused;

		l = strlen(n)+1;

		/* If it's a vetoed file, pretend it doesn't even exist */
		if (use_veto && conn && IS_VETO_PATH(conn, n))
			continue;

		if (used + l > dirp->mallocsize)
		{
			size_t s = MAX(used + l, used + 2000);
			if (!g_xrenew(char, &dirp->data, s))
			{
				DEBUG(0,("Out of memory in Dir_load\n"));
				break;
			}
			dirp->mallocsize = s;
		}
		unsafe_strcpy(dirp->data + used, n);
		oldused = used;
		used += unix_to_dos(dirp->data+used);
		DEBUG(9, ("Dir_load: Added entry %u: [%s] (%u)\n",
			  dirp->numentries, dirp->data + oldused,
			  used - oldused));
		if ((used - oldused) == 0)
		{
			DEBUG(3, ("Dir_load: Problems converting string [%s], "
				  "skipping.\n", n));
			continue;
		}
		dirp->numentries++;
	}
	DEBUG(6, ("Dir_load: Loaded %u entries for %s\n",
		  dirp->numentries, dirp->path));
	dirp->current = dirp->data;
	dirp->pos = 0;
}


/*******************************************************************
 Open a directory.
********************************************************************/

void *OpenDir(connection_struct *conn, const char *_name, BOOL use_veto)
{
  Dir *dirp;
  DIR *p = conn->vfs_ops->opendir(_name);

  if (!p) return(NULL);
  dirp = g_new1(Dir);
  if (!dirp) {
    DEBUG(0, ("Out of memory in OpenDir\n"));
    conn->vfs_ops->closedir(conn, p);
    return(NULL);
  }
  dirp->mallocsize = 0;
  dirp->data = NULL;
  dirp->conn = conn;
  dirp->sysdirp = p;
  dirp->use_veto = use_veto;
  dirp->path = strdup(_name);

  DLIST_ADD(conn->open_dirs, dirp);

  Dir_load(dirp);

  return((void *)dirp);
}


/*******************************************************************
 Close a directory.
********************************************************************/

void CloseDir(void *p)
{
	Dir *dirp = (Dir *)p;
	if (!dirp)
		return;
	DLIST_REMOVE(dirp->conn->open_dirs, dirp);
	dirp->conn->vfs_ops->closedir(dirp->conn, dirp->sysdirp);
	safe_free(dirp->path);
	safe_free(dirp->data);
	safe_free(dirp);
}

/*******************************************************************
 Read from a directory.
********************************************************************/

const char *ReadDirName(void *p)
{
  const char *ret;
  Dir *dirp = (Dir *)p;

  if (!dirp || !dirp->current || dirp->pos >= dirp->numentries) return(NULL);

  ret = dirp->current;

	DEBUG(9, ("ReadDirName(%s) at pos %u gives %s\n", dirp->path,
		  dirp->pos, ret));

  dirp->current = skip_string(dirp->current,1);
  dirp->pos++;

  return(ret);
}


/*******************************************************************
 Seek a dir.
********************************************************************/

BOOL SeekDir(void *p,int pos)
{
  Dir *dirp = (Dir *)p;

  if (!dirp) return(False);

  if (pos < dirp->pos) {
    dirp->current = dirp->data;
    dirp->pos = 0;
  }

  while (dirp->pos < pos && ReadDirName(p)) ;

  return(dirp->pos == pos);
}

/*******************************************************************
 Tell a dir position.
********************************************************************/

int TellDir(void *p)
{
  Dir *dirp = (Dir *)p;

  if (!dirp) return(-1);
  
  return(dirp->pos);
}


/**
 * \brief Reload any "Dir"s, that might be affected
 */
void DirReload(connection_struct *conn, const char *_name)
{
	Dir *p;

	for(p = conn->open_dirs; p; p = p->next)
	{
		int old_pos;
		if (!strcsequal(p->path, _name))
			continue;

		p->conn->vfs_ops->rewinddir(p->conn, p->sysdirp);

		old_pos = TellDir(p);
		Dir_load(p);
		if (!SeekDir(p, old_pos))
			SeekDir(p, 0);
	}
}
