/* 
   Unix SMB/Netbios implementation.
   Version 1.9.
   error packet handling
   Copyright (C) Andrew Tridgell 1992-1998
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/
/* MKPROTO: smbd_proto.h		(do not remove this line) */

#include "includes.h"
#include <errno.h>
#include "libsamba.h"
#include "byteorder.h"
#include "libsmb.h"
#include "nterr.h"
#include "doserr.h"
#include "smbd.h"


/* these can be set by some functions to override the error codes */
int unix_ERR_class=SMB_SUCCESS;
int unix_ERR_code=0;
NTSTATUS unix_ERR_ntstatus = STATUS_SUCCESS;

static void error_base_debug(int level, char *inbuf,
			     const char *file, int line)
{
	int cmd = CVAL(inbuf,smb_com);

	DEBUG(level, ("packet with error set for cmd=0x%x (%s) from %s:%d: ",
		      cmd, smb_fn_name(cmd),
		      src_filename_shorten(file), line));
}

static void error_check_flags2(char *outbuf, const char *file, int line)
{
	uint16 flgs2 = SVAL(outbuf, smb_flg2);

	if (flgs2 & FLAGS2_32_BIT_ERROR_CODES)
	{
		DEBUG(1, ("FIXME: errr-packet called from %s:%d "
			  "with FLAGS2_32_BIT_ERROR_CODES set\n",
			  src_filename_shorten(file), line));
		flgs2 ^= FLAGS2_32_BIT_ERROR_CODES;
		SSVAL(outbuf, smb_flg2, flgs2);
	}
}

static void cifs_error_put_dos(char *outbuf, uchar error_class, uint16 error_code)
{
	SSVAL(outbuf, smb_flg2,
	      SVAL(outbuf, smb_flg2) & ~FLAGS2_32_BIT_ERROR_CODES);
	SCVAL(outbuf,smb_rcls,error_class);
	SSVAL(outbuf,smb_err,error_code);  

	DEBUGADD(3, ("eclass=%d ecode=%d [%s]\n",
		     error_class, error_code, smb_errstr(outbuf)));
}

static void cifs_error_put_nt(char *outbuf, NTSTATUS error_code)
{
	SSVAL(outbuf, smb_flg2,
	      SVAL(outbuf, smb_flg2) | FLAGS2_32_BIT_ERROR_CODES);
	SIVAL(outbuf,smb_rcls,error_code);

	DEBUGADD(3, ("ntstatus=%08x [%s]\n",
		     error_code, smb_errstr(outbuf)));
}

static void cifs_error_put_nt_as_dos(char *outbuf, NTSTATUS ntstatus)
{
	uint8 eclass;
	uint32 ecode;

	DEBUGADD(4, ("(mapped from %s) ", get_nt_error_msg(ntstatus)));

	ntstatus_to_dos(ntstatus, &eclass, &ecode);
	cifs_error_put_dos(outbuf, eclass, ecode);
}

/****************************************************************************
  create an error packet from a cached error.
****************************************************************************/
int cached_error_packet(char *inbuf,char *outbuf,files_struct *fsp,
			const char *file,int line)
{
  write_bmpx_struct *wbmpx = fsp->wbmpx_ptr;

  int32 eclass = wbmpx->wr_errclass;
  int32 err = wbmpx->wr_error;

  /* We can now delete the auxiliary struct */
  free((char *)wbmpx);
  fsp->wbmpx_ptr = NULL;
  return error_packet(inbuf,outbuf,eclass,err,file,line);
}


/****************************************************************************
  create an error packet from errno
****************************************************************************/
int unix_error_packet(char *inbuf,char *outbuf,int def_class,uint32 def_code,
		      const char *file, int line)
{
  int save_errno = errno;
  int eclass=def_class;
  int ecode=def_code;
  NTSTATUS ntstatus = STATUS_SUCCESS;
  int i=0;

	uint16 flgs2 = SVAL(outbuf, smb_flg2);

	if (flgs2 & FLAGS2_32_BIT_ERROR_CODES)
	{
		DEBUG(1, ("FIXME: UNIXERROR(%d, %u) called from %s:%d "
			  "with FLAGS2_32_BIT_ERROR_CODES set\n",
			  def_class, def_code, file, line));
		flgs2 ^= FLAGS2_32_BIT_ERROR_CODES;
		SSVAL(outbuf, smb_flg2, flgs2);
	}

	ntstatus = unix_ERR_ntstatus;
	if (!NT_STATUS_IS_OK(ntstatus) && (unix_ERR_class == SMB_SUCCESS))
	{
		/* only ntcode is set */
		unix_ERR_ntstatus = STATUS_SUCCESS;
		return error_packet_nt(inbuf, outbuf, ntstatus, file, line);
	}

  if (unix_ERR_class != SMB_SUCCESS)
    {
      eclass = unix_ERR_class;
      ecode = unix_ERR_code;
      unix_ERR_class = SMB_SUCCESS;
      unix_ERR_code = 0;
    }
  else
    {
      if (save_errno != 0)
	DEBUG(3, ("unix error string = %s\n", strerror(save_errno)));

      while (unix_dos_nt_errmap[i].unix_error != 0)
      {
	    if (unix_dos_nt_errmap[i].unix_error == save_errno)
	    {
	      eclass = unix_dos_nt_errmap[i].dos_class;
	      ecode = unix_dos_nt_errmap[i].dos_code;
	      ntstatus = unix_dos_nt_errmap[i].nt_error;
	      break;
	    }
	  i++;
      }
    }

  if (! NT_STATUS_IS_OK(ntstatus))
    return error_packet_nt_and_dos(inbuf, outbuf, ntstatus, eclass, ecode, file, line);

  return(error_packet(inbuf,outbuf,eclass,ecode,file,line));
}


/****************************************************************************
  create an error packet. Normally called using the ERROR() macro
****************************************************************************/
int error_packet(char *inbuf,char *outbuf,int error_class,uint32 error_code,
		 const char *file, int line)
{
  int outsize = set_message(outbuf,0,0,True);
  int cmd = CVAL(inbuf,smb_com);
  int flgs2 = SVAL(outbuf,smb_flg2); 

  DEBUG(3, ("error packet for cmd=0x%x (%s) from %s:%d: ",
	    cmd, smb_fn_name(cmd), src_filename_shorten(file), line));

  if ((flgs2 & FLAGS2_32_BIT_ERROR_CODES) == FLAGS2_32_BIT_ERROR_CODES)
  {
    SIVAL(outbuf,smb_rcls,error_code);

    DEBUGADD(3, ("ntstatus=%08x [%s]\n",
		 error_code, smb_errstr(outbuf)));
  }
  else
  {
    SCVAL(outbuf,smb_rcls,error_class);
    SSVAL(outbuf,smb_err,error_code);  

    DEBUGADD(3, ("eclass=%d ecode=%d [%s]\n",
		 error_class, error_code, smb_errstr(outbuf)));
  }

  return outsize;
}

int error_packet_nt(char *inbuf, char *outbuf, uint32 ntstatus,
		    const char *file, int line)
{
	uint16 flgs2 = SVAL(outbuf, smb_flg2);

	DEBUG(3, ("ERROR_NT(%s) from %s:%d\n", get_nt_error_msg(ntstatus),
		  file, line));

	if (flgs2 & FLAGS2_32_BIT_ERROR_CODES)
	{
		DEBUG(1, ("FIXME: ERROR_NT(%s) called from %s:%d "
			  "with FLAGS2_32_BIT_ERROR_CODES set\n",
			  get_nt_error_msg(ntstatus), file, line));
		flgs2 ^= FLAGS2_32_BIT_ERROR_CODES;
		SSVAL(outbuf, smb_flg2, flgs2);
	}

	if (global_client_caps & CAP_STATUS32)
	{
		SSVAL(outbuf, smb_flg2, flgs2 | FLAGS2_32_BIT_ERROR_CODES);
		return error_packet(inbuf, outbuf, 0, ntstatus, file, line);
	}
	else
	{
		uint8 eclass;
		uint32 ecode;

		ntstatus_to_dos(ntstatus, &eclass, &ecode);
		return error_packet(inbuf, outbuf, eclass, ecode, file, line);
	}
}

int error_packet_nt_and_dos(char *inbuf, char *outbuf,
			    NTSTATUS ntstatus,
			    int error_class, uint32 error_code,
			    const char *file, int line)
{
	DEBUG(3, ("ERROR_BOTH(%s, %d, %u) from %s:%d\n",
		  get_nt_error_msg(ntstatus), error_class, error_code,
		  file, line));

	if (global_client_caps & CAP_STATUS32)
		return error_packet_nt(inbuf, outbuf, ntstatus, file, line);
	else
		return error_packet(inbuf, outbuf, error_class, error_code, file, line);
}


void smbd_error_put(char *inbuf, char *outbuf, NTSTATUS ntstatus,
		    const char *file, int line)
{
	error_check_flags2(outbuf, file, line);
	error_base_debug(3, inbuf, file, line);

	if (global_client_caps & CAP_STATUS32)
		cifs_error_put_nt(outbuf, ntstatus);
	else
		cifs_error_put_nt_as_dos(outbuf, ntstatus);
}

void smbd_error_put_both(char *inbuf, char *outbuf,
			 NTSTATUS ntstatus,
			 uchar error_class, uint16 error_code,
			 const char *file, int line)
{
	error_check_flags2(outbuf, file, line);
	error_base_debug(3, inbuf, file, line);

	if (global_client_caps & CAP_STATUS32)
		cifs_error_put_nt(outbuf, ntstatus);
	else
		cifs_error_put_dos(outbuf, error_class, error_code);
}


/****************************************************************************
 *  Utility function to set bad path error.
 ****************************************************************************/

void set_bad_path_error(int err, BOOL bad_path)
{
	DEBUG(4, ("set_bad_path_error(%s(%d), %s)\n", strerror(err),
		  err, BOOLSTR(bad_path)));

	if(bad_path && (err == ENOENT || err == ENOTDIR))
	{
		unix_ERR_class = ERRDOS;
		unix_ERR_code = ERRbadpath;
		unix_ERR_ntstatus = STATUS_OBJECT_PATH_NOT_FOUND;
	}
}
