/* 
 *  Unix SMB/Netbios implementation.
 *  RPC Pipe Proxy code in smbd
 *  Copyright (C) Andrew Tridgell              1992-1998
 *  Copyright (C) Luke Kenneth Casson Leighton 1996-2001,
 *  Copyright (C) Paul Ashton                  1997-1998.
 *  
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *  
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software Foundation,
 *  Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
/* MKPROTO: smbd_proto.h		(do not remove this line) */

#include "includes.h"
#include "libsamba.h"
#include "tng_misc.h"
#include "byteorder.h"
#include "dlinklist.h"
#include "rpc_parse.h"
#include "rpc_client.h"
#include "smbd.h"
#include "others.h"

#define	PIPE		"\\PIPE\\"
#define	PIPELEN		strlen(PIPE)

static pipes_struct *chain_p;
static int pipes_open;

#ifndef MAX_OPEN_PIPES
#define MAX_OPEN_PIPES 64
#endif

static pipes_struct *Pipes;
static struct bitmap *bmap;

/* this must be larger than the sum of the open files and directories */
static int pipe_handle_offset;

/****************************************************************************
 Set the pipe_handle_offset. Called from smbd/files.c
****************************************************************************/

void set_pipe_handle_offset(int max_open_files)
{
	if (max_open_files < 0x7000)
		pipe_handle_offset = 0x7000;
	else
		pipe_handle_offset = max_open_files + 10;	/* For safety. :-) */
}

/****************************************************************************
  reset pipe chain handle number
****************************************************************************/
void reset_chain_p(void)
{
	chain_p = NULL;
}

/****************************************************************************
  initialise pipe handle states...
****************************************************************************/
void init_rpc_pipe_hnd(void)
{
	bmap = bitmap_allocate(MAX_OPEN_PIPES);
	if (!bmap)
	{
		exit_server("out of memory in init_rpc_pipe_hnd\n");
	}
}


pipes_struct *open_rpc_pipe_p(connection_struct * conn,
			      const char *pipe_name, uint16 vuid)
{
	int i;
	pipes_struct *p;
	static int next_pipe;
	struct msrpc_local *m = NULL;
	extern fstring remote_machine;

	DEBUG(4, ("Open pipe requested %s by [%x] (pipes_open=%d)\n",
		  pipe_name, vuid, pipes_open));

	if (! (vuid_get_flags(vuid) & VUID_PUBLIC))
	{
		DEBUG(1, ("open_rpc_pipe_p: invalid vuid 0x%x\n", vuid));
		return NULL;
	}

	/* not repeating pipe numbers makes it easier to track things in 
	   log files and prevents client bugs where pipe numbers are reused
	   over connection restarts */
	if (next_pipe == 0)
	{
		next_pipe = (sys_getuniqid() ^ time(NULL)) % MAX_OPEN_PIPES;
	}

	i = bitmap_find(bmap, next_pipe);

	if (i == -1)
	{
		DEBUG(0, ("ERROR! Out of pipe structures\n"));
		return NULL;
	}

	next_pipe = (i + 1) % MAX_OPEN_PIPES;

	for (p = Pipes; p; p = p->next)
	{
		DEBUG(5, ("open pipes: name %s pnum=%x\n", p->name, p->pnum));
	}

	{
		BOOL is_new;
		become_root();	/* to make pipe connection */
		m = ncalrpc_l_use_add(remote_machine, pipe_name,
				vuid, False, &is_new);
		unbecome_root();
		if (m == NULL)
		{
			DEBUG(5, ("open pipes: msrpc redirect failed\n"));
			return NULL;
		}
	}

	p = g_new(pipes_struct, 1);
	if (!p)
		return NULL;

	ZERO_STRUCTP(p);
	DLIST_ADD(Pipes, p);

	bitmap_set(bmap, i);
	i += pipe_handle_offset;

	pipes_open++;

	p->pnum = i;
	p->m = m;

	p->device_state = 0;
	p->priority = 0;
	p->vuid = vuid;

	fstrcpy(p->name, pipe_name);

	DEBUG(4, ("Opened pipe %s with handle %x (pipes_open=%d)\n",
		  pipe_name, i, pipes_open));

	chain_p = p;

	/* OVERWRITE p as a temp variable, to display all open pipes */
	for (p = Pipes; p; p = p->next)
	{
		DEBUG(5, ("open pipes: name %s pnum=%x\n", p->name, p->pnum));
	}

	return chain_p;
}



/****************************************************************************
  wait device state on a pipe.  exactly what this is for is unknown...
****************************************************************************/
BOOL wait_rpc_pipe_hnd_state(pipes_struct * p, uint16 priority)
{
	if (p == NULL)
		return False;

	DEBUG(3,
	      ("%s Setting pipe wait state priority=%x on pipe (name=%s)\n",
	       timestring(False), priority, p->name));

	p->priority = priority;
	return True;
}


/****************************************************************************
  set device state on a pipe.  exactly what this is for is unknown...
****************************************************************************/
BOOL set_rpc_pipe_hnd_state(pipes_struct * p, uint16 device_state)
{
	if (p == NULL)
		return False;

	DEBUG(3, ("%s Setting pipe device state=%x on pipe (name=%s)\n",
		  timestring(False), device_state, p->name));

	p->device_state = device_state;
	return True;
}

/****************************************************************************
  close an rpc pipe
****************************************************************************/
BOOL close_rpc_pipe_hnd(pipes_struct * p)
{
	if (!p)
	{
		DEBUG(0, ("Invalid pipe in close_rpc_pipe_hnd\n"));
		return False;
	}

	bitmap_clear(bmap, p->pnum - pipe_handle_offset);

	pipes_open--;

	DEBUG(4, ("closed pipe name %s pnum=%x (pipes_open=%d)\n",
		  p->name, p->pnum, pipes_open));

	DLIST_REMOVE(Pipes, p);

	if (p->m != NULL)
	{
		DEBUG(4, ("closing msrpc redirect\n"));
		ncalrpc_l_use_unref(p->m);
	}

	ZERO_STRUCTP(p);
	free(p);

	return True;
}

/****************************************************************************
 get an rpc pipe 
****************************************************************************/
pipes_struct *get_rpc_pipe_p(char *buf, int where)
{
	int pnum = SVAL(buf, where);

	if (chain_p)
		return chain_p;

	return get_rpc_pipe(pnum);
}

/****************************************************************************
  close an rpc pipe
****************************************************************************/
pipes_struct *get_rpc_pipe(int pnum)
{
	pipes_struct *p;

	DEBUG(4, ("search for pipe pnum=%x\n", pnum));

	for (p = Pipes; p; p = p->next)
	{
		DEBUG(5, ("pipe name %s pnum=%x (pipes_open=%d)\n",
			  p->name, p->pnum, pipes_open));
	}

	for (p = Pipes; p; p = p->next)
	{
		if (p->pnum == pnum)
		{
			chain_p = p;
			return p;
		}
	}

	return NULL;
}


/*******************************************************************
 entry point from msrpc to smb.  adds data received to pdu; checks
 pdu; hands pdu off to msrpc, which gets a pdu back (except in the
 case of the RPC_BINDCONT pdu).
 ********************************************************************/
BOOL read_then_write_pipe(pipes_struct * p, char *data, int len,
		    char **rdata, int *rlen, BOOL *pipe_outstanding)
{
	int selrtn;

	DEBUG(10, ("rpc_to_smb_readwrite: len %d\n", len));

	if (write_to_pipe(p, data, len) != len)
	{
		return False;
	}

	if ((*rlen) == 0)
	{
		return False;
	}

	if (!g_xrenew(char, rdata, *rlen))
	{
		return False;
	}

	/*
	 * timeout waiting for the rest for 30 seconds
	 */
	(*rlen) = read_from_pipe(p, (*rdata), (*rlen));
	if ((*rlen) < 0)
	{
		return False;
	}
	if (!g_xrenew(char, rdata, *rlen))
	{
		return False;
	}

	/* now check whether there is outstanding data on the pipe.
	 * this is needed to as to report a STATUS message back to
	 * the NT client.  yes, NT clients fail to operate if you
	 * don't set a STATUS_BUFFER_OVERFLOW warning on the SMBtrans
	 * response.  *dur*!  what's msrpc got to do with smb, ANYWAY!!
	 */

	selrtn = read_data_outstanding(p->m->fd, 0);

	/* Check if error */
	if (selrtn == -1)
	{
		/* something is wrong. Maybe the socket is dead? */
		return False;
	}

	*pipe_outstanding = (selrtn > 0);

	return True;
}

/****************************************************************************
writes data to a pipe.
****************************************************************************/
ssize_t write_to_pipe(pipes_struct * p, char *data, size_t n)
{
	DEBUG(6, ("write_to_pipe: %x", p->pnum));
	DEBUG(6, ("name: %s len: %d", p->name, n));

	dump_data(50, data, n);

	return write_socket(p->m->fd, data, n);
}


/****************************************************************************
 reads data from a pipe.

 headers are interspersed with the data at regular intervals.  by the time
 this function is called, the start of the data could possibly have been
 read by an SMBtrans (file_offset != 0).

 ****************************************************************************/
int read_from_pipe(pipes_struct * p, char *data, int max_len)
{
	DEBUG(6, ("read_from_pipe: %x name: %s len: %d",
				p->pnum, p->name, max_len));

	if (!p)
	{
		DEBUG(6, ("pipe not open\n"));
		return -1;
	}

	/* read a minimum of 1 byte */
	return read_socket_with_timeout(p->m->fd, data, 1, max_len, 30000);
}
