/*
   Unix SMB/Netbios implementation.
   handle SMBsessionsetup
   Copyright (C) Luke Kenneth Casson Leighton 1996-2001
   Copyright (C) Andrew Tridgell 1992-1998
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/
/* MKPROTO: smbd_proto.h		(do not remove this line) */

#include "includes.h"
#include <pwd.h>
#include "libsamba.h"
#include "byteorder.h"
#include "libsmb.h"
#include "rpc_client.h"
#include "nterr.h"
#include "doserr.h"
#include "smbprofile.h"
#include "smbd.h"
#include "others.h"

/* look in server.c for some explanation of these variables */
extern int Protocol;
extern int max_send;
pstring sesssetup_user="";
uint32 global_client_caps = 0;


/****************************************************************************
 always return an error: it's just a matter of which one...
 ****************************************************************************/
static int session_trust_account(connection_struct * conn, char *inbuf,
				 char *outbuf, char *user, char *domain,
				 char *smb_passwd, int smb_passlen,
				 char *smb_nt_passwd, int smb_nt_passlen)
{
	uchar last_chal[8];
	uint32 status = 0x0;

	if (lp_security() != SEC_USER)
	{
		DEBUG(0,
		      ("session_trust_account: Trust account %s only supported with security = user\n",
		       user));
		SSVAL(outbuf, smb_flg2, FLAGS2_32_BIT_ERROR_CODES);
		return (ERROR_NT(NT_STATUS_LOGON_FAILURE));
	}

	if (last_challenge(last_chal))
	{
		NET_USER_INFO_3 info3;
		ZERO_STRUCT(info3);
		status = check_domain_security(user, domain,
					       last_chal,
					       (uchar *) smb_passwd,
					       smb_passlen,
					       (uchar *) smb_nt_passwd,
					       smb_nt_passlen, &info3);
	}
	else
	{
		status = 0xc0000000 | NT_STATUS_LOGON_FAILURE;
	}

	if (status == 0x0)
	{
		status = 0xc0000000 | NT_STATUS_LOGON_FAILURE;
	}

	SSVAL(outbuf, smb_flg2, FLAGS2_32_BIT_ERROR_CODES);
	return ERROR_NT(status);
}

/****************************************************************************
 Return a bad password error configured for the correct client type.
****************************************************************************/

static int bad_password_error(char *inbuf, char *outbuf)
{
	enum remote_arch_types ra_type = get_remote_arch();

	if (((ra_type == RA_WINNT) || (ra_type == RA_WIN2K)) &&
	    (global_client_caps & (CAP_NT_SMBS | CAP_STATUS32)))
	{
		return (ERROR_NT(NT_STATUS_LOGON_FAILURE));
	}

	return (ERROR_DOS(ERRSRV, ERRbadpw));
}


/****************************************************************************
reply to a session setup command
****************************************************************************/

int reply_sesssetup_and_X(connection_struct * conn, char *inbuf, char *outbuf,
			  int length, int bufsize)
{
	uint16 sess_vuid;
	NET_USER_INFO_3 info3;
	gid_t gid;
	uid_t uid;
	int smb_bufsize;
	int smb_apasslen = 0;
	pstring smb_apasswd;
	int smb_ntpasslen = 0;
	pstring smb_ntpasswd;
	pstring domain;
	pstring user;
	pstring orig_user;
	fstring native_os;
	fstring native_lanman;
	BOOL guest = False;
	static BOOL done_sesssetup = False;
	BOOL doencrypt = SMBENCRYPT();
	START_PROFILE(SMBsesssetupX);

	ZERO_STRUCT(info3);

	*smb_apasswd = 0;
	*smb_ntpasswd = 0;

	unsafe_strcpy(domain, "");

	smb_bufsize = SVAL(inbuf, smb_vwv2);

	DEBUG(10, ("Protocol Level: %d\n", Protocol));

	if (Protocol < PROTOCOL_NT1)
	{
		smb_apasslen = SVAL(inbuf, smb_vwv7);
		if (smb_apasslen > MAX_PASS_LEN)
		{
			overflow_attack(smb_apasslen);
			return(ERROR_DOS(ERRDOS,ERRbuftoosmall));
		}

		memcpy(smb_apasswd, smb_buf(inbuf), smb_apasslen);
		smb_apasswd[smb_apasslen] = 0;
		srvstr_pull_buf(inbuf, user, smb_buf(inbuf)+smb_apasslen, sizeof(user), STR_TERMINATE|STR_ALIGN);

		if (!doencrypt && (lp_security() != SEC_SERVER))
		{
			smb_apasslen = strlen(smb_apasswd);
		}

		if (lp_server_ntlmv2() == True)
		{
			DEBUG(1, ("NTLMv2-only accepted with NT LANMAN 1.0 and above.\n"
				  "user %s attempted down-level SMB connection\n",
				  user));
			return (ERROR_DOS(ERRSRV, ERRbadpw));
		}
	}
	else
	{
		uint16 passlen1 = SVAL(inbuf, smb_vwv7);
		uint16 passlen2 = SVAL(inbuf, smb_vwv8);
		enum remote_arch_types ra_type = get_remote_arch();
		char *p = smb_buf(inbuf);

		if (global_client_caps == 0)
			global_client_caps = IVAL(inbuf, smb_vwv11);

		/* client_caps is used as final determination if client is NT or Win95. 
		   This is needed to return the correct error codes in some
		   circumstances.
		 */

		if (ra_type == RA_WINNT || ra_type == RA_WIN2K
		    || ra_type == RA_WIN95)
		{
			if (!
			    (global_client_caps &
			     (CAP_NT_SMBS | CAP_STATUS32)))
			{
				set_remote_arch(RA_WIN95);
			}
		}

		if (passlen1 != 24 && passlen2 != 24)
			doencrypt = False;

		if (passlen1 > MAX_PASS_LEN)
		{
			overflow_attack(passlen1);
			return(ERROR_DOS(ERRDOS,ERRbuftoosmall));
		}

		passlen1 = MIN(passlen1, MAX_PASS_LEN);
		passlen2 = MIN(passlen2, MAX_PASS_LEN);

		if (!doencrypt)
		{
			/* both Win95 and WinNT stuff up the password
			 * lengths for non-encrypting systems. Uggh. 
			 * if passlen1==24 its a win95 system, and its 
			 * setting the password length incorrectly. Luckily 
			 * it still works with the default code because 
			 * Win95 will null terminate the password anyway 
			 * if passlen1>0 and passlen2>0 then maybe its a 
			 * NT box and its setting passlen2 to some random
			 * value which really stuffs things up. we need to 
			 * fix that one.  
			 *
			 * LKCLXXXX: the random value can be random 16 bit.
			 * old test used to have ... && passlen <= 24)
			 * which of course fails most of the time.
			 *
			 * so we increase this to 128 in order to detect
			 * NTLMv2.  *sigh*
			 */

			if (passlen1 > 0 && passlen2 > 0 && passlen2 > 128)
				passlen2 = 0;
		}

		if (lp_restrict_anonymous())
		{
			/* there seems to be no reason behind the differences in MS clients formatting
			 * various info like the domain, NativeOS, and NativeLanMan fields. Win95
			 * in particular seems to have an extra null byte between the username and the
			 * domain, or the password length calculation is wrong, which throws off the
			 * string extraction routines below.  This makes the value of domain be the
			 * empty string, which fails the restrict anonymous check further down.
			 * This compensates for that, and allows browsing to work in mixed NT and
			 * win95 environments even when restrict anonymous is true. AAB
			 */
			dump_data(100, p, 0x70);
			DEBUG(9, ("passlen1=%d, passlen2=%d\n", passlen1, passlen2));
			if (ra_type == RA_WIN95 && !passlen1 && !passlen2 && p[0] == 0 && p[1] == 0) {
				DEBUG(0, ("restrict anonymous parameter used in a win95 environment!\n"));
				DEBUG(0, ("client is win95 and broken passlen1 offset -- attempting fix\n"));
				DEBUG(0, ("if win95 cilents are having difficulty browsing, you will be unable to use restrict anonymous\n"));
				passlen1 = 1;
			}
		}

		if (doencrypt
		    || ((lp_security() == SEC_SERVER)
			|| (lp_security() == SEC_DOMAIN)))
		{
			/* Save the lanman2 password and the NT md4 password. */
			smb_apasslen = passlen1;
			memcpy(smb_apasswd, p, smb_apasslen);
			smb_apasswd[smb_apasslen] = 0;
			smb_ntpasslen = passlen2;
			memcpy(smb_ntpasswd, p + passlen1, smb_ntpasslen);
			smb_ntpasswd[smb_ntpasslen] = 0;

			/*
			 * Ensure the plaintext passwords are in UNIX format.
			 */
			if (!doencrypt)
			{
				dos_to_unix(smb_apasswd, True);
				dos_to_unix(smb_ntpasswd, True);
			}

		} else {
			/* we use the first password that they gave */
			smb_apasslen = passlen1;
			StrnCpy(smb_apasswd, p, smb_apasslen);
			/*
			 * Ensure the plaintext password is in UNIX format.
			 */
			dos_to_unix(smb_apasswd, True);

			/* trim the password */
			smb_apasslen = strlen(smb_apasswd);

			/* wfwg sometimes uses a space instead of a null */
			if (strequal(smb_apasswd, " ")) {
				smb_apasslen = 0;
				*smb_apasswd = 0;
			}
		}

		if (passlen2 == 0 && smb_apasslen == 0 && ra_type == RA_WIN95)
		{
			/* work-around for win95 NULL sessions, where NULL password is
			   actually put in the data stream before the domain name etc */
			p++;
		}
		else
		{
			p += passlen1 + passlen2;
		}

		p += srvstr_pull_buf(inbuf, user, p, sizeof(user), STR_TERMINATE|STR_ALIGN);
		p += srvstr_pull_buf(inbuf, domain, p, sizeof(domain), STR_TERMINATE|STR_ALIGN);
		p += srvstr_pull_buf(inbuf, native_os, p, sizeof(native_os), STR_TERMINATE|STR_ALIGN);
		p += srvstr_pull_buf(inbuf, native_lanman, p, sizeof(native_lanman), STR_TERMINATE|STR_ALIGN);

		DEBUG(3, ("passlen: %d %d Domain=[%s] NativeOS=[%s] NativeLanMan=[%s]\n",
			  passlen1, passlen2,
			  domain, native_os, native_lanman));
	}


	DEBUG(3, ("sesssetupX: domain=[%s] name=[%s] passlens: %d %d\n",
		  domain, user, smb_apasslen, smb_ntpasslen));

	/*
	 * If name ends in $ it's a trust account.
	 */

	if (*user && (user[strlen(user) - 1] == '$') && (smb_apasslen == 24)
	    && (smb_ntpasslen == 24))
	{
		return session_trust_account(conn, inbuf, outbuf, user,
					     domain,
					     smb_apasswd, smb_apasslen,
					     smb_ntpasswd, smb_ntpasslen);
	}

	if (done_sesssetup && lp_restrict_anonymous()) {
		/* tests show that even if browsing is done over already validated connections
		 * without a username and password the domain is still provided, which it
		 * wouldn't be if it was a purely anonymous connection.  So, in order to
		 * restrict anonymous, we only deny connections that have no session
		 * information.  If a domain has been provided, then it's not a purely
		 * anonymous connection. AAB
		 */
		if (!*user && !*smb_apasswd && !*domain) {
			DEBUG(0,
			      ("restrict anonymous is True and anonymous "
			       "connection attempted. Denying access.\n"));
			END_PROFILE(SMBsesssetupX);
			return (ERROR_DOS(ERRDOS, ERRnoaccess));
		}
	}

	/*
	 * Save the username before mapping. We will use
	 * the original username sent to us for security=server
	 * and security=domain checking.
	 */

	pstrcpy(orig_user, user);

	/* If no username is sent use the guest account */
	if (!*user) {
		pstrcpy(user, lp_guestaccount(-1));
		guest = True;
	}

	/*
	 * In share level security, only overwrite sesssetup_use if
	 * it's a non null-session share. Helps keep %U and %G
	 * working.
	 */

	if ((lp_security() != SEC_SHARE) || (*user && !guest))
		pstrcpy(sesssetup_user, user);

	reload_services(True);

	map_nt_and_unix_username(domain, user, user, NULL);

	add_session_user(user);

	/*
	 * Check if the given username was the guest user with no password.
	 */

	if (!guest && strequal(user, lp_guestaccount(-1))
	    && (*smb_apasswd == 0))
		guest = True;

	if (!guest && !check_hosts_equiv(user))
	{
		NTSTATUS status;
		/* 
		 * Check with orig_user for security=server and
		 * security=domain.
		 */

		DEBUG(10, ("Checking SMB password, user %s domain %s\n",
			   user, domain));
		status = password_ok(orig_user, domain,
				     smb_apasswd, smb_apasslen,
				     smb_ntpasswd, smb_ntpasslen,
				     NULL, &info3);

		if (!NT_STATUS_IS_OK(status))
		{
			DEBUG(0, ("SMB LM/NT Password did not match "
				  "for `%s\\%s' (%s): %s!\n",
				  domain, orig_user, user,
				  get_nt_error_msg(status)));

			if (lp_security() >= SEC_USER)
			{
				if (lp_map_to_guest() == NEVER_MAP_TO_GUEST)
				{
					DEBUG(1, ("Rejecting user '%s': authentication failed\n", user));
					END_PROFILE(SMBsesssetupX);
					return bad_password_error(inbuf, outbuf);
				}

				if (lp_map_to_guest() == MAP_TO_GUEST_ON_BAD_USER
				    && status != NT_STATUS_NO_SUCH_USER)
				{
					DEBUG(1, ("Rejecting user '%s\\%s': "
						  "known user but wrong pw\n",
						  domain, orig_user));
					END_PROFILE(SMBsesssetupX);
					return bad_password_error(inbuf, outbuf);
				}

				/*
				 * ..else if lp_map_to_guest() == MAP_TO_GUEST_ON_BAD_PASSWORD
				 * Then always map to guest account - as done below.
				 */
			}

			if (*smb_apasswd || !Get_Pwnam(user, True))
				pstrcpy(user, lp_guestaccount(-1));
			DEBUG(3,
			      ("Registered username %s for guest access\n",
			       user));
			guest = True;
		}
	}

	if (!Get_Pwnam(user, True))
	{
		DEBUG(3, ("No such user %s [%s] - using guest account\n", user, domain));
		pstrcpy(user, lp_guestaccount(-1));
		guest = True;
	}

	if (!strequal(user, lp_guestaccount(-1)) &&
	    lp_servicenumber(user) < 0)
	{
		add_home_service(user,get_user_home_dir(user));
	}


	/* it's ok - setup a reply */
	if (Protocol < PROTOCOL_NT1)
	{
		set_message(outbuf, 3, 0, True);
	} else {
		char *p;
		set_message(outbuf, 3, 0, True);
		p = smb_buf(outbuf);
		p += srvstr_push(outbuf, p, "Unix", -1, STR_TERMINATE|STR_ALIGN);
		p += srvstr_push(outbuf, p, PACKAGE_NAME, -1, STR_TERMINATE|STR_ALIGN);
		p += srvstr_push(outbuf, p, global_myworkgroup, -1, STR_TERMINATE|STR_ALIGN);
		set_message_end(outbuf, p);
		/* perhaps grab OS version here?? */
	}

	/* Set the correct uid in the outgoing and incoming packets
	   We will use this on future requests to determine which
	   user we should become.
	 */
	{
		const struct passwd *pw = Get_Pwnam(user, False);
		if (!pw)
		{
			DEBUG(1, ("Username %s is invalid on this system\n", user));
			END_PROFILE(SMBsesssetupX);
			return bad_password_error(inbuf, outbuf);
		}
		gid = pw->pw_gid;
		uid = pw->pw_uid;
	}

	if (guest)
	{
		DEBUG(2, ("reply_sesssetup_and_X: Setting guest bit on smb for %s\\%s (%s)\n",
			  domain, orig_user, user));
		SSVAL(outbuf, smb_vwv2, 1);
	}

	/* register the name and uid as being validated, so further connections
	   to a uid can get through without a password, on the same VC */

	sess_vuid = register_vuid(uid, gid,
				  user, sesssetup_user,
				  guest, VUID_PUBLIC,
				  &info3);

	SSVAL(outbuf, smb_uid, sess_vuid);
	SSVAL(inbuf, smb_uid, sess_vuid);

	if (!done_sesssetup)
		max_send = MIN(max_send, smb_bufsize);

	DEBUG(6, ("Client requested max send size of %d\n", max_send));

	done_sesssetup = True;

	END_PROFILE(SMBsesssetupX);
	return chain_reply(inbuf, outbuf, length, bufsize);
}
