/* 
 *  Unix SMB/Netbios implementation.
 *  Version 1.9.
 *  RPC Pipe client / server routines
 *  Copyright (C) Andrew Tridgell              1992-1997,
 *  Copyright (C) Luke Kenneth Casson Leighton 1996-1997,
 *  Copyright (C) Paul Ashton                       1997,
 *  Copyright (C) Lars Kneschke                     2000. 
 *  
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *  
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software Foundation,
 *  Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */


#include "includes.h"
#include "libsamba.h"
#include "rpc_parse.h"
#include "nterr.h"
#include "winerror.h"
#include "others.h"


/****************************************************************************
  set reg name 
****************************************************************************/
static BOOL set_policy_reg_name(struct policy_cache *cache, POLICY_HND * hnd,
				fstring name)
{
	char *dev = strdup(name);
	if (dev != NULL)
	{
		if (set_policy_state(cache, hnd, NULL, (void *)dev))
		{
			DEBUG(3, ("Registry setting policy name=%s\n", name));
			return True;
		}
		free(dev);
	}

	DEBUG(3, ("Error setting policy name=%s\n", name));
	return False;
}

/****************************************************************************
  get reg name 
****************************************************************************/
static BOOL get_policy_reg_name(struct policy_cache *cache, POLICY_HND * hnd,
				fstring name)
{
	char *dev = (char *)get_policy_state_info(cache, hnd);

	if (dev != NULL)
	{
		fstrcpy(name, dev);
		DEBUG(5, ("getting policy reg name=%s\n", name));
		return True;
	}

	DEBUG(3, ("Error getting policy reg name\n"));
	return False;
}

/*******************************************************************
 _reg_close
 ********************************************************************/
uint32 _reg_close(POLICY_HND * pol)
{
	/* close the policy handle */
	if (!close_policy_hnd(get_global_hnd_cache(), pol))
	{
		return ERROR_INVALID_HANDLE;
	}
	return ERROR_SUCCESS;
}

/*******************************************************************
 _reg_open
 ********************************************************************/
uint32 _reg_open_hklm(PRS_HANDLE *con_hnd, POLICY_HND * pol, uint32 access_mask)
{
	/* get a (unique) handle.  open a policy on it. */
	if (!open_policy_hnd(get_global_hnd_cache(), get_sec_ctx(),
			     pol, access_mask))
	{
		return ERROR_FILE_NOT_FOUND;
	}

	return ERROR_SUCCESS;
}

/*******************************************************************
 _reg_open_entry
 ********************************************************************/
uint32 _reg_open_entry(const POLICY_HND * pol, const UNISTR2 * uni_name,
		       uint32 unknown_0, uint32 access_mask,
		       POLICY_HND * entry_pol)
{
	fstring name;

	if (find_policy_by_hnd(get_global_hnd_cache(), pol) == -1)
	{
		return ERROR_INVALID_HANDLE;
	}

	if (!open_policy_hnd_link(get_global_hnd_cache(),
				  pol, entry_pol, access_mask))
	{
		return ERROR_TOO_MANY_SECRETS;	/* ha ha very droll */
	}

	unistr2_to_ascii(name, uni_name, sizeof(name) - 1);

	DEBUG(2, ("_reg_open_entry(%s)\n", name));
	/* lkcl XXXX do a check on the name, here */
	if (!strequal(name,
		      "SYSTEM\\CurrentControlSet\\Control\\ProductOptions")
	    && !strequal(name,
			 "Software\\Microsoft\\Windows NT\\CurrentVersion")
	    && !strequal(name,
			 "SYSTEM\\CurrentControlSet\\Services\\NETLOGON\\Parameters\\")
	    && !strequal(name,
			 "System\\CurrentControlSet\\Control\\Terminal Server\\DefaultUserConfiguration")
	    && !strequal(name, ""))
	{
		DEBUG(2, ("winregd: Denying access to %s\n", name));
		return ERROR_ACCESS_DENIED;
		/* return ERROR_FILE_NOT_FOUND; */
	}

	if (!set_policy_reg_name(get_global_hnd_cache(), entry_pol, name))
	{
		return ERROR_TOO_MANY_SECRETS;	/* ha ha very droll */
	}

	return ERROR_SUCCESS;
}

/*******************************************************************
 _reg_info
 ********************************************************************/
uint32 _reg_info(POLICY_HND *pol, const UNISTR2 *uni_valuename,
		 BUFFER2 *buf, uint32 *type)
{
	fstring name;
	char *valuename;

	if (!type)
		return ERROR_INVALID_PARAMETER;

	*type = REGVAL_NONE;

	if (!get_policy_reg_name(get_global_hnd_cache(), pol, name))
	{
		return ERROR_INVALID_HANDLE;
	}

	valuename = unistr2_to_ascii(NULL, uni_valuename, 0);

	DEBUG(3, ("_reg_info: query for [%s]\\[%s]\n",
		  name, valuename));

	if (strequal(name,
		     "Software\\Microsoft\\Windows NT\\CurrentVersion"))
	{
		const char *key = NULL;
		if (strequal(valuename, "CurrentVersion"))
			key = "4.0";
		if (strequal(valuename, "CurrentBuildNumber"))
			key = "1381";
		if (strequal(valuename, "CSDVersion"))
			key = "Service Pack 6";

		if (key)
		{
			make_buffer2(buf, key, strlen(key)+1);
			*type = REGVAL_SZ;
		}
	}

	if (strequal(name,
		     "SYSTEM\\CurrentControlSet\\Control\\ProductOptions"))
	{
		const char *key;
		switch (lp_server_role())
		{
			case ROLE_DOMAIN_PDC:
			case ROLE_DOMAIN_BDC:
			{
				key = "LanmanNT";
				break;
			}
			case ROLE_STANDALONE:
			{
				key = "ServerNT";
				break;
			}
			case ROLE_DOMAIN_MEMBER:
			default:	/* shut up gcc unused var warning */
			{
				key = "WinNT";
				break;
			}
		}
		if (strequal(valuename, "ProductSuite"))
			key = "Enterprise";
		make_buffer2(buf, key, strlen(key)+1);
		*type = REGVAL_SZ;
	}

	if (*type == REGVAL_NONE)
	{
		DEBUG(2, ("_reg_info: Rejecting request for [%s]\\[%s]\n",
			  name, valuename));
		safe_free(valuename);
		return ERROR_FILE_NOT_FOUND;
	}

	safe_free(valuename);

	return ERROR_SUCCESS;
}
